
##
##   dfgbench.sh
##
##   Copyright (C) Stephan Zimmer
##
##   This program is free software; you can redistribute
##   it and/or modify it under the terms of the GNU
##   General Public License as published by the Free
##   Software Foundation; either version 2 of the License,
##   or (at your option) any later version.
##
##   This program is distributed in the hope that it will
##   be useful, but WITHOUT ANY WARRANTY; without even
##   the implied warranty of MERCHANTABILITY or FITNESS
##   FOR A PARTICULAR PURPOSE.  See the GNU General Public
##   License for more details.
##
##   You should have received a copy of the GNU General
##   Public License along with this program; if not, write
##   to the Free Software Foundation, Inc., 59 Temple
##   Place, Suite 330, Boston, MA  02111-1307  USA
##
##
##
##  File       : "dfgbench"
##  Author     : Stephan Zimmer <st.zimmer@web.de>
##
##  Description: Runs SPASS and compares the DFG status to the SPASS result;
##               furhtermore the total time SPASS spent is measured using the
##               UNIX time utility; output is of the form
##                   (OK|MISMATCH) DFG_STATUS SPASS_RESULT TIME
##               where DFG_STATUS,SPASS_RESULT e {"Unsatisfiable",
##                   "Satisfiable", "Unknown"}
##
##  Usage      : There are two ways invoking the script:
##                 dfgbench <DFG FILE> (SPASS+T input file)
##                   [ --spass_quiet ] (do not show the output of SPASS)
##                   [<LIST OF SPASS OPTIONS>] (further SPASS
##                   options, e.g. -DocProof=1)
##
##                 dfgbench --(o)all (<OUTPUT FILENAME>) [--total] [--spass_quiet]
##                   (do not show the output of SPASS)
##                   [<LIST OF SPASS OPTIONS>] (further SPASS
##                   options, e.g. -DocProof=1)
##
##                 dfgbench --(o)list <LIST OF DFG FILES> (<OUTPUT FILENAME>) [--total] [--spass_quiet]
##                   (do not show the output of SPASS)
##                   [<LIST OF SPASS OPTIONS>] (further SPASS
##                   options, e.g. -DocProof=1)
##
##               In case either the "all" or the "list" mode is used, errors are written
##               to the file "$0.errors".
##
##  Note       : This tool depends on the format of the SPASS STDOUT output
##               and represents only a pragmatic solution rather than a
##               real generic one.
##


##
## global settings
##
SCRIPT_NAME="${0}"
EXIT_SUCCESS=0
EXIT_FAILURE=1
if [ "${TSPASS}" == "" ] ; then
    SPASS="TSPASS"
else
    SPASS="${TSPASS}"
fi
nl='
'
tab='    '
temp_output="tmp~"
pattern_spass_result="SPASS beiseite: "
pattern_spass_spent_time="SPASS spent"
pattern_dfg_status="status("
dfg_status=""
spass_result=""
spent_time=""
spass_return=0
spass_quiet=0
time_util=""

##
## check for existence of UNIX time utility
##
time_util="$(which time)"
if [ "$time_util" == "" ] ; then
    echo "Error, this tool needs the UNIX \"time\" utility in orderto measure process time."
    exit ${EXIT_FAILURE}
fi

##
## check and extract command line arguments
##
if [[ ( -z "${1}" ) || ( "${1}" == "--help" ) ]] ; then
    echo "There are three different forms of invoking the script:"
    echo "${tab}Usage 1: $(basename "${SCRIPT_NAME}") <DFG FILE> (--spass_quiet) (<LIST OF SPASS OPTIONS>)"
    echo "${tab}${tab}proceeds file <DFG_FILE>."
    echo "${tab}Usage 2: $(basename "${SCRIPT_NAME}") --{o}all {<OUTPUT FILENAME>} (--total) (--spass_quiet) (<LIST OF SPASS OPTIONS>)"
    echo "${tab}${tab}proceeds all \".dfg\" files in current directory."
    echo "${tab}Usage 3: $(basename "${SCRIPT_NAME}") --{o}list <LIST OF DFG FILES> {<OUTPUT FILENAME>} (--total) (--spass_quiet) (<LIST OF SPASS OPTIONS>)"
    echo "${tab}${tab}proceeds all files from <LIST OF DFG FILES>."
    echo "${tab}--spass_quiet: do not show SPASS output"
    echo "${tab}--total: at end of the run give the total time spent by SPASS"
    echo "${tab}In case \"--oall\" or \"--olist\" is used, give the name of the file containing the output."
    echo "Note that all options have to be given in the same order as described above. "
    echo -n "${nl}"
    if [ -z "${1}" ] ; then
	exit ${EXIT_FAILURE}
    else
	exit ${EXIT_SUCCESS}
    fi
fi

##
## shall the script be applied to all DFGs that are located in
## the current directory or to a given list of DFG files?
##
if [[ ( ( "${1}" == "--all" ) || ( "${1}" == "--list" ) ) ||
	    ( ( "${1}" == "--oall" ) || ( "${1}" == "--olist" ) ) ]] ; then
    output_log_filename=""
    if [ "${1}" == "--all" ] ; then
	shift 1
	dfgs_to_proceed="$(ls *.dfg)"
    elif [ "${1}" == "--oall" ] ; then
	shift 1
	dfgs_to_proceed="$(ls *.dfg)"
	output_log_filename="${1}"
	shift 1
	if [ "${output_log_filename}" == "" ]; then
	    echo "Give name of output file as argument to option \"--oall\"."
	    exit ${EXIT_FAILURE}
	fi
    else
	tmp="${1}"
	shift 1
	list_of_dfgs="${1}"
	shift 1

	if [ -z "${list_of_dfgs}" ] ; then
	    echo "Error, give filename of file containing DFG files after \"--list\".${nl}"
	    exit ${EXIT_FAILURE}
	fi
	if [ ! -e "${list_of_dfgs}" ] ; then
	    echo "Error, argument of \"--list\" \"${list_of_dfgs}\" does not exist.${nl}"
	    exit ${EXIT_FAILURE}
	fi
	dfgs_to_proceed=$(cat "${list_of_dfgs}")

	if [ "${tmp}" == "--olist" ] ; then
	   output_log_filename="${1}"
	   shift 1

	   if [ "${output_log_filename}" == "" ]; then
	       echo "Give name of output file as argument to option \"--olist\"."
	       exit ${EXIT_FAILURE}
	   fi
	fi
    fi

    if [ "${output_log_filename}" != "" ] ; then
	rm -rf "${output_log_filename}"
    fi

    if [ "${1}" == "--total" ] ; then
	shift 1
	total_secs="0"
	total_msecs="0"
	total_mode_errors=""
	for d in ${dfgs_to_proceed} ; do
	    result=$("${SCRIPT_NAME}" "$d" $@)
	    echo "${result}"
	    if [ ! "${output_log_filename}" == "" ] ; then
		echo "${result}" >> "${output_log_filename}"
	    fi
	    secs=$(echo $result | egrep "[0-9]+s" -o | egrep "[0-9]+" -o)
	    msecs=$(echo $result | egrep "s[0-9]+ms" -o | egrep "[0-9]+" -o)
	    if [[ ( "${secs}" == "" ) || ( "${msecs}" == "" ) ]] ; then
		total_mode_errors="${total_mode_errors}${tab}Error, could not extract times for \"${d}\".${nl}"
	    else
		secs_pending=$(echo "(${msecs} + ${total_msecs}) / 100" | bc)
		total_msecs=$(echo "(${msecs} + ${total_msecs}) % 100" | bc)
		total_secs=$(echo "${secs} + ${total_secs} + ${secs_pending}" | bc)
	    fi
	done
	if [ "$(echo "${total_msecs} < 10"|bc)" == "1" ] ; then
	    total_msecs="0${msecs}"
	fi
	spent_time="${total_secs}s${total_msecs}ms"
	echo "${nl}--------------------------------------------------------------"
	echo "Total spent time: ${spent_time}${nl}"
	if [ "${total_mode_errors}" != "" ]; then
	    echo "Note that the global time measuring is affected by the following errors: "
	    echo "${total_mode_errors}${nl}"
	fi
	if [ "${output_log_filename}" != "" ] ; then
	    echo "${nl}--------------------------------------------------------------" >> "${output_log_filename}"
	    echo "Total spent time: ${spent_time}${nl}" >> "${output_log_filename}"
	    if [ "${total_mode_errors}" != "" ]; then
		echo "Note that the global time measuring is affected by the following errors: " >> "${output_log_filename}"
		echo "${total_mode_errors}${nl}" >> "${output_log_filename}"
	    fi
	fi
    else
	if [ "${output_log_filename}" == "" ] ; then
	    for d in ${dfgs_to_proceed} ; do
		"${SCRIPT_NAME}" "$d" $@
	    done
	else
	    for d in ${dfgs_to_proceed} ; do
		"${SCRIPT_NAME}" "$d" $@ | tee -a "${output_log_filename}"
	    done
	fi
    fi

    exit ${EXIT_SUCCESS}
fi

##
## normal mode, invoke SPASS on given file
##
dfg="${1}"
if [ ! -e "${dfg}" ] ; then
    echo "Error, file \"${dfg}\" does not exist.${nl}"
    exit ${EXIT_FAILURE}
fi

shift 1
if [ "${1}" == "--total" ]; then
    echo "Error, parameter \"--total\" is not allowed in this mode.${nl}"
    exit ${EXIT_FAILURE}
fi

rm -rf "${temp_output}"
if [ "${1}" == "--spass_quiet" ] ; then
    spass_quiet=1
    shift 1
    "${time_util}" -p "${SPASS}" "${dfg}" $@ > "${temp_output}" 2>&1
else
    "${time_util}" -p "${SPASS}" "${dfg}" $@ 2>&1 | tee -a "${temp_output}"
fi
spass_return=$?

##
## extract status out of DFG file
##
status=$(grep "${pattern_dfg_status}" "${dfg}")
if [ $? != 0 ] ; then
    echo "${dfg} ERROR, could not extract status out of DFG file.${nl}"
    exit ${EXIT_FAILURE}
fi

echo "${status}" | grep "unsatisfiable" > /dev/null
if [ $? == 0 ] ; then
    dfg_status="Unsatisfiable"
else
    echo "${status}" | grep "satisfiable" > /dev/null
    if [ $? == 0 ] ; then
	dfg_status="Satisfiable"
    else
	## given that DFG input is syntactically wellformed
	dfg_status="Unknown"
    fi
fi

##
## extract status out of SPASS output
##
status=$(grep "${pattern_spass_result}" "${temp_output}")
if [ $? != 0 ] ; then
    echo "${dfg} ERROR, could not extract result out of SPASS output.${nl}"
    exit ${EXIT_FAILURE}
fi

echo "${status}" | grep "Proof found" > /dev/null
if [ $? == 0 ] ; then
    spass_result="Unsatisfiable"
else
    echo "${status}" | grep "Satisfiable" > /dev/null
    if [ $? == 0 ] ; then
	spass_result="Satisfiable"
    else
	echo "${status}" | grep "Ran out of time." > /dev/null
	if [ $? == 0 ] ; then
	    spass_result="Timeout"
	else
	    spass_result="Unknown"
	fi
    fi
fi

##
## extract time found by UNIX time utility
##
extract=$(egrep "user [0-9]+\.[0-9][0-9]" "${temp_output}")
user_secs=$(echo "$extract" | egrep "[0-9]+\." -o | egrep "[0-9]+" -o)
user_msecs=$(echo "$extract" | egrep "\.[0-9][0-9]" -o | egrep "[0-9]+" -o)
extract=$(egrep "sys [0-9]+\.[0-9][0-9]" "${temp_output}")
sys_secs=$(echo "$extract" | egrep "[0-9]+\." -o | egrep "[0-9]+" -o)
sys_msecs=$(echo "$extract" | egrep "\.[0-9][0-9]" -o | egrep "[0-9]+" -o)

msecs=$(echo "(${user_msecs} + ${sys_msecs}) % 100" | bc)
if [ "$(echo "${msecs} < 10"|bc)" == "1" ] ; then
    msecs="0${msecs}"
fi

secs=$(echo "(${user_secs} + ${sys_secs}) + ((${user_msecs} + ${sys_msecs}) / 100)" | bc)
spent_time="${secs}s${msecs}ms"

##
## provide output
##
if [ ${spass_quiet} -ne 1 ] ; then
    echo "${nl}"
fi
echo -n "${dfg} "
if [ "${dfg_status}" == "Unknown" ] ; then
    echo -n "OK "
elif [ "${dfg_status}" == "${spass_result}" ] ; then
    echo -n "OK "
else
    echo -n "MISMATCH "
fi
echo "${dfg_status} ${spass_result} ${spent_time}"
if [ ${spass_quiet} -ne 1 ] ; then
    echo ""
fi

## erase temporary file
rm -rf "${temp_output}"

exit ${spass_return}
