
##
##   prf2dot.pl
##
##   Copyright (C) Stephan Zimmer
##
##   This program is free software; you can redistribute
##   it and/or modify it under the terms of the GNU
##   General Public License as published by the Free
##   Software Foundation; either version 2 of the License,
##   or (at your option) any later version.
##
##   This program is distributed in the hope that it will
##   be useful, but WITHOUT ANY WARRANTY; without even
##   the implied warranty of MERCHANTABILITY or FITNESS
##   FOR A PARTICULAR PURPOSE.  See the GNU General Public
##   License for more details.
##
##   You should have received a copy of the GNU General
##   Public License along with this program; if not, write
##   to the Free Software Foundation, Inc., 59 Temple
##   Place, Suite 330, Boston, MA  02111-1307  USA
##
##
##  File       : "prf2dot"
##  Author     : Stephan Zimmer <st.zimmer@web.de>
##
##  Description: Perl script for visualization of SPASS proof files (created by
##               command line option "-FPDFGPRoof") in the same style as
##               with "spassproof2viz".
##
##  Usage      : prf2dot <input filename> (name of input .prf file)
##                       [<output filename>] (name of output file, if dropped
##                            output is given on stdout)
##                       [-partition_branches={0,1}] (whether branches are
##                            encapsulated in boxes)
##                       [-information_node={0,1}] (whether an information node
##                            should be drawn)
##
##
## needed package
##
use Getopt::Long;

##
## return values
##
$EXIT_SUCCESS=0;
$EXIT_FAILURE=1;

##
## global configuration switches
##
$color_information_node="lightgrey";
$color_derived_formula="cornflowerblue";
$color_input_formula="green";
$color_theory_axiom_formula="violet";
#$color_split_part_formula="orange";
$color_split_formula="orange";
$color_empty_clause="red";
$normal_edge_color="green";
$split_edge_color="orange";
$theory_refutation_edge_color="blue";
$color_branch_component="gold";
# do we partition the graph into subgraphs, for every branch one?
$partition_branches=1;
# do we provide split information?
$show_split_information=1;
# do we provide an additional node containing information?
$show_information_node=1;
# our dictionary, mapping clause numbers to positions
%branch_dictionary=();
# our current position, operational view
$branch_current_position="";
# signature we insert as command into file
$signature = "/*\n\tVISUALIZATION OF SPASS(+T) PROOFS\n\n\tThis file was created using prf2dot\n\tAuthor: st.zimmer\@web.de \n*/\n\n";

##
## function definitions
##

##
## specialized functions working on positions (previously in a package)
##
sub positions_empty
{
    return "e";
}

sub positions_is_root
{
    return (&positions_empty eq $_[0]);
}

sub positions_left_successor
{
    if (&positions_is_root($_[0])) {
	return "0";
    } else {
	return $_[0]."0";
    }
}

sub positions_right_successor
{
    if (&positions_is_root($_[0])) {
	return "1";
    } else {
	return $_[0]."1";
    }
}

sub positions_swap_branch
{
    my $tmp="";

    if (&positions_is_root($_[0])) {
	return &positions_empty;
    } else {
	$tmp = substr($_[0], 0, length($_[0])-1);
	if (substr($_[0], -1) eq "0") {
	    return &positions_right_successor($tmp);
	} else {
	    return &positions_left_successor($tmp);
	}
    }
}

sub positions_equal
{
    return ($_[0] eq $_[1]);
}

sub positions_prefix_order_strict
{
    my $i=0;

    if (&positions_is_root($_[0])) {
	return (! &positions_is_root($_[1]));
    }

    if (length($_[0]) >= length($_[1])) {
	return 0;
    }

    for($i=0; $i<length($_[0]); ++$i) {
	if (substr($_[0], $i, 1) ne substr($_[1], $i, 1)) {
	    return 0;
	}
    }

    return 1;
}

sub positions_branch_order_strict
{
    my $i=0, $max=0;

    if ((&positions_is_root($_[0])) || (&positions_is_root($_[1]))) {
	return 0;
    }

    if (length($_[0]) < length($_[1])) {
	$max = length($_[0]);
    } else {
	$max = length($_[1]);
    }

    for($i=0; $i<$max; ++$i) {
	if ((substr($_[0],$i, 1) eq "0") && (substr($_[1],$i, 1) eq "1")) {
	    return 1;
	} elsif ((substr($_[1],$i, 1) eq "0") && (substr($_[0],$i, 1) eq "1")) {
	    return 0;
	}
    }
    return 0;
}

sub positions_temporal_visiting_order_strict
{
    if (&positions_prefix_order_strict($_[0], $_[1])) {
	return 1;
    } else {
	return &positions_branch_order_strict($_[0], $_[1]);
    }
}

sub positions_temporal_visiting_order_equal
{
    if (&positions_equal($_[0], $_[1])) {
	return 1;
    } else {
	return &positions_temporal_visiting_order_strict($_[0], $_[1]);
    }
}

sub positions_max2
{
    return $_[1] if &positions_temporal_visiting_order_equal($_[0], $_[1]);
    return $_[0];
}

sub positions_max3
{
    return &positions_max2($_[0], &positions_max2($_[1], $_[2]));
}



##      --dec2bin--
## converts a number in decimal representation into binary representation
## --> $1: number in decimal representation
## <-- number of binary representation
sub dec2bin
{
    my $str = unpack("B32", pack("N", shift));
    $str =~ s/^0+(?=\d)//;
    return $str;
}

##      --proceed_proof_step--
## tries to match the proof step given as string (taken from the "Perl
## Kochbuch", recipy 2.4, pp. 51)
## --> $1: proof step representation
## --> $2: whether formulae should be negated (1) or not (0);
##         this is needed for conjecture clauses
## <-- whether matching was successful (EXIT_SUCCESS) or not (EXIT_FAILURE)
sub proceed_proof_step
{
    my $clause_no="";
    my $clause_split_level="";
    my $clause_origin="";
    my $clause_parent_clauses="";
    my $clause="";
    my $to_be_matched = $_[0];
    my $to_be_negated = $_[1];
    my $color="";
    my $additional_node_information="";

    ## try to match proof step
    if ($to_be_matched =~ m/
	clause               # keyword
	\(                   # open bracket
	(.*)                 # clause itself in $1
	\s*,\s*              # seperator
	(\d+)                # clause number in $2
	\)                   # closed bracket
	\.                   # marking end of line
	/x                   # x for "here documentation"
	)
    {
	$clause = $1;
	if ($to_be_negated == 1) {
	    $clause = "not(".$clause.")";
	}
	$clause_no = $2;
	$clause_origin = "Inp";
	$clause_split_level = 0;
	$clause_parent_clauses="";

	if ($show_split_information != 0) {
	    $additional_node_information="[$clause_split_level]";
	}

	## add new node
	$nodes .= "\tnode [shape=box,color=$color_input_formula,style=filled,label=\"\\N$additional_node_information:$clause \"]; $clause_no\n";

	## update dictionary
	$branch_dictionary{$clause_no} = &positions_empty;

	return ${EXIT_SUCCESS};

    } elsif ($to_be_matched =~ m/
	     step                 # keyword
	     \(                   # open bracket
	     (\d+)                # clause number in $1
	     \s*,\s*              # seperator
	     (.*)                 # clause itself in $2
	     \s*,\s*              # seperator
	     (App|EmS|SoR|EqR|EqF|MPm|SpR|SPm|OPm|SpL|Res|SHy|OHy|URR|Fac|Spt|Inp|Rew|CRw|Con|AED|Obv|SSi|MRR|UnC|Def|Ter|ThI|IOE|ArS|ThA|ThR|BLD|OCE|OCh|TOC)
	                          # clause origin in $3
	     \s*,\s*              # seperator
	     \[                   # marker
	     (\d+(\.\d*)?)(([ ]*,[ ]*\d+(\.\d*)?)*)
	     # list of parent clauses in $4,$6
	     \]                   # marker
	     \s*,\s*              # seperator
	     \[                   # marker
	     splitlevel\s*:\s*    # keyword
	     (\d+)                # splitlevel in $7
	     \]                   # marker
	     \)                   # closed bracket
	     \.                   # marking end of line

	     /x                   # x for "here documentation"
	     )
    {

	## extract information out of matching
	$clause_no=$1;
	$clause=$2;
	$clause_origin=$3;
	$clause_parent_clauses=$4.$6;
	$clause_split_level=$9;
	if ($to_be_negated == 1) {
	    $clause = "not(".$clause.")";
	}

	my @parents=split(",", $clause_parent_clauses);
	my $len=0;
	foreach $d (@parents) {
	    $d =~ s/(\d+)(\.?\d*)/$1/;
	    $len++;
	}


	## extract elements from parent that are unique
	my %seen = ();
        my @uparents = ();
	foreach $item (@parents) {
	    push(@uparents, $item) unless $seen{$item}++;
	}
	$len = @uparents;

	## if node results from splitting, re-draw splitted clause in special color
	if (($clause_origin eq "Spt") && ($len == 1)) {
	    $parent=$uparents[0];
	    $nodes =~ s/.*node \[.*(label.*).*\];.*$parent/\tnode \[shape=box,color=$color_split_formula,style=filled,$1\]; $parent/;
	}

	## draw edge in suitable color
	if ($clause_origin eq "Spt") {
	    $color=$split_edge_color;
	} elsif ($clause_origin eq "ThR") {
	    $color=$theory_refutation_edge_color;
	} else {
	    $color=$normal_edge_color;
	}

	my $additional_origin_comment="";
	foreach $d (@uparents) {
	    $additional_origin_comment="";
	    if (($clause_origin eq "Spt") && ($len == 1)) {
		# left splitting edge
		$additional_origin_comment="left";
	    } elsif (($clause_origin eq "Spt") && ($len >= 2)) {
		# right splitting edge (len==2: right split part,
		# len==3: negation of left split part)
		$additional_origin_comment="right";
	    }
	    $edges .= "\t$d -> $clause_no [color=$color,label=\"$clause_origin $additional_origin_comment\"]\n";
	}

	## add new node in suitable color (empty vs. non-empty clause)
	if ($clause =~ m/or\(false\)*/) {
	    $color=$color_empty_clause;
	} elsif ($clause_origin eq "ThA") {
	    $color=$color_theory_axiom_formula;
	} else {
	    $color=$color_derived_formula;
	}

	## do we also output splitting information for clauses?
	if ($show_split_information != 0) {
	    $additional_node_information="[$clause_split_level]";
	}

	$nodes .= "\tnode [shape=box,color=$color,style=filled,label=\"\\N$additional_node_information:$clause\"]; $clause_no\n";

	## update dictionary
	if ($clause_origin eq "Spt") {
	    my $new_pos="";
	    if ($len == 1) {
		# clause is left split part of split clause
		# $new_pos = $branch_dictionary{$uparents[0]};
		# $new_pos = &positions_left_successor($new_pos);
		# we must not take the new position from the position of the parent
		#  clause of the split since it could happen, that the parent clause
		#  does not depend on the most recent split; this would yield two
		#  branches with the same splitting branch as label (this corresponds
		#  to the dichotomy between the operational and the abstract view of the
		#  inferences)
		$branch_current_position = &positions_left_successor($branch_current_position);
		$new_pos = $branch_current_position;
	    } elsif (($clause_origin eq "Spt") && ($len == 2)) {
		# right splitting edge, len==2: right split part
		$new_pos = &positions_max2($branch_dictionary{$uparents[0]},
					    $branch_dictionary{$uparents[1]});
		$new_pos = &positions_swap_branch($new_pos);
		$branch_current_position = $new_pos;
	    } elsif (($clause_origin eq "Spt") && ($len == 3)) {
		# right splitting edge, len==3: negation of left split part
		$new_pos = &positions_max3($branch_dictionary{$uparents[0]},
					    $branch_dictionary{$uparents[1]},
					    $branch_dictionary{$uparents[2]});
		$new_pos = &positions_swap_branch($new_pos);
		$branch_current_position = $new_pos;
	    }
	    $branch_dictionary{$clause_no} = $new_pos;
	} else {
	    my $new_pos=$branch_dictionary{$uparents[0]};
	    foreach $d (@uparents) {
		$new_pos = &positions_max2($new_pos, $branch_dictionary{$d});
	    }
	    $branch_dictionary{$clause_no} = $new_pos;
	}

	return ${EXIT_SUCCESS};
    } else {
	return ${EXIT_FAILURE}
    }
}

##      --show_commandline_usage--
## shows the command line options
##
sub show_commandline_usage
{
    print "Usage: $0 <input filename> [<output filename>] [options]\n\n";
    print "\t<input filename>: name of file containing the proof in DFG format\n";
    print "\t<output filename>: name of file containing visualization in DOT format;\n";
    print "\t                   if not given, it is printed to STDOUT\n";
    print "\t-partition_branches={0,1}: whether branches should be drawn clustered\n";
    print "\t-information_node={0,1}: whether an additional information node should be drawn\n";
    print "\n\n";
    return $EXIT_SUCCESS;
}


##
## main entrance
##

## proceed command line arguments
$proof_file = "";
$output_file_name = "";
my $help = 0;
GetOptions( '<>' =>
	    sub {
		if ($proof_file eq "") {
		    $proof_file = $_[0];
		} else {
		    $output_file_name = $_[0];
		}
	    },
	    "partition_branches=i" => \$partition_branches,
	    "help+" => \$help,
	    "information_node=i" => \$show_information_node);

if ($help == 1) {
    &show_commandline_usage();
    exit $EXIT_SUCCESS;
}

if ($proof_file eq "") {
    print "Error: no proof file given.\n\n";
    exit $EXIT_FAILURE;
}

if ( ! -e $proof_file ) {
    print "Error, file \'$proof_file\' does not exist.\n\n";
    exit $EXIT_FAILURE;
}

## prepare global variables
$graph="";
$nodes="";
$edges="";
$clusters="";

## read file
open(proof_file_handle, "< $proof_file") || die "File $proof_file not found.";
@all_lines = <proof_file_handle>;
close(proof_file_handle);

## for operational view on branching
$branch_current_position=&positions_empty;

## extract header information and filter out input clauses and proof steps
my @proof_steps = ();
my $dfg_author = "";
my $dfg_status = "";
my $dfg_description = "";
my $dfg_name = "";
my $flag_conjectures = 0;

foreach $current_line (@all_lines)
{
    if ($current_line =~ m/
	name                  # keyword
	\(                    # open bracket
	{\*                   # start marker: "{*"
	(.*)                  # name in $1
	\*}                   # start marker: "}*"
	\)                    # closed bracket
	/x                    # x for "here documentation"
	) {
	$dfg_name = $1;
    } elsif ($current_line =~ m/
	status                # keyword
	\(                    # open bracket
	\s*(satisfiable|unknown|unsatisfiable)\s*
                              # logical status in $1
	\)                    # closed bracket
	/x                    # x for "here documentation"
	) {
	$dfg_status = $1;
    } elsif ($current_line =~ m/
	author                # keyword
	\(                    # open bracket
	{\*                   # start marker: "{*"
	(.*)                  # author in $1
	\*}                   # start marker: "}*"
	\)                    # closed bracket
	/x                    # x for "here documentation"
	) {
	$dfg_author = $1;
    } elsif ($current_line =~ m/
	description           # keyword
	\(                    # open bracket
	{\*                   # start marker: "{*"
	(.*)                  # description in $1
	\*}                   # start marker: "}*"
	\)                    # closed bracket
	/x                    # x for "here documentation"
	) {
	$dfg_description = $1;
    } elsif ($current_line =~ m/
	(clause|step)       # keyword indicating an entry contributing to the proof
	\(                  # open bracket
	.*                  # details
	\)                  # closed bracket
        \.                  # end of entry
	/x                  # x for "here documentation"
	) {
	if ($flag_conjectures == 1) {
	    $current_line = "C".$current_line;   # hack
	}
	push(@proof_steps, $current_line);
    } elsif ($current_line =~ m/list_of_clauses\(conjectures.*/) {
	$flag_conjectures = 1;
    } elsif (($flag_conjectures == 1) && ($current_line =~ m/end_of_list\./)) {
	$flag_conjectures = 0;
    } else {
	# ignore line
    }
}

## build a node containing additional information
my $information_node="\tnode [shape=box,color=$color_information_node,style=filled,label=\"Problem: $dfg_name\\nAuthor: $dfg_author\\nDescription: $dfg_description\\nStatus: $dfg_status\"]; 0\n";

## proceed all proof steps
my $proceed_proof_step_result = 0;
foreach $current_line (@proof_steps)
{
    if ($current_line =~ m/^C(.*)/) {
	$proceed_proof_step_result = proceed_proof_step($1, 1);
    } else {
	$proceed_proof_step_result = proceed_proof_step($current_line, 0);
    }


    if ($proceed_proof_step_result == $EXIT_FAILURE)
    {
	die("\n\n\nError, could not match line \'$current_line\'\n\n\n");

    }
}

## build the graph of the nodes and edges
if ($show_information_node == 1) {
    $graph = $signature."digraph TSPASS {\n\tfontsize=14;\n\n".$information_node.$nodes."\n".$edges."\n";
} else {
    $graph = $signature."digraph TSPASS {\n\tfontsize=14;\n\n".$nodes."\n".$edges."\n";
}

## if we provide an information node draw into an own subgraph
if ($show_information_node == 1) {
    $clusters .= "\tsubgraph cluster_information { color=$color_branch_component; label=\"Information\"; 0 }\n";
}

## do we draw the branches in the graph in seperate regions?
if ($partition_branches != 0) {
    # problem: without reversing a "foreach values(...)" loop on "branch_dictionary"
    #  would consider equal values several times; a foreach on the keys instead
    #  considers the key elements modulo equality
    %r_branch_dictionary=reverse(%branch_dictionary);
    # for every occuring branch print the subgraph cluserting information
    foreach $branch (keys(%r_branch_dictionary)) {
	$clusters .= "\tsubgraph cluster_".$branch." { color=$color_branch_component; label=\"branch ".$branch."\"; ";
	foreach $node (keys(%branch_dictionary)) {
	    if ($branch_dictionary{$node} eq $branch) {
		$clusters .= $node.";";
	    }
	}
	$clusters .= " }\n";
    }
}

$graph .= $clusters;

## terminate the graph
$graph .= "}\n\n";

## print the graph
if ($output_file_name eq "") {
    # print to stdout
    print $graph;
} else {
    # print to file
    open(DOTFILE,">$output_file_name") || die("Cannot open \"".$output_file_name."\"\n");
    print DOTFILE "$graph\n";
    close(DOTFILE);
}


## exist with suitable return code
exit $EXIT_SUCCESS;

