/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 DICTIONARY OF CLAUSES                  * */
/* *                                                        * */
/* *  $Module: clause_dictionary                            * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "clause_dictionary.h"
#include "memory.h"
#include "misc.h"


/**************************************************************/
/* Internal functions                                         */
/**************************************************************/

int clause_dictionary_evector_dummy_equal(void*, void*);
int clause_dictionary_evector_dummy_compare(void*, void*);
void clause_dictionary_evector_dummy_destroy_position_list(void*);

int clause_dictionary_evector_dummy_equal(void* e1, void* e2)
{
  return 0;
}

int clause_dictionary_evector_dummy_compare(void* e1, void* e2)
{
  return 0;
}

void clause_dictionary_evector_dummy_destroy_position_list(void* vector)
{
  evector_deconstructEvector((EVECTOR*) vector,
			     &position_evector_destroyPosition);
}


/**************************************************************/
/* Allocation and disallocation                               */
/**************************************************************/

CLAUSE_DICTIONARY* clause_dictionary_create(int base_size, int expand)
/**************************************************************
  INPUT:   Two numbers.
  RETURNS: A newly allocated sturcture <CLAUSE_DICTIONARY>.
  EFFECT:  Memory is allocated using the memory mode. The two
           <EVECTOR> members are created with base size
           <base_size> and expand factor <expand>. Memory is
           allocated using the memory module.
***************************************************************/
{
  CLAUSE_DICTIONARY* local;

  /* allocate memory for structure itself */
  local = (CLAUSE_DICTIONARY*) memory_Malloc(sizeof(CLAUSE_DICTIONARY));

  /* allocate vector of clause numbers */
  local->clause_numbers = evector_createEvectorInt(base_size, expand);

  /* we are creating a list of lists of positions */
  local->dependent_positions = evector_createEvector(base_size, expand,
						     &clause_dictionary_evector_dummy_equal,
						     &clause_dictionary_evector_dummy_compare);

  /* vector of all positions that appear in <dependent_positions> */
  local->dependent_positions_accumulated =
    evector_createEvector(100, expand,
			  &position_evector_equal, &position_evector_tree_order);

  return local;
}

void clause_dictionary_destroy(CLAUSE_DICTIONARY* dictionary)
/**************************************************************
  INPUT:   A clause dictionary.
  EFFECT:  Deconstructs the clause dictionary.
  CAUTION: For every clause the evector of dependent positions
           is deconstructed. In function <...insertClause>
           no deep copy of the evector is done however!
***************************************************************/
{
  /* disallocate member structures */
  evector_destroyEvectorInt(dictionary->clause_numbers);
  evector_deconstructEvector(dictionary->dependent_positions,
			     &clause_dictionary_evector_dummy_destroy_position_list);
  evector_deconstructEvector(dictionary->dependent_positions_accumulated,
			     &position_evector_destroyPosition);

  /* disallocate structure itself */
  memory_Free(dictionary, sizeof(CLAUSE_DICTIONARY));
}


/**************************************************************/
/* Accessability                                              */
/**************************************************************/

void clause_dictionary_insertClause(CLAUSE_DICTIONARY* dictionary, int number, EVECTOR* dep_poss)
/**************************************************************
  INPUT:   A dictionary, a SPASS clause number, and a
           sorted evector of positions the clause depends on.
  EFFECT:  Inserts the pair (<number>, <dep_poss>) into the
           dictionary <dictionary>. Adapted the list of
           accumulated positions.
  CAUTION: No deep copy of dep_poss is done! This would cause
           an unneeded overhead.
***************************************************************/
{
  int i;
  POSITION *temp;

  evector_insertAppendedInt(dictionary->clause_numbers, number);
  evector_insertAppended(dictionary->dependent_positions, dep_poss);

  for(i=0; i<evector_getNumberOfElements(dep_poss); ++i) {
    temp = position_copyPosition((POSITION*) evector_getNth(dep_poss, i));
    if (evector_insertSortedNonDuplicate(dictionary->dependent_positions_accumulated,
					 temp) == 0)
      position_destroyPosition(temp);
  }
}

void clause_dictionary_insertClauseNonFiltering(CLAUSE_DICTIONARY* dictionary, int number)
/**************************************************************
  INPUT:   A dictionary, a SPASS clause number.
  EFFECT:  Inserts the pair (<number>, <dep_poss>) into the
           dictionary <dictionary>, <dep_poss> is allocated
           and only contains the empty position.
  CAUTION: Uses the function <clause_dictionary_insertClause>.
***************************************************************/
{
  EVECTOR * dep_poss;

  dep_poss =
    evector_createFromString("e\n", ",\n",
			     &position_evector_createFromString,
			     &position_evector_destroyPosition,
			     &position_evector_equal,
			     &position_evector_tree_order);

  clause_dictionary_insertClause(dictionary, number, dep_poss);

  /* empty position already inserted? */
  if (evector_getNumberOfElements(dictionary->dependent_positions_accumulated) == 0)
    evector_insertSortedNonDuplicate(dictionary->dependent_positions_accumulated,
				     position_createRootPosition());
}

int clause_dictionary_getNumberOfElements(CLAUSE_DICTIONARY* dictionary)
/**************************************************************
  INPUT:   A clause dictionary.
  RETURNS: The number of clauses stored in the dictionary.
  NOTE:    By the invariant of the datastructure we can either
           return <length(dictionary->clause_numbers)> or
           <length(dictionary->dependent_positions)>.
***************************************************************/
{
#ifdef CHECK
  if (evector_getNumberOfElements(dictionary->clause_numbers) !=
      evector_getNumberOfElements(dictionary->dependent_positions)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In clause_dictionary_getNumberOfElements:");
    misc_ErrorReport(" Inconsistent datastructure.");
    misc_FinishErrorReport();
  }
#endif

  return(evector_getNumberOfElements(dictionary->clause_numbers));
}

int clause_dictionary_getNthClauseNo(CLAUSE_DICTIONARY *dictionary, int n)
/**************************************************************
  INPUT:   A clause dictionary and an index.
  RETURNS: The SPASS clause number of the <n>-th stored
           clause.
  NOTE:    Usally the <n>-th stored clause from the dictionary
           is the <n-1>-th assumed clause from the SMT
           procedure.
***************************************************************/
{
  return(evector_getNthInt(dictionary->clause_numbers, n));
}

EVECTOR* clause_dictionary_getNthDependentPositions(CLAUSE_DICTIONARY *dictionary, int n)
/**************************************************************
  INPUT:   A clause dictionary and an index.
  RETURNS: The evector of positions which the <n>-th stored
           clause depends on.
  NOTE:    Usally the <n>-th stored clause from the dictionary
           is the <n-1>-th assumed clause from the SMT
           procedure. Thus the positions the <n-1>-th by the SMT
           procedure assumed clause is returned.
***************************************************************/
{
  return(evector_getNth(dictionary->dependent_positions, n));
}

EVECTOR* clause_dictionary_getAllDependentPositionsCopy(CLAUSE_DICTIONARY *dictionary)
/**************************************************************
  INPUT:   A clause dictionary.
  RETURNS: A copy of the accumulated list of all those
           positions at least one of the clauses in the
           dictionary depends on.
***************************************************************/
{
  EVECTOR *local;

  local = evector_copyEvector(dictionary->dependent_positions_accumulated,
			      &position_evector_copyLiterally);

  return local;
}

POSITION* clause_dictionary_getWorkingPosition(CLAUSE_DICTIONARY *dictionary)
/**************************************************************
  INPUT:   A CLAUSE_DICTIONARY object.
  RETURNS: The highest position a clause in <dictionary>
           depends on.
  CAUTION: This position can be understood as the current
           working position of the SMT procedure.
***************************************************************/
{
  return((POSITION*)
	 evector_getMaxElement(dictionary->dependent_positions_accumulated));
}
