/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 DICTIONARY OF CLAUSES                  * */
/* *                                                        * */
/* *  $Module: clause_dictionary                            * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _CLAUSEDICTIONARY_
#define _CLAUSEDICTIONARY_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "evector.h"
#include "position.h"


/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

/*
 * Explanation:
 *
 *    The SPASS clause number of the n-th exported clause (ie.
 *    the clause the SMT procedure has seen as n-th clause) is
 *    <dictionary->clause_numbers[n]>; it depends exactly on the
 *    branches in <dictionary->dependent_positions[n]>.
 *
 *    In particular <length(dictionary->clause_numbers)> ==
 *                  <length(dictionary->dependent_positions)>.
 *
 *    The list of _all_ dependent positions is given by
 *        dictionary->dependent_positions_accumulated.
 *    This list is sorted, i.e. the maximum branch the SMT
 *    procedure is currently working on, is given by the
 *    last entry.
 *
 *
 */
typedef struct
{
  EVECTOR *clause_numbers;
  EVECTOR *dependent_positions;
  EVECTOR *dependent_positions_accumulated;
} CLAUSE_DICTIONARY;


/**************************************************************/
/* Allocation and disallocation                               */
/**************************************************************/

CLAUSE_DICTIONARY* clause_dictionary_create(int, int);
void clause_dictionary_destroy(CLAUSE_DICTIONARY*);


/**************************************************************/
/* Accessability                                              */
/**************************************************************/

void clause_dictionary_insertClause(CLAUSE_DICTIONARY*, int, EVECTOR*);
void clause_dictionary_insertClauseNonFiltering(CLAUSE_DICTIONARY*, int);
int clause_dictionary_getNumberOfElements(CLAUSE_DICTIONARY*);
int clause_dictionary_getNthClauseNo(CLAUSE_DICTIONARY*, int);
EVECTOR* clause_dictionary_getNthDependentPositions(CLAUSE_DICTIONARY*, int);
EVECTOR* clause_dictionary_getAllDependentPositionsCopy(CLAUSE_DICTIONARY*);
POSITION* clause_dictionary_getWorkingPosition(CLAUSE_DICTIONARY*);

static __inline__
EVECTOR* clause_dictionary_getVectorOfClauseNumbers(CLAUSE_DICTIONARY *dictionary)
/*********************************************************
  INPUT:   A CLAUSE_DICTIONARY object.
  EFFECT:  The vector of clause numbers present in the
           dictionary.
**********************************************************/
{
  return dictionary->clause_numbers;
}

static __inline__
EVECTOR* clause_dictionary_getVectorOfDependentPositions(CLAUSE_DICTIONARY *dictionary)
/*********************************************************
  INPUT:   A CLAUSE_DICTIONARY object.
  EFFECT:  The vector of dependent positions present in the
           dictionary.
**********************************************************/
{
  return dictionary->dependent_positions;
}

static __inline__
EVECTOR* clause_dictionary_getAllDependentPositions(CLAUSE_DICTIONARY *dictionary)
/**************************************************************
  INPUT:   A CLAUSE_DICTIONARY object.
  RETURNS: The member of <dictionary> that refers to the list
           of all accumulated branches that at least one of the
           clauses depends on.
***************************************************************/
{
  return(dictionary->dependent_positions_accumulated);
}

#endif
