/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              SCOPING OF CONDENSATIONS                  * */
/* *                                                        * */
/* *  $Module: condensationscope                            * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "condensationscope.h"
#include "misc.h"
#include "memory.h"


/**************************************************************/
/* Creation and deletion                                      */
/**************************************************************/

CONDENSATION_SCOPE* condensationscope_createCONDENSATIONSCOPE(int base_size)
/**************************************************************
  INPUT:   An integer.
  RETURNS: Creates a condensation scope for initially
           <base_size> elements.
  CAUTION: Memory is allocated.
***************************************************************/
{
  CONDENSATION_SCOPE *local;

  local = evector_createEvector(base_size, 2,
				position_evector_equal,
				position_evector_tree_order);

  return(local);
}

void condensationscope_destroyCONDENSATIONSCOPE(CONDENSATION_SCOPE *scope)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object.
  CAUTION: Disallocates memory for <scope>.
***************************************************************/
{
  evector_deconstructEvector(scope, &position_evector_destroyPosition);
}


/**************************************************************/
/* Adding and retrieval                                       */
/**************************************************************/

void condensationscope_addCondensation(CONDENSATION_SCOPE *scope, POSITION* split)
{
  int result;
  POSITION* copy;

  copy = position_copyPosition(split);
  result = evector_insertSortedNonDuplicate(scope, (void*) copy);

#ifdef CHECK
  if (result == 0) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nIn condensationscope_addCondensation:");
    misc_UserErrorReport("\nCondensation of split added twice.\n");
    misc_FinishUserErrorReport();
  }
#endif
}

int condensationscope_getNumberOfCondensationsUntil(CONDENSATION_SCOPE *scope, POSITION* split)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object, a positions of a
           split.
  RETURNS: The number of branches above <p> (wrt. <=_p) that
           have been condensed.
***************************************************************/
{
  int result;
  int i, Index;

  result = 0;
  if (scope->number_of_elements > 0) {
    Index = evector_getIndexToElement(scope, split);

    for(i=0; i<=Index; ++i)
      if (position_PrefixOrderEqual((POSITION*) (scope->data[i]), split))
	++result;
  }

  return result;
}

int condensationscope_dependsOnCondensedBranch(CONDENSATION_SCOPE *scope, EVECTOR* positions)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object, a list of positions of
           branches..
  RETURNS: 1 if at least one of the _branches_ in <positions>
           was condensed (i.e. the corresponding split was
           condensed).
***************************************************************/
{
  int i;

  for(i=0; i<evector_getNumberOfElements(positions); ++i)
    if (condensationscope_wasCondensedBranch(scope, (POSITION*) evector_getNth(positions, i)))
      return 1;

  return 0;
}

int condensationscope_wasCondensedBranch(CONDENSATION_SCOPE *scope, POSITION *p)
/**************************************************************
  INPUT:   A CONDENSATION_SCOPE object, a position.
  RETURNS: 1 if  the branch with position <p> was condensed
           already, 0 otherwise.
***************************************************************/
{
  POSITION *tmp;
  int result;

  if (position_isRootPosition(p))
    /* empty branch cannot be condensed */
    result = 0;
  else {
    /* get the corresponding split to branch <p> */
    tmp = position_getAbovePosition(p);

    /* if the split is condensed then also the corresponding branches */
    result = evector_isMember(scope, (void*) tmp);

    position_destroyPosition(tmp);
  }

  return result;
}
