/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          MANAGING THE EXTERNAL SMT PROCEDURE           * */
/* *                                                        * */
/* *  $Module:   dpconnect                                  * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dpconnect.h"
#include "search.h"
#include "clock.h"
#include <time.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>


/**************************************************************/
/* Global variables                                           */
/**************************************************************/

#define _DPCONNECT_BIGSTRING_SIZE 8192
char dpconnect_big_string[_DPCONNECT_BIGSTRING_SIZE];
int dpconnect_term_signal = SIGTERM;

/* new timeout if SMT procedure terminated due to timeout
   <runsuccessfull> is the running time of an execution that terminated regularly */
#define new_timeout(old, runsuccessfull) ((runsuccessfull != -1) ? runsuccessfull + 2 : 2*old)


/**************************************************************/
/* Internal Functions                                         */
/**************************************************************/

int dpconnect_wasAbortedDueToLimits(int signr);
int checkPositionList(EVECTOR *positions);
int dpconnect_checkPositionList(EVECTOR *positions);

int dpconnect_wasAbortedDueToLimits(int signr)
{
  return ((signr == SIGSEGV) /* maximum size of the process' virtual memory */
	  || (signr == SIGXCPU) /* CPU  time  limit  in seconds */
	  || (signr == SIGXFSZ) /* maximum size of files */
	  || (signr == SIGALRM) /* stop watch, either intern or extern from calling process */
	  );
}

int dpconnect_checkPositionList(EVECTOR *positions)
/*********************************************************
  INPUT:   An expandable vector of positions.
  RETURNS: Whether the list of positions is consistent
           for a formula scope (0 if not, 1 if): let <mp>
           be the maximum position in <positions> wrt.
           <=_t; we check whether for all <p> in
           <positions> <p> <=_p <mp>; if not, then for
           at least one <p'> in <positions> <p'> >_b <mp>
           or <p'> and <mp> are not comparable
**********************************************************/
{
  POSITION *p_max;
  int i;

  p_max = (POSITION*) evector_getMaxElement(positions);

  for(i=0; i<evector_getNumberOfElements(positions); ++i)
    if (! position_PrefixOrderEqual( ((POSITION*) evector_getNth(positions, i)),
				     p_max))
      return 0;

  return 1;
}


/**************************************************************/
/* Access Functions                                           */
/**************************************************************/

POSITION* dpconnect_getCurrentSMTPosition(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  RETURNS: The current working position of the SMT
           procedure.
  EFFECT:  Delegates the work to the clause dictionary
           and uses function
           <clause_dictionary_getWorkingPosition>.
**********************************************************/
{
#ifdef CHECK
  if (dpconnect_getClauseDictionary(Search) == NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\nIn dpconnect_getCurrentSMTPosition:");
    misc_ErrorReport("\nexternal SMT procedure currently not running.\n");
    misc_FinishErrorReport();
  }
#endif

  return(clause_dictionary_getWorkingPosition(dpconnect_getClauseDictionary(Search)));
}

SMT_RESULT* dpconnect_getCurrentSMTResult(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  RETURNS: The most recent SMT result.
  CAUTION: The most recent SMT result is a NULL pointer,
           if the SMT procedure was not invoked so far or
           is still running on the first invocation.
**********************************************************/
{
  return prfs_DpConnect(Search)->current_smt_result;
}

int dpconnect_isChecking(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  RETURNS: Whether the SMT procedure is still running
           (or the subprocess is waiting to be resolved),
           1, or 0 otherwise (no process running at all).
**********************************************************/
{
  return (prfs_DpConnect(Search)->is_checking > 0);
}

CLAUSE_DICTIONARY* dpconnect_getClauseDictionary(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  RETURNS: The currently allocated clause dictionary.
  CAUTION: The dictionary can be NULL if the SMT
           procedure is currently not running.
**********************************************************/
{
  return (prfs_DpConnect(Search)->clause_dictionary);
}

void dpconnect_setClauseDictionary(struct PROOFSEARCH_HELP* Search, CLAUSE_DICTIONARY* new )
/**************************************************************
  INPUT:   A proofsearch object, a clause dictionary.
  EFFECT:  Sets the clause dictionary attribute of
           <Search->dpconnect> to <new>.
***************************************************************/
{
  prfs_DpConnect(Search)->clause_dictionary = new;
}

/**************************************************************/
/* Allocation and deconstruction                              */
/**************************************************************/

void dpconnect_Init(struct PROOFSEARCH_HELP* ps) {
/**************************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  Initialize the "dpconnect" member of <ps> by
           allocating memory and doing all initializations. It
           also checks the environment variables.
***************************************************************/
  DPCONNECT dpc;
  int i;
  char env_name[50];

  /* check whether a suitable theory was specified */
  /*
  if (flag_GetFlagValue(prfs_Store(ps), flag_DPCTHEORY) == flag_OFF) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nPlease set the flag \"flag_DpcTheory\" to a value unequal 0.\n");
    misc_FinishUserErrorReport();
  }
  */

  /* allocate memory for structure */
  dpc = (DPCONNECT_NODE*) memory_Malloc(sizeof(DPCONNECT_NODE));

  /* which SMT procedure are we using? -> set up name for environment variable
     containing the full path to program appropriately */
  /*
  sprintf(env_name, "TSPASS_DP%d",
	  flag_GetFlagValue(prfs_Store(ps), flag_DPCPROVER));
  */

  /* try to read contents of environment variable indicating full name of SMT procedure
  if (getenv(env_name) == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Environment variable 'TSPASS_DP<PROVER>' is not set. \n");
    misc_FinishUserErrorReport();
  }
  */
  if (flag_GetFlagValue(prfs_Store(ps), flag_DPCPROVER) == 1) {
    dpc->full_name_of_smt_procedure = string_StringCopy(TSPASS_DP1);
  } else if (flag_GetFlagValue(prfs_Store(ps), flag_DPCPROVER) == 2) {
    dpc->full_name_of_smt_procedure = string_StringCopy(TSPASS_DP2);
  } else {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Illegal value for flag_DPCPROVER. \n");
    misc_FinishUserErrorReport();
  }

  /* dpc->full_name_of_smt_procedure = string_StringCopy(getenv(env_name)); */

  for(i=strlen(dpc->full_name_of_smt_procedure)-1; i>=0; --i)
    if ((dpc->full_name_of_smt_procedure)[i] == '/')
      break;
  dpc->base_name_of_smt_procedure = string_Suffix(dpc->full_name_of_smt_procedure, i+1);

  /* create a temporary directory */
  if (getenv("TSPASS_TMPDIR") != NULL) {
    dpc->dpdir = string_StringCopy(getenv("TSPASS_TMPDIR"));
  } else
    dpc->dpdir = string_StringCopy("/tmp");
  dpc->dpdir = string_Nconc1(dpc->dpdir, "/spdpcXXXXXX");


  if (mkdtemp(dpc->dpdir) == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_Init:");
    misc_UserErrorReport("\n Cannot create temporary directory.\n");
    misc_FinishUserErrorReport();
  }

  /* setup the filenames for export/smt input/smt output */
  dpc->smt_export_filtering_filename = string_Conc(dpc->dpdir, "/export_filtering");
  dpc->smt_export_non_filtering_filename = string_Conc(dpc->dpdir, "/export_non_filtering");
  dpc->smt_input_filename = string_Conc(dpc->dpdir, "/smt_input");
  dpc->smt_output_filename = string_Conc(dpc->dpdir, "/smt_output");

#ifdef SMT_LOG
  dpc->smt_log_filename = string_Conc(dpc->dpdir, "/TSPASS.log");
#endif

  /* set up the positions */
  dpc->current_spass_position = position_createRootPosition();

  /* so far, no result */
  dpc->current_smt_result = (SMT_RESULT*) NULL;

  /* bookkeeping for checking */
  dpc->asserted_since_last_check = 0;
  dpc->asserted_non_filtered_clauses = 0;

  /* flags about state of SMT procedure */
  dpc->is_checking = 0;
  dpc->smt_pid = -1;
  dpc->got_all_problem_input_clauses = 0;

  /* attach to SMT_INTERFACE, init SMT module */
  dpc->smt_interface = smt_interface_Init(ps);

  /* set up signature string */
  switch(flag_GetFlagValue(prfs_Store(ps), flag_DPCTHEORY)) {
  case flag_DPCTINT:
    dpc->signature_string = smt_interface_SigToString(dpc->smt_interface, "int");
    break;
  case flag_DPCTRAT:
    dpc->signature_string = smt_interface_SigToString(dpc->smt_interface, "real");
    break;
  case flag_DPCTREAL:
    dpc->signature_string = smt_interface_SigToString(dpc->smt_interface, "real");
    break;
  default:
    dpc->signature_string = smt_interface_SigToString(dpc->smt_interface, "data");
    /* misc_StartUserErrorReport(); */
    /* misc_UserErrorReport("\n In dpconnect_Init:"); */
    /* misc_UserErrorReport("\n Illegal theory for DPCONNECT.\n"); */
    /* misc_FinishUserErrorReport(); */
    break;
  } /* switch theory */

  /* mark clause dictionary as unused */
  dpc->clause_dictionary = (CLAUSE_DICTIONARY*) NULL;

  /* create dictionary of clauses that are not subject to filtering */
  dpc->non_filtered_clauses = evector_createEvectorInt(1000, 2);

  /* prepare export files */
  dpc->smt_export_filtering_file = fopen(dpc->smt_export_filtering_filename, "w");
  if (dpc->smt_export_filtering_file == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nUnable to open file for exporting clauses (filtering)!");
    misc_FinishUserErrorReport();
  }
  dpc->smt_export_non_filtering_file = fopen(dpc->smt_export_non_filtering_filename, "w");
  if (dpc->smt_export_non_filtering_file == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nUnable to open file for exporting clauses (non-filtering)!");
    misc_FinishUserErrorReport();
  }

  /* some standard preparations for the clauses not subject to filtering
     needed for the smt procedure */
  smt_interface_prepareNonFilteredClauseInput(dpc->smt_interface,
					      ps, dpc->smt_export_non_filtering_file);

  /* prepare statistics */
  dpc->smt_invocations_total = 0;
  dpc->smt_invocations_got_a_result = 0;
  dpc->smt_invocations_killed = 0;
  dpc->smt_invocations_killed_timeout = 0;

  /* check whether SMT procedure supports unsatisfiable cores */
  if (flag_GetFlagValue(prfs_Store(ps), flag_DPCEVIDENCE) == flag_ON)
    if (smt_interface_supportsUnsatisfiableCores(dpc->smt_interface) != 1) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\nEvidence is not supported by the SMT procedure!\n");
      misc_UserErrorReport("Please disable the flag \"DpcEvidence\"!\n");
      misc_FinishUserErrorReport();
    }

  /* check whether SMT procedure supports export of non-ground formulae */
  if (flag_GetFlagValue(prfs_Store(ps), flag_DPCNONGROUNDEXPORT) == flag_ON)
    if (smt_interface_supportsNonGroundExport(dpc->smt_interface) != 1) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\nExport of non-ground formulae is not supported by the SMT procedure!\n");
      misc_UserErrorReport("Please disable the flag \"DpcNonGroundExport\" and set the switch \"IThI\" to a value >= 1!\n");
      misc_FinishUserErrorReport();
    }

  /* assign member of ProofSearch object */
  ps->dpconnect = dpc;

  /* initially, no swapping necessary */
  dpc->swap = FALSE;

#ifdef SMT_LOG
  /* prepare statistical log */
  dpc->smt_log_file = fopen(dpc->smt_log_filename, "w");
  dpconnect_startLogging(ps);
#endif

  /* set threshold for invoking a check */
  dpconnect_setNewThreshold(ps);

  /* set initial timeout */
  dpc->current_smt_timeout = 0.0;
  dpconnect_setNewSMTTimelimit(ps);

  return;
}

void dpconnect_Free(struct PROOFSEARCH_HELP* Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  Deconstructs all members of the dpconnect
           structure belonging to <Search> and finally frees
           memory of the dpconnection structure itself.
***************************************************************/
{
  DPCONNECT dpc;

  /* abbreviation */
  dpc = prfs_DpConnect(Search);

  /* check whether we are still checking */
  if (dpc->is_checking > 0)
    dpconnect_interruptExternalSMT(Search, NULL, "dpconnect_Free", TRUE);

  /* close export files */
  if (dpc->smt_export_filtering_file != NULL) {
    fclose(dpc->smt_export_filtering_file);
    /* possibly ignore any error */
    errno = 0;
  }
  if (dpc->smt_export_non_filtering_file != NULL) {
    fclose(dpc->smt_export_non_filtering_file);
    /* possibly ignore any error */
    errno = 0;
  }

#ifdef SMT_LOG
  /* close statistical log file */
  dpconnect_finishLogging(Search);
  fclose(dpc->smt_log_file);
#endif

  /* possibly delete the temporary files */
  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCDEBUG) != flag_DPCDEBUGON) {

    remove(dpc->smt_input_filename);
    remove(dpc->smt_output_filename);
    remove(dpc->smt_export_filtering_filename);
    remove(dpc->smt_export_non_filtering_filename);
#ifdef SMT_LOG
    remove(dpc->smt_log_filename);
#endif
    remove(dpc->dpdir);

    /* don't care if something went wrong */
    errno = 0;
  }

  /* free filenames */
  string_StringFree(dpc->dpdir);
  string_StringFree(dpc->full_name_of_smt_procedure);
  string_StringFree(dpc->base_name_of_smt_procedure);

  string_StringFree(dpc->smt_input_filename);
  string_StringFree(dpc->smt_output_filename);
  string_StringFree(dpc->smt_export_filtering_filename);
  string_StringFree(dpc->smt_export_non_filtering_filename);
#ifdef SMT_LOG
  string_StringFree(dpc->smt_log_filename);
#endif

  /* free positions */
  if (dpc->current_spass_position != NULL)
    position_destroyPosition(dpc->current_spass_position);

  /* free signature string */
  estring_StringFree(dpc->signature_string);

  /* possibly free existing clause dictionry, if not already done */
  if (dpc->clause_dictionary != NULL)
    clause_dictionary_destroy(dpc->clause_dictionary);

  /* free dictionary for clauses not subject to filtering */
  evector_destroyEvectorInt(dpc->non_filtered_clauses);

  /* possibly free memory for current SMT result */
  if (dpc->current_smt_result != NULL)
    smt_result_Free(dpc->current_smt_result);

  /* shutdown the SMT connection module */
  smt_interface_FreeProver(dpc->smt_interface);

  /* free smt interface structure */
  smt_interface_Disalloc(dpc->smt_interface);

  /* free structure itself */
  memory_Free(dpc, sizeof(DPCONNECT_NODE));
}



/**************************************************************/
/* Proof Control                                              */
/*    Note: These functions *should* be used from outside.    */
/**************************************************************/

BOOL dpconnect_CanContributeMostRecentResult(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  RETURNS: Whether the most recent result unsat or sat and
           is relevant wrt. to the current SPASS configuration,
           1. Otherwise 0. In particular, if there is no
           recent result also 0.
  EFFECT:  Calls <smt_result_isRelevant>.
***************************************************************/
{
  BOOL result;
  DPCONNECT dpc;
  POSITION *p_spass;

  /* abbreviation */
  dpc = prfs_DpConnect(Search);

  result = FALSE;

  if (dpc->current_smt_result == NULL)
    result = FALSE;
  else if (!smt_result_isUnsat(dpc->current_smt_result) &&
	   !smt_result_isSat(dpc->current_smt_result))
    result = FALSE;
  else {
    /* get the current SPASS working position */
    p_spass = prfs_getPositionOfCurrentLevelInCurrentConfiguration(Search);

    result = smt_result_isRelevant(dpc->current_smt_result, Search,
				   p_spass);

    position_destroyPosition(p_spass);
  }

  return result;
}

int dpconnect_updateState(struct PROOFSEARCH_HELP* Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  RETURNS: Whether the state changed (1) or nothing happened
           since the function was called last time (0). A
           state is said to have changed if the SMT procedure
           was working but delivered a result which was not
           not grapped so far.
  EFFECT:  It looks for the SMT procedure: is it still
           working? If so, do we care about a timeout? If so,
           then we check whether the timeout already expired
           and possibly kill the decision procedure
           subprocess, set the last result to "unknown" and
           return 1 (state changed). Otherwise, either if we do
           not care about a timeout or the SMT procedure is
           still running but so far the timeout did not expire,
           we just return 0 (indicating that nothing changed).
           If the SMT procedure already returned (either due
           to normal behaviour or because of resource limits)
           we correspondingly create the result of its
           invocation, i.e. if it returned normally, we
           delegate the parsing of the output to the
           specialized function of the corresponding
           connection module. Moreover, if it was interrupted
           by a signal or we stopped it because of the expired
           timelimit, we create an "unknown" result.
  CAUTION: The function does not initiate another check, even
           if the conditions for a new check might be
           fulfilled.
***************************************************************/
{
  DPCONNECT dpc;
  int result, smt_status, status, smt_return_status;
  EVECTOR *explanation, *dependent_positions;
  struct rusage usage;
  float started_before;

#ifdef SMT_DEBUG
  clock_StartCounter(clock_SMTGETRESULT);
#endif

  /* abbreviation */
  dpc = prfs_DpConnect(Search);

  /* amount of seconds since the process was started */
  started_before = clock_GetSeconds(clock_SMTSUBPROCESS);

  /* is SMT currently working at all? */
  if (dpc->is_checking == 0)
    result = 0; /* smt procedure not running at all */
  else if ((smt_return_status=
	    dpconnect_hasFinishedExternalSMT(Search, &smt_status, &usage)) == 0) {
    /* smt procedure is still running */
    if (dpc->use_timelimit) {
      /* check whether our timeout expired */
      if (started_before > dpc->current_smt_timeout) {

#ifdef SMT_DEBUG
	smt_debug_start();
	printf("The given timeout (%f seconds) expired (%f), subprocess will be killed!\n",
	       dpc->current_smt_timeout, started_before);
	smt_debug_end();
#endif
#ifdef SMT_LOG
	dpconnect_logMessage(Search, 1, "The given timeout (%f seconds) expired (%f), subprocess will be killed!\n", dpc->current_smt_timeout, started_before);
#endif

	/* smt procedure is still running, the given timelimit expired */
	dpconnect_interruptExternalSMT(Search, NULL, "timelimit expired", FALSE);

	++(prfs_DpConnect(Search)->smt_invocations_killed_timeout);

	/* adjust timeout for next invocation: if SMT procedure was interrupted
	   due to timeout, we increase the current timeout */
	dpconnect_setNewSMTTimelimit(Search);

	/* disallocate previous explanation */
	if (dpconnect_getCurrentSMTResult(Search) != NULL)
	  smt_result_Free(dpconnect_getCurrentSMTResult(Search));

	/* generate result "unknown" */
	dpc->current_smt_result =
	  smt_result_CreateUnknown(clause_dictionary_getAllDependentPositionsCopy(dpconnect_getClauseDictionary(Search)));

	/* disallocate clause dictionary */
	if (dpconnect_getClauseDictionary(Search) != NULL) {
	  clause_dictionary_destroy(dpconnect_getClauseDictionary(Search));
	  dpconnect_setClauseDictionary(Search, (CLAUSE_DICTIONARY*) NULL);
	}

	/* now, not checking any more */
	dpc->is_checking = 0;
	dpc->smt_pid = -1;

	/* here: start new checking? if, then only if the caller is not
	   function "exportClause" because there possibly a new start will be
	   invoked anyway
	*/

	result = 1; /* something changed */
      } else {
	/* smt procedure is still running, it still has time left */
	result = 0; /* nothing changed */
      }
    } else {
      /* smt procedure is still running, we do not care about any timeout */
      result = 0; /* nothing changed */
    }
  } else {
    /* SMT procedure either terminated or was interrupted due to resource limits */

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("SMT procedure terminated.\n");
    printf("Spent user time: %ld.%.6ld\n", usage.ru_utime.tv_sec, usage.ru_utime.tv_usec);
    printf("Spent system time: %ld.%.6ld\n", usage.ru_stime.tv_sec, usage.ru_stime.tv_usec);
    printf("In total: %ld.%ld\n",
	   ((usage.ru_utime.tv_usec + usage.ru_stime.tv_usec)/1000000)+usage.ru_utime.tv_sec+usage.ru_stime.tv_sec,
	   (usage.ru_utime.tv_usec+usage.ru_stime.tv_usec)% 1000000);
    printf("The process was started %f seconds ago by DPCONNECT.\n", started_before);
    smt_debug_end();
#endif

#ifdef SMT_LOG
    dpconnect_logMessage(Search, 1, "SMT procedure terminated.\n");
    dpconnect_logMessage(Search, 0, "\tThis was the %d-th invocation.\n",
			 (prfs_DpConnect(Search))->smt_invocations_total);
    dpconnect_logMessage(Search, 0, "\tSpent user time: %ld.%.6ld\n", usage.ru_utime.tv_sec, usage.ru_utime.tv_usec);
    dpconnect_logMessage(Search, 0, "\tSpent system time: %ld.%.6ld\n", usage.ru_stime.tv_sec, usage.ru_stime.tv_usec);
    dpconnect_logMessage(Search, 0, "\tIn total: %ld.%.6ld\n",
	   ((usage.ru_utime.tv_usec + usage.ru_stime.tv_usec)/1000000)+usage.ru_utime.tv_sec+usage.ru_stime.tv_sec,
	   (usage.ru_utime.tv_usec+usage.ru_stime.tv_usec)% 1000000);
    dpconnect_logMessage(Search, 0, "\tThe process was started %f seconds ago by DPCONNECT.\n", started_before);
#endif

    if (smt_return_status == 2) {
      /* finished due to signal; since we do not give timeouts on the command line,
         but rely on our timing control of the subprocess the executions must have
         been interrupted due to resource limits or other strange errors */

      /* adapt timeout */

      /* disallocate previous result */
      if (dpc->current_smt_result != NULL)
	smt_result_Free(dpconnect_getCurrentSMTResult(Search));

      /* generate result "unknown" */
      dpc->current_smt_result =
	smt_result_CreateUnknown(clause_dictionary_getAllDependentPositionsCopy(dpconnect_getClauseDictionary(Search)));
#ifdef SMT_LOG
      dpconnect_logMessage(Search, 0, "\tTermination due to signal %d.\n",
			   smt_status);
#endif
    } else {
      /* finished with exit value */

      /* note that <explanation> and <dependent_positions> are allocated in
	 <smt_parseSMTOutput> but disallocated in <dpconnect_freeSMTResult> */
      explanation = (EVECTOR*) NULL;
      dependent_positions = (EVECTOR*) NULL;
      status = smt_interface_parseSMTOutput((prfs_DpConnect(Search))->smt_interface,
					    Search,
					    (prfs_DpConnect(Search))->smt_output_filename,
					    &explanation,
					    &dependent_positions,
					    smt_status);

      if (status == 4) {
	/* some problems within the <smt_parseSMTOutput> function */
	misc_StartErrorReport();
	misc_ErrorReport("\n In dpconnect_updateState:");
	misc_ErrorReport("\n SMT procedure did not deliver a recognizable output or reported an error.\n");
	misc_FinishErrorReport();
      }
      /* disallocate previous explanation */
      if (dpconnect_getCurrentSMTResult(Search) != NULL)
	smt_result_Free(dpconnect_getCurrentSMTResult(Search));

      /* store result for later use */
      if (status == 1) {
	if (explanation == NULL) /* no unsatisfiable cores,
				    and thus also <dependent_positions> == NULL */
	  dpc->current_smt_result =
	    smt_result_CreateUnsatWithoutEvidence(dpconnect_getClauseDictionary(Search));
	else /* unsatisf. cores */
	  dpc->current_smt_result =
	    smt_result_CreateUnsatWithEvidence(explanation, dependent_positions);
      } else if (status == 2) {
	dpc->current_smt_result =
	  smt_result_CreateSat(clause_dictionary_getAllDependentPositionsCopy(dpconnect_getClauseDictionary(Search)));
      } else /* (status == 3) */ {
	dpc->current_smt_result =
	  smt_result_CreateUnknown(clause_dictionary_getAllDependentPositionsCopy(dpconnect_getClauseDictionary(Search)));
      }

#ifdef SMT_LOG
      dpconnect_logMessage(Search, 0, "\tTermination with exit status %d.\n",
			   smt_status);
      dpconnect_logMessage(Search, 0, "\tResult sucessfully created.\n",
			   smt_status);
#endif

    }

    /* disallocate clause dictionary */
    if (dpconnect_getClauseDictionary(Search) != NULL) {
      clause_dictionary_destroy(dpconnect_getClauseDictionary(Search));
      dpconnect_setClauseDictionary(Search, (CLAUSE_DICTIONARY*) NULL);
    }

    /* now, not checking any more */
    dpc->is_checking = 0;
    dpc->smt_pid = -1;

    result = 1; /* something changed */
  }

#ifdef SMT_DEBUG
  clock_StopAddPassedTime(clock_SMTGETRESULT);

  if (result == 1) {
    smt_debug_start();
    printf("State change of SMT procedure subprocess\n");
    smt_debug_end();
  }
#endif
#ifdef SMT_LOG
  if (result == 1) {
    dpconnect_logMessage(Search, 1, "dpconnect_updateState will report state change of SMT procedure.\n");
  }
#endif

  return result;
}

void dpconnect_exportClause(struct PROOFSEARCH_HELP* Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A Proofsearch object, a clause.
  EFFECT:  It inserts the clause, as well as the positions
           on which it depends, into the list of exported
           clauses (until now this is a file appending in
           effect). Furthermore <update_state> is invoked and
           if the state changes (i.e. the SMT procedure was
           checking and returned) the number of exported
           clauses is compared to the treshhold and possibly
           another check is invoked (provided that certain
           other conditions are fulfilled).
           Note:
              (1) Standard axioms are not exported.
              (2) Clauses are exported at most once.
              (3) Depending on the flag "DpcNonGroundExport"
                  non-ground clauses are exported or not.
              (4) Clause from branch "e" are exported to the
                  export file containing clauses that are not
                  subject to filtering, all others to the file
                  containing clauses that are possibly
                  filtered out later.
  CAUTION: Besides urgent checks this is the only place where
           new invocations of the SMT procedure are triggered.
***************************************************************/
{
  ESTRING *translation;
  EVECTOR *dependent_positions;
  char *positions_string;
  char type[20];
  int tmp_cond;
  DPCONNECT dpc;

  /* abbreviation */
  dpc = prfs_DpConnect(Search);

  if (prfs_ValidLevel(Search) < clause_SplitLevel(Clause)) {
    return;
  }

  /* clause already exported ? */
  if (clause_ExportedToDpc(Clause))
    return;

  /* clause due to standard axiom? then do not export it.*/
  if (clause_IsTheoryStandardAxiom(Clause) == 1)
    return;
  /* TheoryExtensionAxioms are exported! */

  /* auxilliary clause coming from the theory instantiation rule */
  if (clause_IsThGroundClause(Clause))
    return;

  /* check whether we can export the clause: do less expensive checks first */
  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCNONGROUNDEXPORT) != flag_DPCNONGROUNDEXPORTFULL) {
    if (flag_GetFlagValue(prfs_Store(Search),
			  flag_DPCNONGROUNDEXPORT) == flag_DPCNONGROUNDEXPORTONLYINPUT) {
      if (! clause_IsFromInput(Clause))
	if (clause_ContainsVariables(Clause))
	  return;
    } else if (flag_GetFlagValue(prfs_Store(Search),
				 flag_DPCNONGROUNDEXPORT) == flag_OFF) {
      if (clause_ContainsVariables(Clause))
	return;
    }
  }

  /* what kind of type do we take for quantifier prefixes? */
  switch(flag_GetFlagValue(prfs_Store(Search), flag_DPCTHEORY)) {
  case flag_DPCTINT:
    strcpy(type, "int");
    break;
  case flag_DPCTRAT:
    strcpy(type, "real");
    break;
  case flag_DPCTREAL:
    strcpy(type, "real");
    break;
  default:
    strcpy(type, "data");
    /* misc_StartUserErrorReport(); */
    /* misc_UserErrorReport("\n In dpconnect_SendSignature:"); */
    /* misc_UserErrorReport("\n Illegal theory for CONTROL.\n"); */
    /* misc_FinishUserErrorReport(); */
    break;
  } /* switch */

  /* translate the clause into string representation */
  translation = smt_interface_clauseToString((prfs_DpConnect(Search))->smt_interface,
					     Clause, type);

  /* only send non-empty (in string representation!) clauses */
  if (!estring_EqualString(translation, "")) {

    /* distingish between clauses that are not subject to filtering (splitlevel 0 and
       depending only on branch e) and clauses that possibly are filtered out later */
    if (clause_SplitLevel(Clause) == 0) {

      /* print it in the format suitable for the used SMT procedure */
      smt_interface_addAssertion((prfs_DpConnect(Search))->smt_interface,
				   Search,
				   (prfs_DpConnect(Search))->smt_export_non_filtering_file,
				   estring_GetString(translation),
				   clause_Number(Clause),
				   (prfs_DpConnect(Search))->asserted_non_filtered_clauses+1);

      /* maintain dictionary for non-filtered clauses */
      evector_insertAppendedInt((prfs_DpConnect(Search))->non_filtered_clauses,
				clause_Number(Clause));

      /* maintain the number of non-filtered clauses; this is done no matter
         we use the dictionary or not because we're in any case interested in the total
         number of exported clauses but are not willing to count the lines with contents
	 of the file for the exported clauses with splitlevel zero */
      ++((prfs_DpConnect(Search))->asserted_non_filtered_clauses);

    } else {
      /* get dependent positions of clause */
      dependent_positions =
	prfs_getPositionsOfDependentLevelsOfClauseInCurrentConfigurationSorted(Search, Clause);

      /* convert positions to list of strings */
      positions_string =
	evector_convertToString(dependent_positions, &position_evector_convertToString, ",");

      fprintf((prfs_DpConnect(Search))->smt_export_filtering_file,
	      "%s\n%d\n%s\n", positions_string, clause_Number(Clause),
	      estring_GetString(translation));

      /* tidy up allocated memory */
      evector_deconstructEvector(dependent_positions, &position_evector_destroyPosition);
      memory_Free(positions_string, sizeof(char)*(1+strlen(positions_string)));
    }

    /* mark clause as "exported" */
    clause_SetExportedToDpc(Clause, TRUE);
    ++((prfs_DpConnect(Search))->asserted_since_last_check);

  } /* string translation of Clause != "" */

  /* tidy up string translation of clause */
  estring_StringFree(translation);

  /* we invoke a new check if the conjunctions of the following is fulfilled:
     (1) "threshold condition": asserted_since_last_check >= check_after
     (2) we already got all problem input clauses
     (3) the disjunction of the following:
           (a) SMT procedure currently not working and either no most recent result or
               most recent result is not relevant any more
           (b) SMT procedure currently working and an updateState gives a new state such
               that the result is not relevant any more (in particular if check was
               interrupted due to expired timeouts)
         Possibly one could also take "splittings_since_most_recent_check" into
         consideration instead of just considering the threshold condition */

  if ( (dpc->asserted_since_last_check >= dpc->check_after) &&
       (dpc->got_all_problem_input_clauses == 1) ) {
    tmp_cond = 0;
    if (dpc->is_checking == 0) {
      if (!dpconnect_CanContributeMostRecentResult(Search)) {
	tmp_cond = 1;
#ifdef SMT_LOG
	dpconnect_logMessage(Search, 1, "Condition for a check fulfilled: %d >= %d\n",
			     dpc->asserted_since_last_check,
			     dpc->check_after);
	dpconnect_logMessage(Search, 0, "\tSMT procedure was sleeping, either no previous result or not relevant any more\n");
#endif
      }
    } else {
      if (dpconnect_updateState(Search) == 1)
	if (!dpconnect_CanContributeMostRecentResult(Search)) {
	  tmp_cond = 1;
#ifdef SMT_LOG
	  dpconnect_logMessage(Search, 1, "Condition for a check fulfilled: %d >= %d\n",
			       dpc->asserted_since_last_check,
			       dpc->check_after);
	  dpconnect_logMessage(Search, 0, "\tSMT procedure terminated, but its result is not relevant [maybe due to timeout expiration].\n");
#endif
	}
    }

    if (tmp_cond == 1)
      dpconnect_performCheck(Search);
  }

  return;
}

void dpconnect_registerSplitting(struct PROOFSEARCH_HELP* Search, CLAUSE Clause)
/*********************************************************
  INPUT:   A Proofsearch object, a clause.
  EFFECT:  Registers a splitting, i.e. changes the
           variable indicating the current SPASS working
           position (<pos := pos.0>) by considering the
           split stack of <Search>.
           Also increments the variable
           <splittings_since_most_recent_check> by 1.
**********************************************************/
{
#ifdef SMT_LOG
  char *log_temp;
#endif

  /* get new position of SPASS, always a left branch */
  if ((prfs_DpConnect(Search))->current_spass_position != NULL)
    position_destroyPosition((prfs_DpConnect(Search))->current_spass_position);
  (prfs_DpConnect(Search))->current_spass_position =
    prfs_getPositionOfCurrentLevelInCurrentConfiguration(Search);

  /* bookeeping: how many splittings since last/current check */
  ++((prfs_DpConnect(Search))->splittings_since_most_recent_check);

#ifdef SMT_DEBUG
  smt_debug_start();
  printf("Splitting registered, new position is ");
  position_print((prfs_DpConnect(Search))->current_spass_position);
  printf(".\n");
  smt_debug_end();
#endif
#ifdef SMT_LOG
  log_temp = position_convertToString((prfs_DpConnect(Search))->current_spass_position);
  dpconnect_logMessage(Search, 1, "Splitting registered, new position is %s.\n",
		       log_temp);
  memory_Free(log_temp, (1+strlen(log_temp)*sizeof(char)));
  if ((prfs_DpConnect(Search))->is_checking > 0) {
    log_temp = position_convertToString(dpconnect_getCurrentSMTPosition(Search));
    dpconnect_logMessage(Search, 0, "\tCurrent position of SMT procedure is %s.\n",
			 log_temp);
    memory_Free(log_temp, (1+strlen(log_temp)*sizeof(char)));
  }
#endif

  return;
}

void dpconnect_registerBacktracking(struct PROOFSEARCH_HELP* Search, CLAUSE Clause)
/*********************************************************
  INPUT:   A Proofsearch object, a clause.
  EFFECT:  Registers a backtracking, i.e. changes the
           variable indicating the current SPASS working
           position appropriately by considering the
           split stack of <Search>.
           Possibly terminates the SMT procedure
           (if working on a position not relevant any
           more).
**********************************************************/
{
#ifdef SMT_LOG
  char *log_temp;
#endif
  /* additional check: did we go to the 0-th level? */
  if (prfs_ValidLevel(Search) == 0)
    return;

  /* get the current position of SPASS, always a right branch */
  if ((prfs_DpConnect(Search))->current_spass_position != NULL)
    position_destroyPosition((prfs_DpConnect(Search))->current_spass_position);
  (prfs_DpConnect(Search))->current_spass_position =
    prfs_getPositionOfCurrentLevelInCurrentConfiguration(Search);

  /* for next invocation swapping is necessary */
  prfs_DpConnect(Search)->swap = TRUE;

#ifdef SMT_DEBUG
  smt_debug_start();
  printf("Backtracking registered, new position is ");
  position_print((prfs_DpConnect(Search))->current_spass_position);
  printf(".\n");
  smt_debug_end();
#endif
#ifdef SMT_LOG
  log_temp = position_convertToString((prfs_DpConnect(Search))->current_spass_position);
  dpconnect_logMessage(Search, 1, "Backtracking registered, new position is %s.\n",
		       log_temp);
  memory_Free(log_temp, (1+strlen(log_temp)*sizeof(char)));

  if ((prfs_DpConnect(Search))->is_checking > 0) {
    log_temp = position_convertToString(dpconnect_getCurrentSMTPosition(Search));
    if (!position_PrefixOrderEqual(dpconnect_getCurrentSMTPosition(Search),
				    (prfs_DpConnect(Search))->current_spass_position))
      dpconnect_logMessage(Search, 0, "\tSMT procedure has to be terminated (is working in %s)\n",
			   log_temp);
    else
      dpconnect_logMessage(Search, 0, "\tSMT procedure has NOT to be terminated (is working in %s)\n",
			   log_temp);
    memory_Free(log_temp, (1+strlen(log_temp)*sizeof(char)));
  }
#endif

  /* possibly interrupt check of SMT procedure */
  if ((prfs_DpConnect(Search))->is_checking > 0)
    if (!position_PrefixOrderEqual(dpconnect_getCurrentSMTPosition(Search),
				   (prfs_DpConnect(Search))->current_spass_position))
      dpconnect_interruptExternalSMT(Search, NULL, "relevant backtracking", TRUE);

  /* Possibly: directly re-start SMT procedure again */

  return;
}

void dpconnect_registerCondensation(struct PROOFSEARCH_HELP* Search, POSITION *p)
/*********************************************************
  INPUT:   A Proofsearch object, a position.
  EFFECT:  Registers a condensation, i.e. updates the
           condensation database.
           Possibly terminates the SMT procedure
           (if working on a position under a condensed
           branch).
  CAUTION: <p> is the position of the _SPLIT_ that is
           condensed. If currently the left (right)
           branch of this split is entered, this will
           mean that all clauses depending on position
           <p.0> (<p.1>) are going to be removed.
           The branch that was condensed has to be added
           to the database of positions of condensed
           splits _BEFORE_ this function is called.
**********************************************************/
{
  BOOL interrupt;
#ifdef SMT_LOG
  char *log_temp;
#endif

  /* for next invocation swapping is necessary */
  prfs_DpConnect(Search)->swap = TRUE;

  /* do we have to interrupt the external decision procedure? */
  interrupt = FALSE;
  if ((prfs_DpConnect(Search))->is_checking > 0)
    if (condensationscope_dependsOnCondensedBranch(prfs_CondensationScope(Search), clause_dictionary_getAllDependentPositions(dpconnect_getClauseDictionary(Search))))
      interrupt = TRUE;

#ifdef SMT_DEBUG
  smt_debug_start();
  printf("Condensation registered: ");
  position_print(p);
  printf(".\n");
  smt_debug_end();
#endif
#ifdef SMT_LOG
  log_temp = position_convertToString(p);
  dpconnect_logMessage(Search, 1, "Condensation registered: %s.\n",
		       log_temp);
  string_StringFree(log_temp);
  if ((prfs_DpConnect(Search))->is_checking > 0) {
    log_temp = evector_convertToString(clause_dictionary_getAllDependentPositions(dpconnect_getClauseDictionary(Search)), &position_evector_convertToString, ",");
    if (interrupt)
      dpconnect_logMessage(Search, 0, "\tSMT procedure has to be terminated (is working with branches %s)\n",
			   log_temp);
    else
      dpconnect_logMessage(Search, 0, "\tSMT procedure has NOT to be terminated (is working with branches %s)\n",
			   log_temp);
    string_StringFree(log_temp);
  }
#endif

  if (interrupt)
    dpconnect_interruptExternalSMT(Search, NULL, "relevant condensation", TRUE);
  /* Possibly:  directly re-start SMT procedure again */

  return;
}

void dpconnect_gotAllProblemInputClauses(struct PROOFSEARCH_HELP  *Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  Remembers that all input clauses have been
           exported already. Invokes a check of the SMT
           procedure for the first time, if there has
           been exported at least one clause (not
           necessarily the case, even for non-empty input
           problems if no non-ground clauses are
           exported).
**********************************************************/
{
#ifdef SMT_DEBUG
  smt_debug_start();
  printf("All input clauses exported to be given to SMT procedure.\n");
  smt_debug_end();
#endif
#ifdef SMT_LOG
  dpconnect_logMessage(Search, 1,
		       "All input clauses exported to be given to SMT procedure.\n");
#endif

  (prfs_DpConnect(Search))->got_all_problem_input_clauses = 1;

  /* write a comment into the export file for clauses not subject to filtering that
     we got all input clauses */
  smt_interface_printComment((prfs_DpConnect(Search))->smt_interface,
			     Search,
			     (prfs_DpConnect(Search))->smt_export_non_filtering_file,
			     "Input clauses until *here*\n");
  if ((prfs_DpConnect(Search))->asserted_since_last_check > 0)
    /* this could be zero in case all input formulae are non-ground
       and we only export ground formulae */
    dpconnect_performCheck(Search);
}

void dpconnect_performCheck(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  First the state of the structure is updated. Then,
           if there is no SMT subprocess which is still
           running, the external SMT procedure is invoked.
  CAUTION: This function might also be called from outside.
***************************************************************/
{
  /* maybe SMT procedure was checking but delivered a result */
  dpconnect_updateState(Search);

  if ((prfs_DpConnect(Search))->is_checking == 0)
    dpconnect_invokeExternalSMT(Search, 1);
}

void dpconnect_performCheckUrgent(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  Any existing DP subprocess is interrupted.
           A new DP subprocess is started on all formulae.
  CAUTION: This function might also be called from outside, in
           particular if SPASS is not able to close a branch.
           Note that the state the datastructure is not
           updated because this might take time. Note that no
           time limit is assumed, since we have to wait for a
           result, even if it takes very long.
***************************************************************/
{
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Trying to interrupt external SMT...\n");
          smt_debug_end();
#endif
  dpconnect_interruptExternalSMT(Search, NULL, "urgent check", TRUE);
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Interrupted SMT, invoking SMT again...\n");
          smt_debug_end();
#endif
  dpconnect_invokeExternalSMT(Search, 0);
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Invoked SMT...\n");
          smt_debug_end();
#endif
}

int dpconnect_hasResult(struct PROOFSEARCH_HELP* Search)
/**************************************************************
  INPUT:   A Proofsearch object.
  RETURNS: 0 if until now we don't have any result from the
           SMT procedure and 1 otherwise.
  EFFECT:  First the state is updated, which might yield a
           fresh result. Then we check whether there is a
           result (and return 1) or not (and return 0). A
           result is waiting iff.
               <current_smt_result> != NULL.
  CAUTION: The purpose of this function is to be called from
           outside if one looks whether there is a result
           from the SMT procedure which might contribute to
           the proof. If the function returned one, then we
           grap the result using the function
               <dpconnect_getCurrentSMTResult>.
***************************************************************/
{

  dpconnect_updateState(Search);

  return (dpconnect_getCurrentSMTResult(Search) != NULL);
}


/**************************************************************/
/* Handling of external SMT procedure                         */
/*    Note: These functions shouldn't be called from outside. */
/**************************************************************/

void dpconnect_setNewThreshold(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A PROOFSEARCH object.
  EFFECT:  Calculates the new threshold for the next
           execution.
  CAUTION: Call after <smt_invocations_total> was set /
           increased newly.
***************************************************************/
{
  int a,b,c,d,e,n;
#ifdef SMT_LOG
  dpconnect_logMessage(Search, 1, "Calculating new threshold for %d-th execution.\n",
		       (prfs_DpConnect(Search)->smt_invocations_total)+1);
  dpconnect_logMessage(Search, 0, "\tOld threshold was: %d\n",
		       prfs_DpConnect(Search)->check_after);
#endif
  /* a*b^n + c*n*n + d*n + e */
  n = prfs_DpConnect(Search)->smt_invocations_total;
  a = flag_GetFlagValue(prfs_Store(Search), flag_DPCTHRESHOLDA);
  b = flag_GetFlagValue(prfs_Store(Search), flag_DPCTHRESHOLDB);
  c = flag_GetFlagValue(prfs_Store(Search), flag_DPCTHRESHOLDC);
  d = flag_GetFlagValue(prfs_Store(Search), flag_DPCTHRESHOLDD);
  e = flag_GetFlagValue(prfs_Store(Search), flag_DPCTHRESHOLDE);

  prfs_DpConnect(Search)->check_after = a*misc_pow_i(b,n)+c*n*n+d*n+e;

#ifdef SMT_LOG
  dpconnect_logMessage(Search, 0, "\tNew threshold is: %d\n",
		       prfs_DpConnect(Search)->check_after);
#endif

  return;
}

void dpconnect_setNewSMTTimelimit(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A PROOFSEARCH object.
  EFFECT:  Calculates the new timeout for the next
           execution.
  CAUTION: Call this function after
           <smt_invocations_killed_timeout> was increased.
***************************************************************/
{
  int a,b,c,d,e,n;
#ifdef SMT_LOG
  dpconnect_logMessage(Search, 1, "Calculating new timeout after %d-th interruption due to timelimits.\n", (prfs_DpConnect(Search)->smt_invocations_killed_timeout));
  dpconnect_logMessage(Search, 0, "\tOld timeout was: %f\n",
		       prfs_DpConnect(Search)->current_smt_timeout);
  dpconnect_logMessage(Search, 0, "\tSo far, %d executions of the SMT procedure had been interrupted.\n", prfs_DpConnect(Search)->smt_invocations_killed_timeout);
#endif

  /* maximum timelimit already rechaed? */
  if ((flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITMAX) !=
       flag_DPCSMTTIMELIMITMAXOFF) &&
      (((float) flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITMAX)) <
       prfs_DpConnect(Search)->current_smt_timeout))
    prfs_DpConnect(Search)->current_smt_timeout =
      (float) flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITMAX);
  else {
    /* a*(b+0.01*c)^n + d*n + e */
    n = prfs_DpConnect(Search)->smt_invocations_total;
    a = flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITA);
    b = flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITB);
    c = flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITC);
    d = flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITD);
    e = flag_GetFlagValue(prfs_Store(Search), flag_DPCSMTTIMELIMITE);

    prfs_DpConnect(Search)->current_smt_timeout =
      a*misc_pow_f((((float) b)+0.01*((float) c)),n)+d*n+e;
  }

#ifdef SMT_LOG
  dpconnect_logMessage(Search, 0, "\tNew timeout is: %f\n",
		       prfs_DpConnect(Search)->current_smt_timeout);
#endif

  return;
}

void dpconnect_invokeExternalSMT(struct PROOFSEARCH_HELP *Search, int timelimit)
/*********************************************************
  INPUT:   A proofsearch object, a flag
           whether the invocation of the SMT procedure
           should be limited wrt. time.
  EFFECT:  Starts the SMT procedure subprocess.
             (1) Depending on the "DpcDebug" flag old SMT
                 {in,out}put is saved.
             (2) Aa clause dictionary is created.
             (3) Then the file containing the exported
                 formulae is processed and input to the
                 SMT procedure is created. The clause
                 dictionary is maintained.
             (4) Fianally the subprocess is started and
                 the process bookeeping is done.
  CAUTION: *NEVER* use this function from outside
           directly, instead use
           <dpconnect_performCheck(Urgent)>. In general
           it will not check whether there exists
           an earlier SMT subprocess which is still
           running.
**********************************************************/
{
  FILE *export_file_filtering, *export_file_non_filtering, *smt_input_file;
  int take, clause_number, pid, formulae_reached, i;
  EVECTOR *dependent_positions, *cmd_args;
  POSITION *max_pos;
  char *read_result, **cmd_args_string, *swap_file_name;
  size_t read_bytes;
  FILE *swap_file;
#ifdef SMT_LOG
  char *log_temp;
  int swap_ignored;
#endif

  swap_file = (FILE*) NULL;
  swap_file_name = (char*) NULL;

#ifdef CHECK
  if (((prfs_DpConnect(Search))->is_checking > 0) ||
      ((prfs_DpConnect(Search))->smt_pid > 0)) {
    misc_StartErrorReport();
    misc_ErrorReport("\nIn dpconnect_invokeExternalSMT:");
    misc_ErrorReport("\nexternal SMT procedure already running.\n");
    misc_FinishErrorReport();
  }
#endif

#ifdef SMT_LOG
  swap_ignored = 0;
#endif

  /* possibly save old input */
  if ((flag_GetFlagValue(prfs_Store(Search), flag_DPCDEBUG) == flag_DPCDEBUGON) &&
      ((prfs_DpConnect(Search))->smt_invocations_total > 0)) {
    sprintf(dpconnect_big_string, "%s.%d",
	    (prfs_DpConnect(Search))->smt_input_filename,
	    (prfs_DpConnect(Search))->smt_invocations_total);
    misc_CopyFile((prfs_DpConnect(Search))->smt_input_filename,
		   dpconnect_big_string);

    sprintf(dpconnect_big_string, "%s.%d",
	    (prfs_DpConnect(Search))->smt_output_filename,
	    (prfs_DpConnect(Search))->smt_invocations_total);
    misc_CopyFile((prfs_DpConnect(Search))->smt_output_filename,
		   dpconnect_big_string);
  }

#ifdef SMT_DEBUG
  clock_StartCounter(clock_SMTCREATEINPUT);
#endif

  /* remove current input and output files (ignore any errors) */
  remove((prfs_DpConnect(Search))->smt_input_filename);
  remove((prfs_DpConnect(Search))->smt_output_filename);
  errno = 0;

  /*     create a clause dictionary
    invariant: dictionary[m] = n iff. SMT clause no. (m-1) corresponds to SPASS clause no.
    n (assumed that SMT starts formulae counting with 1) */
  if ((prfs_DerivedClauses(Search) / 2) > 0)
    dpconnect_setClauseDictionary(Search,
				  clause_dictionary_create((prfs_DerivedClauses(Search) / 2),
							   2));
  else
    dpconnect_setClauseDictionary(Search, clause_dictionary_create(100, 2));

  /* make a flush for export file (filtering), because we need its contents */
  fflush((prfs_DpConnect(Search))->smt_export_filtering_file);

  /* make a flush for export file (non-filtering), because we need its contents */
  fflush((prfs_DpConnect(Search))->smt_export_non_filtering_file);

  /* create input to SMT procedure */
  export_file_filtering =
    fopen((prfs_DpConnect(Search))->smt_export_filtering_filename, "r");
  if (export_file_filtering == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
    misc_UserErrorReport("\n Could not open file for reading formulae context.\n");
    misc_FinishUserErrorReport();
  }

  smt_input_file = fopen((prfs_DpConnect(Search))->smt_input_filename, "w");
  if (smt_input_file == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
    misc_UserErrorReport("\n Could not open file for providing input to SMT.\n");
    misc_FinishUserErrorReport();
  }

  /* print initial comment to input file */
  smt_interface_printSPASSSignature((prfs_DpConnect(Search))->smt_interface,
				    Search,
				    smt_input_file);

  /* print signature */
  fprintf(smt_input_file, "%s\n\n",
	  estring_GetString((prfs_DpConnect(Search))->signature_string));

  /* in the following pass the input and derived formulae */
  formulae_reached = 0;

  /* pass formulae with splitlevel 0 that cannot be subject to filtering */
#ifdef CHECK
  if ((prfs_DpConnect(Search))->asserted_non_filtered_clauses !=
      evector_getNumberOfElements((prfs_DpConnect(Search))->non_filtered_clauses)) {
    misc_StartErrorReport();
    misc_ErrorReport("\nIn dpconnect_invokeExternalSMT:");
    misc_ErrorReport("\nInconsistency in datastructures: number of exported non-filtered");
    misc_ErrorReport("\nunequal to total number in bookkeeping.\n");
      misc_FinishErrorReport();
  }
#endif

  /* insert them into clause dictionary */
  for(i=0;
      i<evector_getNumberOfElements((prfs_DpConnect(Search))->non_filtered_clauses);
      ++i) {
    clause_dictionary_insertClauseNonFiltering(dpconnect_getClauseDictionary(Search),
					       evector_getNthInt((prfs_DpConnect(Search))->non_filtered_clauses, i));
  }

  /* now append the input to the SMT procedure with the contents of the file containing
     non-filtered clauses */
  export_file_non_filtering =
    fopen((prfs_DpConnect(Search))->smt_export_non_filtering_filename, "r");
  if (export_file_non_filtering == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
    misc_UserErrorReport("\n Could not open file for reading formulae context.\n");
    misc_FinishUserErrorReport();
  }

  while((read_bytes=fread((void*) dpconnect_big_string, sizeof(char),
			  _DPCONNECT_BIGSTRING_SIZE, export_file_non_filtering)))
    fwrite((void*) dpconnect_big_string, sizeof(char), read_bytes, smt_input_file);

  fclose(export_file_non_filtering);
  /* ignore any errors */
  errno = 0;

  /* maintain the number of formulae reached */
  formulae_reached += (prfs_DpConnect(Search))->asserted_non_filtered_clauses;

  smt_interface_printComment(prfs_DpConnect(Search)->smt_interface,
			     Search,
			     smt_input_file,
			     "Clauses with splitlevel 0 until *here*\n");

  /* open temporary swap file */
  if (prfs_DpConnect(Search)->swap) {
    swap_file_name = string_Conc(prfs_DpConnect(Search)->smt_export_filtering_filename,
			       ".swap");
    swap_file = fopen(swap_file_name, "w");
    if (swap_file == NULL) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_UserErrorReport("\n Could not open file for swapping.\n");
      misc_FinishUserErrorReport();
    }
  }

  /* read formulae from export file, that possibly will be filtered out
       (1) list of positions the formula depends on
       (2) clause number
       (3) formula itself */
  while (NULL != fgets(dpconnect_big_string, _DPCONNECT_BIGSTRING_SIZE,
		       export_file_filtering)) {
    dependent_positions =
      evector_createFromString(dpconnect_big_string, ",\n",
			       &position_evector_createFromString,
			       &position_evector_destroyPosition,
			       &position_evector_equal,
			       &position_evector_tree_order);

    /* we know that it is sorted; it speeds up the relevance check later */
    evector_setSorted(dependent_positions);

#ifdef CHECK
    if(dpconnect_checkPositionList(dependent_positions) != 1) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_ErrorReport("\n Stored formulae context is corrupt.\n");
      misc_FinishErrorReport();
    }
#endif

    /* we check whether the current formula is relevant, i.e. whether the conjunction
       of the following is fulfilled:
          (1) for the maximum dependent position (wrt. <_t) p holds:
	      ! @SPASS >_b max
          (2) none of the positions the formula depends on, was condensed
    */
    /* condition (1) */

    max_pos = (POSITION*) evector_getMaxElement(dependent_positions);
    if (position_BranchOrderStrict(max_pos, (prfs_DpConnect(Search))->current_spass_position))
      take = 0;
    else
      /* condition (2) */
      take = 1- condensationscope_dependsOnCondensedBranch(prfs_CondensationScope(Search),
							   dependent_positions);
    if (take == 0) {
#ifdef SMT_LOG
      if (prfs_DpConnect(Search)->swap)
	++swap_ignored;
#endif

      read_result = fgets(dpconnect_big_string, _DPCONNECT_BIGSTRING_SIZE,
			  export_file_filtering);
      read_result = fgets(dpconnect_big_string, _DPCONNECT_BIGSTRING_SIZE,
			  export_file_filtering);
#ifdef CHECK
      if ((read_result == (char*) NULL) ||
	  (dpconnect_big_string[strlen(dpconnect_big_string)-1] != '\n')) {
	misc_StartErrorReport();
	misc_ErrorReport("\n In dpconnect_invokeExternalSMT:");
	misc_ErrorReport("\n Automatically generated export file is corrupt.\n");
	misc_FinishErrorReport();
      }
#endif
      /* disallocate dependent position since not used any more */
      evector_deconstructEvector(dependent_positions,
				 &position_evector_destroyPosition);
    } else {
      /* save line 1 of data set */
      if (prfs_DpConnect(Search)->swap)
	fprintf(swap_file, "%s", dpconnect_big_string);

      read_result = fgets(dpconnect_big_string, _DPCONNECT_BIGSTRING_SIZE,
			  export_file_filtering);
#ifdef CHECK
      if ((read_result == (char*) NULL) || (dpconnect_big_string[strlen(dpconnect_big_string)-1] != '\n')) {

	misc_StartErrorReport();
	misc_ErrorReport("\n In dpconnect_invokeExternalSMT:");
	misc_ErrorReport("\n Automatically generated export file is corrupt.\n");
	misc_FinishErrorReport();
      }
#endif

      dpconnect_big_string[strlen(dpconnect_big_string)-1] = '\0';
      clause_number = atoi(dpconnect_big_string);

      /* save line 2 of data set */
      if (prfs_DpConnect(Search)->swap)
	fprintf(swap_file, "%d\n", clause_number);

      read_result = fgets(dpconnect_big_string, _DPCONNECT_BIGSTRING_SIZE,
			  export_file_filtering);
#ifdef CHECK
      if ((read_result == (char*) NULL) || (dpconnect_big_string[strlen(dpconnect_big_string)-1] != '\n')) {
	misc_StartErrorReport();
	misc_ErrorReport("\n In dpconnect_invokeExternalSMT:");
	misc_ErrorReport("\n Automatically generated export file is corrupt.\n");
	misc_FinishErrorReport();
      }
#endif

      /* cut off the final newline */
      dpconnect_big_string[strlen(dpconnect_big_string)-1] = '\0';

      /* save line 3 of data set */
      if (prfs_DpConnect(Search)->swap)
	fprintf(swap_file, "%s\n", dpconnect_big_string);

      /* print it in the format suitable for the used SMT procedure */
      smt_interface_addAssertion((prfs_DpConnect(Search))->smt_interface,
				 Search,
				 smt_input_file,
				 dpconnect_big_string,
				 clause_number,
				 formulae_reached+1);

      /* CAUTION: the following call does not create a deep copy of
	 <dependent_positions>, therefore we do not disallocate it afterwards. */
      clause_dictionary_insertClause(dpconnect_getClauseDictionary(Search),
				     clause_number, dependent_positions);
      ++formulae_reached;
    }
  }

  /* finally, print a check request */
  smt_interface_printCheck((prfs_DpConnect(Search))->smt_interface,
			   Search,
			   smt_input_file);

  /* close export file */
  fclose(export_file_filtering);

  /* close input file for SMT procedure */
  fclose(smt_input_file);

  if (prfs_DpConnect(Search)->swap) {
    /* close temporary swap file */
    fclose(swap_file);

    /* temporarly close old export database file */
    fclose(prfs_DpConnect(Search)->smt_export_filtering_file);

    /* do swapping of export database file */
    if (remove((prfs_DpConnect(Search))->smt_export_filtering_filename) != 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_UserErrorReport("\n Could not remove file.\n");
      misc_FinishUserErrorReport();
    }
    if (rename(swap_file_name, (prfs_DpConnect(Search))->smt_export_filtering_filename) != 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_UserErrorReport("\n Could not rename file.\n");
      misc_FinishUserErrorReport();
    }

    /* free memory for name of swap file */
    string_StringFree(swap_file_name);

    /* reopen new export database file in append mode */
    prfs_DpConnect(Search)->smt_export_filtering_file =
      fopen(prfs_DpConnect(Search)->smt_export_filtering_filename, "a");
    if (prfs_DpConnect(Search)->smt_export_filtering_file == NULL) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\nUnable to open file for exporting clauses (filtering)!");
      misc_FinishUserErrorReport();
    }

    /* so far, for next invocation no swapping necessary */
    prfs_DpConnect(Search)->swap = FALSE;

#ifdef SMT_LOG
    dpconnect_logMessage(Search, 1, "Swapping of export database file done.\n");
    dpconnect_logMessage(Search, 0, "\t%d clauses are not relevant any more.\n",
			 swap_ignored);
#endif
  }

  /* ignore any errors */
  errno=0;

#ifdef SMT_DEBUG
  clock_StopAddPassedTime(clock_SMTCREATEINPUT);
#endif

  /* prepare command line arguments */
  cmd_args = smt_interface_getCommandLineArguments((prfs_DpConnect(Search))->smt_interface,
						   Search);

  cmd_args_string =
    (char**) memory_Malloc(sizeof(char*) * (evector_getNumberOfElements(cmd_args)+2));

  cmd_args_string[0] =
    (char*) memory_Malloc(sizeof(char) *
			  (1+strlen((prfs_DpConnect(Search))->base_name_of_smt_procedure)));
  strcpy(cmd_args_string[0], (prfs_DpConnect(Search))->base_name_of_smt_procedure);
  for(i=0; i<evector_getNumberOfElements(cmd_args); ++i)
    cmd_args_string[i+1] = evector_getNthString(cmd_args, i);
  cmd_args_string[i+1] = (char*) NULL;


  /*****************/
  /*    VFORK      */
  /*****************/
  /* now invoke SMT procedure */
  switch(pid = vfork()) {
  case -1:
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
    misc_UserErrorReport("\n \"vfork\" failed.\n");
    misc_FinishUserErrorReport();
    break;
  case 0:
    /*****************/
    /*    CHILD      */
    /*****************/

    /* change process group of subprocess */
    (void) setpgid(0,0);

    /* the following code until "exec..." is guaranteed to be executed first;
       avoid placing here any time consuming calculations or even worse any
       requests to the father process which might be processed after the vfork;
       otherwise this could cause unefficient behaviour and deadlocks respectively */

    /* redirect smt output to a file */
    if (NULL == freopen((prfs_DpConnect(Search))->smt_output_filename, "w", stdout)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT (child process):");
      misc_UserErrorReport("\n \"freopen (stdout)\" failed.\n");
      _exit(1);
    }

    /* waste the stderr output of DP */
    if (NULL == freopen("/dev/null", "w", stderr)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT (child process):");
      misc_UserErrorReport("\n \"freopen (stderr)\" failed.\n");
      _exit(1);
    }

    (prfs_DpConnect(Search))->is_checking = formulae_reached;

    execv((prfs_DpConnect(Search))->full_name_of_smt_procedure,
	  cmd_args_string);

    /* if <execv> returned, something went wrong; we have to tell this to the
       parent process */
    (prfs_DpConnect(Search))->is_checking = 0;

    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In dpconnect_invokeExternalSMT (child process):");
    misc_UserErrorReport("\n \"execv\" failed.\n");
    _exit(1);

  default:
    /*****************/
    /*    FATHER     */
    /*****************/

    /* check whether <execv> went smoothly */
    if ((prfs_DpConnect(Search))->is_checking == 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_UserErrorReport("\n \"execv\" failed.\n");
      misc_FinishUserErrorReport();
    }

    if (setpriority(PRIO_PROCESS,
		    pid,
		    flag_GetFlagValue(prfs_Store(Search), flag_DPCDPSCHEDULING)-20) != 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In dpconnect_invokeExternalSMT:");
      misc_UserErrorReport("\n \"setrpriority\" failed.\n");
      misc_FinishUserErrorReport();
    }

    /* mark (for <dpconnect_updateState>) whether a global time limit
       is considered or not */
    (prfs_DpConnect(Search))->use_timelimit = ((timelimit==0) ? FALSE : TRUE);

    /* set the stop watch (anyway) */
    clock_InitCounter(clock_SMTSUBPROCESS);
    clock_StartCounter(clock_SMTSUBPROCESS);

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("---------------------------------------------------------------------------------\n");
    printf("SMT procedure subprocess started: PID is %d, given scheduling priority is %d.\n",
	   pid, flag_GetFlagValue(prfs_Store(Search), flag_DPCDPSCHEDULING)-20);
    printf("Full path to SMT procedure: '%s'.\n", (prfs_DpConnect(Search))->full_name_of_smt_procedure);
    printf("Command line was as follows:\n");
    for(i=0; i<evector_getNumberOfElements(cmd_args)+1; ++i)
      printf("%s ", cmd_args_string[i]);
    printf("\n");
    if (! (prfs_DpConnect(Search))->use_timelimit)
      printf("No timeout was specified.\n");
    else
      printf("Timeout for invocation is %f seconds.\n", (prfs_DpConnect(Search))->current_smt_timeout);
    printf("%d formulae passed in total for this step.\n", formulae_reached);
    printf("---------------------------------------------------------------------------------\n");
    smt_debug_end();
#endif
#ifdef SMT_LOG
    dpconnect_logMessage(Search, 1, "SMT procedure subprocess started.\n");
    dpconnect_logMessage(Search, 0, "\tThis is the %d-th invocation.\n",
			 (prfs_DpConnect(Search))->smt_invocations_total+1);
    dpconnect_logMessage(Search, 0, "\tPID is %d, given scheduling priority is %d.\n",
			 pid, flag_GetFlagValue(prfs_Store(Search), flag_DPCDPSCHEDULING)-20);
    log_temp = position_convertToString(prfs_DpConnect(Search)->current_spass_position);
    dpconnect_logMessage(Search, 0, "\tSPASS working position is %s.\n", log_temp);
    string_StringFree(log_temp);
    log_temp = position_convertToString(dpconnect_getCurrentSMTPosition(Search));
    dpconnect_logMessage(Search, 0, "\tSMT working position is %s.\n", log_temp);
    string_StringFree(log_temp);
    log_temp = evector_convertToString(clause_dictionary_getAllDependentPositions(dpconnect_getClauseDictionary(Search)), &position_evector_convertToString, ",");
    dpconnect_logMessage(Search, 0, "\tPositions of dependent branches: %s\n",
			 log_temp);
    string_StringFree(log_temp);
    dpconnect_logMessage(Search, 0, "\tCommand line was as follows:\n\t\t");
    for(i=0; i<evector_getNumberOfElements(cmd_args)+1; ++i)
      dpconnect_logMessage(Search, 0, "%s ", cmd_args_string[i]);
    dpconnect_logMessage(Search, 0, "\n");
    if (! (prfs_DpConnect(Search))->use_timelimit)
      dpconnect_logMessage(Search, 0, "\tTimeout: No timeout was specified.\n");
    else
      dpconnect_logMessage(Search, 0, "\tTimeout: Timeout for invocation is %f seconds.\n", (prfs_DpConnect(Search))->current_smt_timeout);
    dpconnect_logMessage(Search, 0, "\t%d formulae passed in total for this step.\n", formulae_reached);
#endif

    /* free memory needed for commandline arguments; note that the strings in
       <cmd_args_string> point to the same memory location that them from the
       evector <cmd_args>; therefore we must not delete the evector including
       deconstruct of all the elements stored in it */
    for(i=0; i<1+evector_getNumberOfElements(cmd_args); ++i)
      memory_Free(cmd_args_string[i], sizeof(char)*(1+strlen(cmd_args_string[i])));
    memory_Free(cmd_args_string, sizeof(char*) * (evector_getNumberOfElements(cmd_args)+2));

    evector_destroyEvectorWithtoutElements(cmd_args);

    /* bookkeeping */
    (prfs_DpConnect(Search))->smt_pid = pid;

    (prfs_DpConnect(Search))->asserted_since_last_check = 0;
    (prfs_DpConnect(Search))->splittings_since_most_recent_check = 0;
    ++(prfs_DpConnect(Search))->smt_invocations_total;

    /* set new threshold */
    dpconnect_setNewThreshold(Search);

    break;
  } /* switch */
}

int dpconnect_interruptExternalSMT(struct PROOFSEARCH_HELP *Search, int *status,
				   const char *reason, BOOL disalloc_clause_dictionary)
/*********************************************************
  INPUT:   A proofsearch object, an integer value used
           as second return value, an explanation, and a
           flag whether the current clause dictionary
           should be disallocated.
  RETURNS: Whether the process was killed (0), or had
           already terminated with an exit status (1,
           status in <status>) or because it caught a
           signal (2, signal in <status>). In case the
           process is not recognizable any more, -1 is
           returned. <status> is only filled with
           information if <!= NULL>.
  EFFECT:  Possibly kills the DP subprocess.
           Uses the <wait_pid> system call. Updates the
           invocation statistics and the process book-
           keeping. Possibly tidies up with the clause
           dictionary.
           Note that we waste the result of the SMT
           procedure in case it already terminated. It
           should be checked in advance by the calling
           function whether there is already a useful
           result of the SMT procedure.
**********************************************************/
{
  int wait_status;
  int result, stat;

  stat = 0;

  result = -1;

  if ((prfs_DpConnect(Search))->is_checking > 0) {

#ifdef CHECK
    if ((prfs_DpConnect(Search))->smt_pid <= 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In dpconnect_interruptExternalSMT:");
      misc_ErrorReport("\n Bookkeeped an invalid pid.\n");
      misc_FinishErrorReport();
    }
#endif
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Trying to interrupt process %d...\n", (prfs_DpConnect(Search))->smt_pid);
          smt_debug_end();
#endif

    switch (waitpid((prfs_DpConnect(Search))->smt_pid,
		    &wait_status,
		    0 | WNOHANG)) {
    case 0:
      /* so far, not terminated */
      result = 0; /* killed */

      /* statistics */
      ++((prfs_DpConnect(Search))->smt_invocations_killed);

#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Trying to kill process %d...\n", (prfs_DpConnect(Search))->smt_pid);
          smt_debug_end();
#endif
      /* kill the process */
      if (kill((prfs_DpConnect(Search))->smt_pid, dpconnect_term_signal) == -1) {
	/* ignore any error */
	errno = 0;
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Got it...\n");
          smt_debug_end();
#endif
      } else {
        /* wait 20 ms, then kill it using SIGKILL. The old method below doesn't seem
           to work reliably anymore for Cvc3 */

        struct timespec delay;
        delay.tv_sec = 0;
        delay.tv_nsec = 20000000L;  
        (void) nanosleep(&delay, NULL);
#ifdef SMT_DEBUG
          smt_debug_start();
          printf("Using SIGKILL...\n");
          smt_debug_end();
#endif
        (void) kill((prfs_DpConnect(Search))->smt_pid, SIGKILL);

	/* grap the process status in order to prevent zombies */

	/* this call is blocking (otherwise we generate zombies in case the signal
	   has not reached the subprocess at the time the call is done); the blocking
	   is no problem if the process handles the signal and terminates; possibly,
	   <dpconnect_term_signal> can be changed to <SIGKILL>; so far, Yices and Cvc3
	   respect the SIGTERM signal and immediately terminate */
	/* waitpid((prfs_DpConnect(Search))->smt_pid, &wait_status, 0); */

	/* ignore errors */
	errno = 0;
      }

#ifdef SMT_DEBUG
      smt_debug_start();
      printf("SMT procedure subprocess killed [%s].\n", reason);
      smt_debug_end();
#endif
#ifdef SMT_LOG
      dpconnect_logMessage(Search, 1, "SMT procedure subprocess killed [%s].\n", reason);
#endif
      break;

    case -1:
      /* error (process not recognizable etc.), ignore it */
      errno = 0;
      result = -1; /* error */
      break;

    default:
      /* process terminated */

#ifdef SMT_DEBUG
      smt_debug_start();
      printf("SMT procedure already terminated but was aimed to be killed [%s].\n",
	     reason);
      smt_debug_end();
#endif
#ifdef SMT_LOG
      dpconnect_logMessage(Search, 1, "SMT procedure already terminated but was aimed to be killed [%s].\n", reason);
#endif

      if (WIFEXITED(wait_status)) {
	result = 1; /* exit status */
	stat = WEXITSTATUS(status);

	/* statistics */
	++((prfs_DpConnect(Search))->smt_invocations_got_a_result);
      }
      else if (WIFSIGNALED(wait_status)) {
	result = 2; /* signal status */
	stat = WTERMSIG(status);

	/* statistics */
	if (dpconnect_wasAbortedDueToLimits(stat))
	  ++((prfs_DpConnect(Search))->smt_invocations_killed);
      }

      if (status != (int*) NULL)
	*status = stat;
      break;
    } /* switch */
  }

  /* update bookkeeping about external SMT subprocess */
  (prfs_DpConnect(Search))->is_checking = 0;
  (prfs_DpConnect(Search))->smt_pid = -1;

  /* possibly tidy up with clause dictionary */
  if (disalloc_clause_dictionary)
    if (dpconnect_getClauseDictionary(Search) != NULL) {
      clause_dictionary_destroy(dpconnect_getClauseDictionary(Search));
      dpconnect_setClauseDictionary(Search, (CLAUSE_DICTIONARY*) NULL);
    }

  return result;
}

int dpconnect_hasFinishedExternalSMT(struct PROOFSEARCH_HELP *Search, int *status,
				     struct rusage *usage)
/*********************************************************
  INPUT:   A proofsearch object and an integer value used
           as second return value.
  RETURNS: Whether the SMT procedure had already
           terminated with an exit status (1, status in
           <status>) or because it caught a signal (2,
           signal in <status>). In case the process is
           not recognizable (for instance  because the
           SMT procedure currently "sleeps"), -1 is
           returned. <status> is only filled with
           information if <!= NULL>. If the process is
           still running, 0 is returned.
  EFFECT:  Uses the <wait_pid> system call. Updates the
           invocation statistics and the process book-
           keeping.
  CAUTION: *NEVER* call this routine from outside because
           it affects the bookkeeping and might cause
           an inconsistent state of the datastructures!
           Instead use the function
           <dpconnect_updateState> in combination with
           <dpconnect_getCurrentSMTResult>
           to check whether SMT procedure already
           delivered a result and grap it!
**********************************************************/
{
  int wait_status;
  int result, stat, wait_result;

  stat = 0;
  result = -1;

  if ((prfs_DpConnect(Search))->is_checking > 0) {
    wait_result = wait4((prfs_DpConnect(Search))->smt_pid, &wait_status, 0 | WNOHANG,
			usage);

    switch (wait_result) {
    case 0:
      /* so far, not terminated */
      result = 0; /* still running */
      break;

    case -1:
      /* error (process not recognizable etc.), just ignore it */
      errno = 0;
      result = -1; /* error */
      break;

    default:
      /* process terminated */

      if (WIFEXITED(wait_status)) {
	result = 1; /* exit status */
	stat = WEXITSTATUS(status);

	/* statistics */
	++((prfs_DpConnect(Search))->smt_invocations_got_a_result);
      } else if (WIFSIGNALED(wait_status)) {
	result = 2; /* signal status */
	stat = WTERMSIG(status);

	/* statistics */
	if (dpconnect_wasAbortedDueToLimits(stat))
	  ++((prfs_DpConnect(Search))->smt_invocations_killed);
      }

      if (status != (int*) NULL)
	*status = stat; /* either exit status or signal number */

      /* update bookkeeping about external SMT subprocess */
      (prfs_DpConnect(Search))->is_checking = 0;
      (prfs_DpConnect(Search))->smt_pid = -1;

      break;
    } /* switch */
  }

  return result;
}

/**************************************************************/
/* Logging for Statistics                                     */
/**************************************************************/

void dpconnect_startLogging(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  Prints initial string to statistical log file.
***************************************************************/
{
#ifdef SMT_LOG
  fprintf((prfs_DpConnect(Search))->smt_log_file, "Statistical logfile created by SPASS+T\n\n\n");
  dpconnect_appendCurrentTime(Search);
  fprintf((prfs_DpConnect(Search))->smt_log_file, ": Logging started.\n");
#endif
}

void dpconnect_finishLogging(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  Prints message indicating end of logging into
           statistical log file.
***************************************************************/
{
#ifdef SMT_LOG
  fprintf((prfs_DpConnect(Search))->smt_log_file, "\n");
  dpconnect_appendCurrentTime(Search);
  fprintf((prfs_DpConnect(Search))->smt_log_file, ": Logging finished.\n");
#endif
}

void dpconnect_appendCurrentTime(struct PROOFSEARCH_HELP *Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  Appends the statistical log by inserting the
           current time and date.
***************************************************************/
{
#ifdef SMT_LOG
  struct timeval tv;
  struct tm t_broken_up;
  char s[50];

  gettimeofday(&tv, NULL);

  localtime_r(&(tv.tv_sec), &t_broken_up);

  sprintf(s, "%02d:%02d:%02d.%06ld (%02u/%02u/%04u)",
	  t_broken_up.tm_hour, t_broken_up.tm_min, t_broken_up.tm_sec,
	  tv.tv_usec,
	  t_broken_up.tm_mday, t_broken_up.tm_mon+1, t_broken_up.tm_year+1900);

  fprintf((prfs_DpConnect(Search))->smt_log_file, "%s", s);
#endif
}

void dpconnect_logMessage(struct PROOFSEARCH_HELP *Search, int withTime, char *Format, ...)
/**************************************************************
  INPUT:   A proof search object, an integer, a message string
           as a variable argument list.
  EFFECT:  Prints the message into the statistical log file.
           Additionally, if <withTime == 1> also a timemark
           is printed.
***************************************************************/
{
#ifdef SMT_LOG
  va_list varg_list;

  /* initialize vargs */
  va_start(varg_list, Format);

  if (withTime == 1) {
    fprintf((prfs_DpConnect(Search))->smt_log_file, "\n");
    dpconnect_appendCurrentTime(Search);
    fprintf((prfs_DpConnect(Search))->smt_log_file, ": ");
  }

  vfprintf((prfs_DpConnect(Search))->smt_log_file,
	   Format, varg_list);

  /* garbage collection for varg_list */
  va_end(varg_list);
#endif
}

void dpconnect_logSettings(struct PROOFSEARCH_HELP *Search)
{
  LIST relevant_flags, Scan;

  relevant_flags = list_Nil();
  relevant_flags = list_Push((POINTER) flag_DPCTHEORY, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_THAXIOMS, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_RARITH, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_RBLD, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_ITHI, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_IIOE, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_IORDCH, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCDEBUG, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCNONGROUNDEXPORT, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_BRANCHCONDENSATION, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCTHRESHOLDA, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCTHRESHOLDB, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCTHRESHOLDC, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCTHRESHOLDD, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCTHRESHOLDE, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSMTTIMELIMITA, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSMTTIMELIMITB, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSMTTIMELIMITC, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSMTTIMELIMITD, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSMTTIMELIMITE, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCDPSCHEDULING, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCSTRATIO, relevant_flags);
  relevant_flags = list_Push((POINTER) flag_DPCEVIDENCE, relevant_flags);

  dpconnect_logMessage(Search, 1, "DPC-relvant flags were initially set to:\n");

  for(Scan=relevant_flags; !list_Empty(Scan); Scan = list_Cdr(Scan))
    dpconnect_logMessage(Search, 0, "\t%s = %d\n",
			 flag_Name((int) list_Car(Scan)),
			 flag_GetFlagValue(prfs_Store(Search), (int) list_Car(Scan)));

  list_Delete(relevant_flags);
}
