/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          MANAGING THE EXTERNAL SMT PROCEDURE           * */
/* *                                                        * */
/* *  $Module:   dpconnect                                  * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _DPCONNECT_
#define _DPCONNECT_

/**************************************************************/
/* Macros                                                     */
/**************************************************************/

/* some ANSI codes for colors in the shell */
#define NORMAL "\033[0m"
#define BLUE "\033[1;34m"
#define RED "\033[1;31m"
#define GREEN "\033[1;32m"
#define MAGENTA "\033[1;36m"

#define smt_debug_start() printf("\n"BLUE);
#define smt_debug_end() printf(NORMAL"\n"); fflush(stdout)


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <sys/wait.h>
#include <stdarg.h>
#include "clause.h"
#include "term.h"
#include "flags.h"
#include "foldfg.h"
#include "estrings.h"
#include "symbol.h"
#include "position.h"
#include "clause_dictionary.h"
#include "smt_interface.h"
#include "smt_result.h"


/**************************************************************/
/* External procedures                                        */
/**************************************************************/

extern char *mkdtemp (char *__template);
extern int kill (__pid_t __pid, int __sig);


/**************************************************************/
/* Datastructures                                             */
/**************************************************************/

struct PROOFSEARCH_HELP;
struct SMT_INTERFACE_HELP;

typedef struct DPCONNECT_HELP {
  /* filenames for temporary files */
  char *dpdir;
  char *full_name_of_smt_procedure;
  char *base_name_of_smt_procedure;
  char *smt_input_filename;
  char *smt_output_filename;
  char *smt_export_filtering_filename;
  char *smt_export_non_filtering_filename;

  /* file pointer to database of exported clauses with splitlevel >0 */
  FILE *smt_export_filtering_file;

  /* file pointer to database of exported clauses with splitlevel 0 */
  FILE *smt_export_non_filtering_file;

#ifdef SMT_LOG
  /* log file */
  char *smt_log_filename;
  FILE *smt_log_file;
#endif

  /* the problem signature */
  ESTRING *signature_string;

  /* the current SPASS position */
  POSITION *current_spass_position;

  /* the connection interface to the decision procedure */
  struct SMT_INTERFACE_HELP *smt_interface;

  /* information about execution of SMT procedure subprocess */
  int is_checking;
  pid_t smt_pid;

  /* the clause dictionary */
  CLAUSE_DICTIONARY *clause_dictionary;

  /* <non_filtered_clauses[i] = j> iff
        - the i-th exported clause with splitlevel 0 has the clause number j and
        - this clause corresponds to assertion in i-th line of
          <smt_export_non_filtering_filename> */
  EVECTOR* non_filtered_clauses;

  /* the most recent SMT result */
  SMT_RESULT *current_smt_result;

  /* number of splittings since the decision procedure was
     started most recently */
  int splittings_since_most_recent_check;

  /* the current maximum time limit */
  float current_smt_timeout;

  /* is the current execution of the decision procedure bounded by a
     maximum time limit? */
  BOOL use_timelimit;

  /* bookkeeping */
  int smt_invocations_total;
  int smt_invocations_got_a_result;
  int smt_invocations_killed;
  int smt_invocations_killed_timeout;

  /* the total number of exported clauses with splitlevel 0 */
  int asserted_non_filtered_clauses;

  /* the number of clauses that have been exported sine the decision
     procedure was started most recently */
  int asserted_since_last_check;

  /* current threshold parameter */
  int check_after;

  /* did we already got all input clauses that could be exported */
  int got_all_problem_input_clauses;

  /* do we inttend to use swapping of the exported clauses database
     (splitlevel >0) the next time? */
  BOOL swap;
} DPCONNECT_NODE,*DPCONNECT;


/**************************************************************/
/* Allocation and deconstruction                              */
/**************************************************************/

void dpconnect_Free(struct PROOFSEARCH_HELP*);
void dpconnect_Init(struct PROOFSEARCH_HELP*);


/**************************************************************/
/* Access Functions                                           */
/**************************************************************/

POSITION* dpconnect_getCurrentSMTPosition(struct PROOFSEARCH_HELP*);
SMT_RESULT* dpconnect_getCurrentSMTResult(struct PROOFSEARCH_HELP*);
int dpconnect_isChecking(struct PROOFSEARCH_HELP*);
CLAUSE_DICTIONARY* dpconnect_getClauseDictionary(struct PROOFSEARCH_HELP*);
void dpconnect_setClauseDictionary(struct PROOFSEARCH_HELP*, CLAUSE_DICTIONARY* );


/**************************************************************/
/* Proof Control                                              */
/*    Note: These functions *should* be used from outside.    */
/**************************************************************/

BOOL dpconnect_CanContributeMostRecentResult(struct PROOFSEARCH_HELP*);
int dpconnect_updateState(struct PROOFSEARCH_HELP*);
void dpconnect_exportClause(struct PROOFSEARCH_HELP*, CLAUSE);
void dpconnect_registerSplitting(struct PROOFSEARCH_HELP*, CLAUSE);
void dpconnect_registerBacktracking(struct PROOFSEARCH_HELP*, CLAUSE);
void dpconnect_registerCondensation(struct PROOFSEARCH_HELP*, POSITION*);
void dpconnect_gotAllProblemInputClauses(struct PROOFSEARCH_HELP*);
void dpconnect_performCheck(struct PROOFSEARCH_HELP*);
void dpconnect_performCheckUrgent(struct PROOFSEARCH_HELP*);
int dpconnect_hasResult(struct PROOFSEARCH_HELP*);


/**************************************************************/
/* Handling of external SMT procedure                         */
/*    Note: These functions shouldn't be called from outside. */
/**************************************************************/

void dpconnect_setNewThreshold(struct PROOFSEARCH_HELP*);
void dpconnect_setNewSMTTimelimit(struct PROOFSEARCH_HELP*);
void dpconnect_invokeExternalSMT(struct PROOFSEARCH_HELP*, int);
int dpconnect_interruptExternalSMT(struct PROOFSEARCH_HELP*, int*, const char*, BOOL);
int dpconnect_hasFinishedExternalSMT(struct PROOFSEARCH_HELP*, int*, struct rusage*);

/**************************************************************/
/* Logging for Statistics                                     */
/**************************************************************/

void dpconnect_startLogging(struct PROOFSEARCH_HELP*);
void dpconnect_finishLogging(struct PROOFSEARCH_HELP*);
void dpconnect_appendCurrentTime(struct PROOFSEARCH_HELP*);
void dpconnect_logMessage(struct PROOFSEARCH_HELP*, int, char*, ...);
void dpconnect_logSettings(struct PROOFSEARCH_HELP*);

#endif
