/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   EXPANDABLE STRINGS                   * */
/* *                                                        * */
/* *  $Module: estrings                                     * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <ctype.h>
#include "estrings.h"
#include "list.h"
#include "spass_gmp.h"


/**************************************************************/
/* Forward Declarations                                       */
/**************************************************************/

void gmp_FreeInternal(void *ptr, size_t size);


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

BOOL      estring_Equal(const ESTRING* es1, const ESTRING* es2)
/**************************************************************
  INPUT:   Two expandable strings.
  RETURNS: TRUE iff the strings consists of the same characters.
***************************************************************/
{
  return (strcmp(estring_GetString(es1),
		 estring_GetString(es2)) == 0);
}

BOOL      estring_EqualString(const ESTRING* estring, const char* string)
/**************************************************************
  INPUT:   An expandable string, a string.
  RETURNS: TRUE iff the strings consists of the same characters.
***************************************************************/
{
  return (strcmp(estring_GetString(estring),
		 string) == 0);
}

ESTRING*  estring_EmptyString(int initial_size)
/**************************************************************
  INPUT:   Initial number of characters the string can hold
           before it has to be expanded (not counting the '\0'
           marker).
  RETURNS: The empty string.
  EFFECT:  Memory is allocated for the returned string using
           the memory module.
**************************************************************/
{
  ESTRING *local;

  local = (ESTRING*) memory_Malloc(sizeof(ESTRING));

  estring_SetCurrentSize(local, initial_size);
  estring_SetString(local, (char*) memory_Malloc(sizeof(char) * (initial_size+1)));
  (estring_GetString(local))[0] = '\0';

  return local;
}

ESTRING*  estring_FromString(const char *origin)
/**************************************************************
  INPUT:   A string.
  RETURNS: The expandable string holding the original string.
  EFFECT:  Memory is allocated for the returned string using
           the memory module. Its initial size is the length
           of the given original string.
**************************************************************/
{
  ESTRING *local;

  local = estring_EmptyString(strlen(origin));

  strcpy(estring_GetString(local), origin);

  return local;
}


void      estring_StringFree(ESTRING* estring)
/**************************************************************
  INPUT:   An expandable string.
  RETURNS: Nothing.
  EFFECT:  Frees the memory used by the string.
  CAUTION: The string had to be allocated using the memory
           module.
***************************************************************/
{
  memory_Free(estring_GetString(estring), sizeof(char)*(estring->current_size+1));
  memory_Free(estring, sizeof(ESTRING));
}

void      estring_AppendString(ESTRING* dest, const char* string)
/**************************************************************
  INPUT:   An expandable string, a string.
  RETURNS: Nothing.
  EFFECT:  Appends the string <string> at the end of the
           expandable string <dest>. Possibly it is resized.
***************************************************************/
{
  int new_size, needed_size;
  char *save_pointer, *new_pointer;

  needed_size = strlen(string) + strlen(estring_GetString(dest));
  if (needed_size > estring_GetCurrentSize(dest)) {
    /* we have to expand the allocated memory block for the string */

    new_size = estring_GetCurrentSize(dest);
    while(new_size < needed_size)
      new_size *= 2;  /* expand factor is two */

    save_pointer = estring_GetString(dest);
    new_pointer = (char*) memory_Malloc(sizeof(char) * (new_size+1));
    strcpy(new_pointer, estring_GetString(dest));

    memory_Free(estring_GetString(dest), sizeof(char)*(estring_GetCurrentSize(dest)+1));
    estring_SetString(dest, new_pointer);
    estring_SetCurrentSize(dest, new_size);
  }

  /* append string */
  strcat(estring_GetString(dest), string);
}

void      estring_AppendEString(ESTRING* dest, const ESTRING* estring)
/**************************************************************
  INPUT:   Two expandable strings.
  RETURNS: Nothing.
  EFFECT:  Appends the expandable string <string> at the end of
           <dest>. Possibly <dest> is resized. Refactures the
           function <estring_AppendString>.
***************************************************************/
{
  estring_AppendString(dest, estring_GetString(estring));
}

void      estring_AppendStringDelete(ESTRING* dest, char* string)
/**************************************************************
  INPUT:   An expandable string, a string.
  RETURNS: Nothing.
  EFFECT:  Like <estring_AppendString>, but the memory for
           <string> is disallocated.
  CAUTION: The memory for <string> had to be allocated by the
           memory module. Refactures <estring_AppendString>.
***************************************************************/
{
  estring_AppendString(dest, string);

  memory_Free(string, (1+strlen(string))*sizeof(char));
}

void      estring_AppendStringDeleteMPN(ESTRING* dest, char* string)
/**************************************************************
  INPUT:   An expandable string, a string.
  RETURNS: Nothing.
  EFFECT:  Like <estring_AppendString>, but the memory for
           <string> is disallocated using the disallocation
           function of the "spass_gmp" module.
  CAUTION: The memory for <string> had to be allocated
           internally by the GMP library, i.e. using the memory
           allocation function of the "spass_gmp" module.
           Refactures <estring_AppendString>.
***************************************************************/
{
  estring_AppendString(dest, string);

  gmp_FreeInternal((void*) string, (1+strlen(string))*sizeof(char));
}

void      estring_AppendEStringDelete(ESTRING* dest, ESTRING* estring)
/**************************************************************
  INPUT:   Two expandable strings.
  RETURNS: Nothing.
  EFFECT:  Like <estring_AppendEString>, but the memory for
           <string> is disallocated.
  CAUTION: The memory for <string> had to be allocated by the
           memory module. Refactures <estring_AppendEString>.
***************************************************************/
{
  estring_AppendEString(dest, estring);

  estring_StringFree(estring);
}

char*     estring_ExtractString(ESTRING* estring)
/**************************************************************
  INPUT:   An expandable string.
  RETURNS: The string contained in <estring>.
  EFFECT:  Memory for the returned string is allocated,
           memory for the expandabe string is disallocated.
  CAUTION: Everything has to be allocated using the memory
           module.
***************************************************************/
{
  char *local;

  local = memory_Malloc(sizeof(char) * (strlen(estring_GetString(estring))+1));

  strcpy(local, estring_GetString(estring));

  estring_StringFree(estring);

  return local;
}
