/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   EXPANDABLE VECTORS                   * */
/* *                                                        * */
/* *  $Module: evector                                      * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "misc.h"
#include "memory.h"
#include "evector.h"
#include "estrings.h"
#include <stdlib.h>
#include <stdio.h>


/**************************************************************/
/* Internal functions                                         */
/**************************************************************/

static int evector_isMemberUnsorted(EVECTOR *evector, void *element);
static int evector_isMemberSorted(EVECTOR *evector, void *element);
static int evector_compareInt(void *e1, void *e2);
static int evector_equalInt(void *e1, void *e2);
static void evector_printInt(void *e1);
static int evector_compareString(void *e1, void *e2);
static int evector_equalString(void *e1, void *e2);
static void evector_printString(void *e1);
static void* evector_IntFromString(char *string);
static void evector_destroyInt(void *ptr);
static void evector_destroyString(void *ptr);
static void* evector_copyInt(void *e);
static void* evector_copyString(void *e);
static int evector_isSeperator(char c, char *seperators);

static int evector_isMemberUnsorted(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and a pointer to a data.
  RETURNS: Whehter the element is contained in the
           evector.
  EFFECT:  <element> is compared to all elements
           using the internal <equal> function.
**********************************************************/
{
  int i, result;

  result = 0;
  for(i=0; (i<evector->number_of_elements) && (result==0); ++i)
    result = (((*evector->equal)(evector->data[i], element)) == 1);

  return result;
}

static int evector_isMemberSorted(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and a pointer to a data.
  RETURNS: Whether <element> is contained in <evector>
           modulo the internal <equal> function.
  EFFECT:  Binary search is applied.
  CAUTION: <evector> has to be sorted wrt. to the
           internal <compare> function.
**********************************************************/
{
  int Index;

  Index = evector_getIndexToElement(evector, element);

  return((*evector->equal)(evector->data[Index], element) == 1);
}

static int evector_compareInt(void *e1, void *e2)
/*********************************************************
  INPUT:   To pointers.
  RETURNS: Whether the pointers interpreted as ints are
           in relation wrt. '<'.
**********************************************************/
{
  return(*((int*) e1) < *((int*) e2));
}

static int evector_equalInt(void *e1, void *e2)
/*********************************************************
  INPUT:   Two pointers.
  RETURNS: Whether the pointers interpreted as ints are
           in relation wrt. '='.
**********************************************************/
{
  return(*((int*) e1) == *((int*) e2));
}

static void evector_printInt(void *e1)
/*********************************************************
  INPUT:   A pointer.
  RETURNS: Print the value the pointer refers to
           interpreted as an integer value.
**********************************************************/
{
  printf("%d", *((int*) e1));

  return;
}

static int evector_compareString(void *e1, void *e2)
/*********************************************************
  INPUT:   To pointers.
  RETURNS: Whether the pointers interpreted as strings
           are in relation wrt. '<'.
**********************************************************/
{
  if (strcmp(((char*) e1), ((char*) e2)) < 0)
    return 1;
  else
    return 0;
}

static int evector_equalString(void *e1, void *e2)
/*********************************************************
  INPUT:   To pointers.
  RETURNS: Whether the pointers interpreted as strings
           are equal.
**********************************************************/
{
  return ((strcmp(((char*) e1), ((char*) e2)) < 0) == 0);
}

static void evector_printString(void *e1)
/*********************************************************
  INPUT:   A pointer.
  RETURNS: Print the value the pointer refers to
           interpreted as a string.
**********************************************************/
{
  printf("%s", ((char*) e1));

  return;
}

static void* evector_copyInt(void *e)
/*********************************************************
  INPUT:   A pointer.
  RETURNS: The pointer contains are interpreted as an int
           and copied to another pointer.
  EFFECT:  Memory is allocated using the memory module.
**********************************************************/
{
  int *copy;

  copy = evector_mmm_malloc(sizeof(int));

  *copy = *((int*)e);

  return((void*) copy);
}

static void* evector_copyString(void *e)
/*********************************************************
  INPUT:   A pointer.
  RETURNS: The pointer contains are interpreted as a
           string and copied to another pointer.
  EFFECT:  Memory is allocated using the memory module.
**********************************************************/
{
  char *copy;

  copy = evector_mmm_malloc(sizeof(char)* (1+strlen((char*)e)));
  strcpy(copy, (char*)e);

  return((void*) copy);
}

static int evector_isSeperator(char c, char *seperators)
/*********************************************************
  INPUT:   A character and a list of characters.
  RETURNS: Whether <c> is a member of <seperators>.
**********************************************************/
{
  int i, result;

  result = 0;
  for(i=0; i<strlen(seperators); ++i)
    if (c == seperators[i]) {
      result = 1;
      break;
    }

  return result;
}

static void* evector_IntFromString(char *string)
/*********************************************************
  INPUT:   A string.
  RETURNS: The integer converted from <string>.
  EFFECT:  Memory for the integer is allocated using the
           memory module.
**********************************************************/
{
  int *space;

  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = atoi(string);

  return (void*) space;
}

static void evector_destroyInt(void *ptr)
/*********************************************************
  INPUT:   A pointer.
  EFFECT:  Memory for <ptr> is disallocated using the
           memory module. <ptr> is interpreted as a
           pointer to an integer.
**********************************************************/
{
  evector_mmm_free((int*) ptr, sizeof(int));

  return;
}

static void evector_destroyString(void *ptr)
/*********************************************************
  INPUT:   A pointer.
  EFFECT:  Memory for <ptr> is disallocated using the
           memory module. <ptr> is interpreted as a
           pointer to a string.
**********************************************************/
{
  evector_mmm_free((char*) ptr, sizeof(char*) * (1+strlen(ptr)));

  return;
}


/**************************************************************/
/* Creation and deconstruction                                */
/**************************************************************/

EVECTOR* evector_createEvector(int base_size, int factor_expand,
			       int (*equal)(void*, void*), int (*compare)(void*, void*))
/*********************************************************
  INPUT:   Two integers, two function pointers.
  RETURNS: An newly created expandable vector.
  EFFECT:  Creates an evector with space for initially
           <base_size> elements, expansion factor
           <factor_expand> for elements compared using
           the functions <equal> (for deciding equality)
           and <compare> (ordering). Initially, the
           evector is assumed to be sorted.
           Since expansion is done wrt. the current size,
           the initial base size is not stored in the
           evector datastructure.
  CAUTION: Memory is allocated using the memory module.
**********************************************************/
{
  EVECTOR *local;

  local = (EVECTOR*) evector_mmm_malloc(sizeof(EVECTOR));

  local->is_sorted = 1;

  local->contains_int_data = 0;
  local->contains_string_data = 0;

  local->number_of_elements = 0;
  local->size = base_size;
  local->factor_expand = factor_expand;

  local->equal = equal;
  local->compare = compare;

  local->data = (void**) evector_mmm_malloc(sizeof(void*) * local->size);

  return local;
}

EVECTOR* evector_createEvectorInt(int base_size, int factor_expand)
/*********************************************************
  INPUT:   Two integers.
  RETURNS: An newly created expandable vector.
  EFFECT:  Creates an evector with space for initially
           <base_size> elements, expansion factor
           <factor_expand> for integers. Initially, the
           evector is assumed to be sorted.
  CAUTION: The function <evector_createEvector> is used.
**********************************************************/
{
  EVECTOR *result;

  /* call to general function */
  result = evector_createEvector(base_size, factor_expand,
				 &evector_equalInt, &evector_compareInt);

  result->contains_int_data = 1;

  return result;
}

EVECTOR* evector_createEvectorString(int base_size, int factor_expand)
/*********************************************************
  INPUT:   Two integers.
  RETURNS: An newly created expandable vector.
  EFFECT:  Creates an evector with space for initially
           <base_size> elements, expansion factor
           <factor_expand> for strings. Initially, the
           evector is assumed to be sorted.
  CAUTION: The function <evector_createEvector> is used.
**********************************************************/
{
  EVECTOR *result;

  /* call to general function */
  result = evector_createEvector(base_size, factor_expand,
				 &evector_equalString, &evector_compareString);

  result->contains_string_data = 1;

  return result;
}

EVECTOR* evector_createFromString(char* string, char* seperators,
				  void* (*convert_literally)(char*),
				  void (*destroy)(void*),
				  int (*equal)(void*, void*), int (*compare)(void*, void*))
/*********************************************************
  INPUT:   A string, a list of seperators, a function
           pointer to a convert function, a function
           pointer for disallocation of individual
           elements, two function pointers to order the
           data.
  RETURNS: A newly created evector containing the
           elements represented in <string> and seperated
           by <seperators>. Elements contained more than
           once in <string> are inserted only once.
           Initially the vector is supposed to be sorted.
           The base size of the vector is 10 greater the
           number of elements in the string. The
           expansion factor is set to 2.
  EFFECT:  Two passes over the string. First the number
           of elements in computed, than the elements are
           converted using <convert_literally> and
           inserted. If they already were inserted before,
           then <destroy> is invoked.
  CAUTION: Memory is allocated using the memory module.
           Uses the functions <evector_createEvector> and
           <evector_insertSortedNonDuplicate>.
**********************************************************/
{
  EVECTOR *result;
  int number_of_entries;
  int slen;
  char *temp_accu;
  void *temp_convert_result;
  int i, k;

  slen = strlen(string);

  if (slen == 0)
    result = evector_createEvector(10, 2, equal, compare);
  else {
    /* two passes over the string:
        (i)  calculate size
	(ii) convert entries */

    /* calculate the size of the evector */
    number_of_entries = 0;
    for(i=0; i<slen; ) {
      if (!evector_isSeperator(string[i], seperators)) {
	++number_of_entries;
	/* goto next seperating character */
	while(i<slen)
	  if (!evector_isSeperator(string[i], seperators))
	    ++i;
	  else
	    break;
      } else
	++i;
    }

    /* build evector - be generous (10 entries more) */
    result = evector_createEvector(number_of_entries+10, 2, equal, compare);
    temp_accu = (char*) evector_mmm_malloc(sizeof(char) * (slen+1));

    sprintf(temp_accu, "%s", "");
    for(i=0; i<slen; ++i)
      if (!evector_isSeperator(string[i], seperators))
	break;
    for(k=0; i<slen; ) {
      if (!evector_isSeperator(string[i], seperators)) {
	temp_accu[k++] = string[i++];
      }
      if (evector_isSeperator(string[i], seperators) || (i == slen)) {
	/* finish string */
	temp_accu[k] = '\0';

	/* convert accumulator literally */
	temp_convert_result = (*convert_literally)(temp_accu);

#ifdef CHECK
	if (temp_convert_result == NULL)
	  evector_my_error("In evector_createFromString: literal conversion faild.");
#endif

	/* add conversion result to vector */
	if (evector_insertSortedNonDuplicate(result, temp_convert_result) == 0)
	  (*destroy)(temp_convert_result); /* entry already in vector */
	temp_convert_result = NULL;

	k=0;
	temp_accu[k] = '\0';
	/* goto next entry (non-serperating character) */
	while(i<slen)
	  if (evector_isSeperator(string[i], seperators))
	    ++i;
	  else
	    break;
      }
    }

    /* free memory for accumulator */
    evector_mmm_free(temp_accu, sizeof(char) * (slen+1));
  }

  return result;
}

EVECTOR* evector_createFromStringInt(char *string, char* seperators)
/*********************************************************
  INPUT:   A string of integers and a list of seperators.
  RETURNS: The evector containing the int data contained
           in <string>.
  EFFECT:  Uses the function <evector_createFromString>.
**********************************************************/
{
  EVECTOR *result;

  result = evector_createFromString(string, seperators,
				    &evector_IntFromString,
				    &evector_destroyInt,
				    &evector_equalInt,
				    &evector_compareInt);

  result->contains_int_data = 1;

  return result;
}

EVECTOR* evector_copyEvector(EVECTOR *source, void* (*copy_literally)(void*))
/*********************************************************
  INPUT:   An evector and a function pointer.
  RETURNS: A newly created evector containing the data
           of <source>.
  EFFECT:  All elements from <source> are copied using
           the function <copy_literally>. The sequence
           order of <source> is preserved. The returned
           vector is sorted iff. <source> is sorted.
           The expansion factor is taken from
           <source>. Initially the new vector has exactly
           the same number of slots already prepared
           than <source> currently has.
  CAUTION: Memory is allocated using the memory module
           (ie. <copy_literally>).
**********************************************************/
{
  EVECTOR *copy;
  int i;

  copy = evector_createEvector(source->size, source->factor_expand,
			       source->equal, source->compare);

  copy->is_sorted = source->is_sorted;
  copy->contains_int_data = source->contains_int_data;
  copy->contains_string_data = source->contains_string_data;
  copy->number_of_elements = source->number_of_elements;

  for(i=0; i<source->number_of_elements; ++i)
    copy->data[i] = (*copy_literally)(source->data[i]);

  return copy;
}

EVECTOR* evector_copyEvectorInt(EVECTOR *source)
/*********************************************************
  INPUT:   An expandable evector.
  RETURNS: An evector containing int data which is the
           copy of <source>.
  EFFECT:  The function <evector_copyEvector> is used.
**********************************************************/
{
  EVECTOR *copy;

#ifdef CHECK
  if (source->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of combined un-re-boxing */
  copy = evector_copyEvector(source, &evector_copyInt);

  return copy;
}

EVECTOR* evector_copyEvectorString(EVECTOR *source)
/*********************************************************
  INPUT:   An expandable evector.
  RETURNS: An evector containing string data which is the
           copy of <source>.
  EFFECT:  The function <evector_copyEvector> is used.
**********************************************************/
{
  EVECTOR *copy;

#ifdef CHECK
  if (source->contains_string_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of combined un-re-boxing */
  copy = evector_copyEvector(source, &evector_copyString);

  return copy;
}

void evector_destroyEvectorWithtoutElements(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  EFFECT:  The evector datastructure as well as all the
           reserved slotes are disallocated.
  CAUTION: The elements itself have to be freed
           elsewhere.
**********************************************************/
{
  evector->compare = evector->equal = NULL;

  evector_mmm_free(evector->data, sizeof(void*) * evector->size);
  evector->data = NULL;

  evector_mmm_free(evector, sizeof(EVECTOR));

  return;
}

void evector_destroyEvectorInt(EVECTOR *evector)
/*********************************************************
  INPUT:   An expandable vector.
  EFFECT:  All integers stored are disallocated. Then the
           evector itself is freed using the function
           <evector_destroyEvectorWithtoutElements>.
  CAUTION: Memory for all stored elements is freed.
           <evector> has to be of type "int evector".
**********************************************************/
{
  int i;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  for(i=0; i<evector->number_of_elements; ++i)
    evector_mmm_free((int*) evector->data[i], sizeof(int));

  /* call to general function */
  evector_destroyEvectorWithtoutElements(evector);

  return;
}

void evector_destroyEvectorString(EVECTOR *evector)
/*********************************************************
  INPUT:   An expandable vector.
  EFFECT:  All strings stored are disallocated. Then the
           evector itself is freed using the function
           <evector_destroyEvectorWithtoutElements>.
  CAUTION: Memory for all stored elements is freed.
           <evector> has to be of type "string evector".

           For all stored strings it is assumed that they
           have been allocated tightly, ie. a string of
           length <n> was allocated as a sequence of
           <n+1> fields of type <char>.
**********************************************************/
{
  int i;

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  for(i=0; i<evector->number_of_elements; ++i)
    evector_mmm_free((char*) evector->data[i], sizeof(char) * (1+strlen(evector->data[i])));

  /* call to general function */
  evector_destroyEvectorWithtoutElements(evector);

  return;
}

void evector_deconstructEvector(EVECTOR *evector, void (*deconstruct)(void*))
/*********************************************************
  INPUT:   An evector and a function pointer.
  EFFECT:  Memory for the evector as well as for all its
           stored elements is freed. For elements the
           function <deconstruct> is used.
  CAUTION: Memory is freed using the memory module.
**********************************************************/
{
  int i;
  evector->compare = evector->equal = NULL;

  for(i=0; i<evector->number_of_elements; ++i) {
    deconstruct(evector->data[i]);
    evector->data[i] = NULL;
  }

  evector_mmm_free(evector->data, sizeof(void*) * evector->size);
  evector->data = NULL;

  evector_mmm_free(evector, sizeof(EVECTOR));

  return;
}

void evector_setSorted(EVECTOR *evector)
/**************************************************************
  INPUT:   An evector.
  EFFECT:  Assumes the evector to be sorted.
  CAUTION: Can create inconsistent datastructures.
***************************************************************/
{
  evector->is_sorted = 1;

  return;
}

/**************************************************************/
/* Insertion and appending                                     */
/**************************************************************/

void evector_insertAppended(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and a pointer to an element.
  EFFECT:  The element is inserted at the end of the
           evector. If the evector has no free slots left
           (ie. <number_of_elements == size>, then it is
           expanded by <factor_expand> (and copied).
  CAUTION: If this function is used, the evector is
           assumed not to be sorted anymore, albeit the
           insertion (intentionally) would preserve the
           invariant. In this case use the function
           <evector_insertAppendedInt>.
**********************************************************/
{
  void **temp;
  int i;

  /* invariant not necessarily preserved */
  evector->is_sorted = 0;

  if (evector->number_of_elements < evector->size)
    evector->data[(evector->number_of_elements)++] = element;
  else {
    temp = (void**) evector_mmm_malloc(sizeof(void*) * evector->factor_expand * evector->size);
    for(i=0; i<evector->number_of_elements; ++i)
      temp[i] = evector->data[i];
    evector_mmm_free(evector->data, sizeof(void*) * evector->size);

    evector->data = temp;
    evector->data[(evector->number_of_elements)++] = element;
    evector->size *= evector->factor_expand;
  }

  return;
}

void evector_insertAppendedInt(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an integer.
  EFFECT:  Inserts <i> at the end of <evector>. Possibly
           it is extended. The algorithm follows the
           remarks from the description of
           <evector_insertAppended>.
  CAUTION: See description of <evector_insertAppended>.
**********************************************************/
{
  int *space;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of "boxing" */
  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = i;

  /* call to general function */
  evector_insertAppended(evector, (void*) space);

  return;
}

void evector_insertAppendedString(EVECTOR *evector, const char *s)
/*********************************************************
  INPUT:   An evector and a string.
  EFFECT:  Inserts <i> at the end of <evector>. Possibly
           it is extended. The algorithm follows the
           remarks from the description of
           <evector_insertAppended>.
  CAUTION: See description of <evector_insertAppended>.
**********************************************************/
{
  char *space;

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* kind of "boxing" */
  space = (char*) evector_mmm_malloc(sizeof(char) * (1+strlen(s)));
  strcpy(space, s);

  /* call to general function */
  evector_insertAppended(evector, (void*) space);

  return;
}

int evector_insertSortedNonDuplicate(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and an element.
  RETURNS: 0 if <element> was already contained in
           <evector>, 1 if it was inserted correctly.
  EFFECT:  <element> is inserted in <evector> such that
           it remains sorted. First the index of
           <element> is computed using binary search.
           Then it is checked whether the element is
           equal to the element at this position. If so,
           nothing is done, the element is not inserted
           twice. Otherwise the element is (temporarly)
           inserted at the end using the function
           <evector_insertAppended>. Then is moved
           to the correct position by pairwise
           exchanging the elements.
  CAUTION: <evector> must be sorted.
**********************************************************/
{
  int Index, result, k;

#ifdef CHECK
  if (evector->is_sorted != 1)
    evector_my_error("vector is not sorted");
#endif

  if (evector->number_of_elements == 0) {
    result = 1;
    evector_insertAppended(evector, element);
    evector->is_sorted = 1;
  } else {
    Index = evector_getIndexToElement(evector, element);

    if ((*evector->equal)(evector->data[Index], element))
      result = 0;

    else {
      result = 1;
      evector_insertAppended(evector, element);
      evector->is_sorted = 1;
      if ((*evector->compare)(evector->data[Index], element))
	++Index;

      for(k=evector->number_of_elements-1; k>Index; --k)
	evector->data[k] = evector->data[k-1];
      evector->data[Index] = element;
    } /* not already contained in evector */
  }

  return result;
}

/* 0: element already there, 1: inserted */
int evector_insertSortedNonDuplicateInt(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an integer.
  RETURNS: Whether <i> was already contained in <evector>
           (0) or not and was inserted now (1).
  EFFECT:  See <evector_insertSortedNonDuplicate>.
  CAUTION: Uses the function
           <evector_insertSortedNonDuplicate>.
           <evector> must be sorted.
**********************************************************/
{
  int *space;
  int result;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of "boxing" */
  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = i;

  /* call to general function */
  result = evector_insertSortedNonDuplicate(evector, (void*) space);

  if (result == 0) /* not inserted */
    evector_mmm_free(space, sizeof(int));

  return result;
}

/* 0: element already there, 1: inserted */
int evector_insertSortedNonDuplicateString(EVECTOR *evector, const char* s)
/*********************************************************
  INPUT:   An evector and an integer.
  RETURNS: Whether <s> was already contained in <evector>
           (0) or not and was inserted now (1).
  EFFECT:  See <evector_insertSortedNonDuplicate>.
  CAUTION: Uses the function
           <evector_insertSortedNonDuplicate>.
           <evector> must be sorted.
**********************************************************/
{
  char *space;
  int result;

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* kind of "boxing" */
  space = (char*) evector_mmm_malloc(sizeof(char) * (1+strlen(s)));
  strcpy(space, s);

  /* call to general function */
  result = evector_insertSortedNonDuplicate(evector, (void*) space);

  if (result == 0) /* not inserted */
    evector_mmm_free(space, sizeof(char) * (1+strlen(s)));

  return result;
}


/**************************************************************/
/* Search, put and take functions                             */
/**************************************************************/

int evector_isMember(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and an element.
  RETURNS: Wether <element> is contained in <evector> (1)
           or not (0).
  EFFECT:  If <evector> is sorted, binary search is
           applied using <evector_isMemberSorted>;
           otherwise <evector_isMemberNonSorted> is
           invoked which uses linear search.
**********************************************************/
{
  if (evector->number_of_elements == 0)
    return 0;
  else {
    if (evector->is_sorted)
      return(evector_isMemberSorted(evector, element));
    else
      return(evector_isMemberUnsorted(evector, element));
  }
}

int evector_isMemberInt(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an integer.
  RETURNS: Whether <i> is contained in <evector> (1) or
           not (0).
  EFFECT:  Uses <evector_isMember>.
**********************************************************/
{
  int *space;
  int result;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = i;

  /* call to general function */
  result = evector_isMember(evector, space);

  evector_mmm_free(space, sizeof(int));

  return result;
}

int evector_isMemberString(EVECTOR *evector, const char *s)
/*********************************************************
  INPUT:   An evector and a string.
  RETURNS: Whether <s> is contained in <evector> (1) or
           not (0).
  EFFECT:  Uses <evector_isMember>.
**********************************************************/
{
  char *space;
  int result;

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  space = (char*) evector_mmm_malloc(sizeof(char)*(1+strlen(s)));
  strcpy(space, s);

  /* call to general function */
  result = evector_isMember(evector, space);

  evector_mmm_free(space, sizeof(char)*(1+strlen(s)));

  return result;
}

int evector_getIndexToElement(EVECTOR *evector, void *element)
/*********************************************************
  INPUT:   An evector and an element.
  RETURNS: The closed index of <evector> for <element>
           under the assumption that <evector> is sorted.
  EFFECT:  Binary search is applied.
  CAUTION: <evector> must be sorted.
**********************************************************/
{
  int index_left, index_right, index_middle;

#ifdef CHECK
  if (evector->is_sorted != 1)
    evector_my_error("vector is not sorted");
#endif

  index_left = 0;
  index_right = evector->number_of_elements-1;
  index_middle = (index_left + index_right) / 2;
  while(index_left <= index_right) {
    index_middle = (index_left + index_right) / 2;

    if ((*evector->equal)(evector->data[index_middle], element))
      break;
    else if ((*evector->compare)(element, evector->data[index_middle]))
      index_right = index_middle - 1;
    else
      index_left = index_middle + 1;
  }

  return index_middle;
}

int evector_getIndexToElementInt(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an integer.
  RETURNS: The closed index of <evector> for <i>
           under the assumption that <evector> is sorted.
  EFFECT:  Uses <evector_getIndexToElement>.
  CAUTION: <evector> must be sorted.
**********************************************************/
{
  int *space;
  int result;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* boxing */
  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = i;

  /* call to general function */
  result = evector_getIndexToElement(evector, space);

  evector_mmm_free(space, sizeof(int));

  return result;
}

int evector_getIndexToElementString(EVECTOR *evector, char *s)
/*********************************************************
  INPUT:   An evector and a string.
  RETURNS: The closed index of <evector> for <s>
           under the assumption that <evector> is sorted.
  EFFECT:  Uses <evector_getIndexToElement>.
  CAUTION: <evector> must be sorted.
**********************************************************/
{
  char *space;
  int result;

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* boxing */
  space = (char*) evector_mmm_malloc(sizeof(char) * (1+strlen(s)));
  strcpy(space, s);

  /* call to general function */
  result = evector_getIndexToElement(evector, space);

  evector_mmm_free(space, sizeof(char) * (1+strlen(s)));

  return result;
}

int evector_getMemberByKey(EVECTOR *evector, void *element, void **found)
/*********************************************************
  INPUT:   An evector, an element and a pointer to a
           pointer to an element used as return value.
  RETURNS: Whether <element> was found in <evector> (1)
           or not (0).
  EFFECT:  Seeks for the element <element> and gives back
           a pointer to it if found, otherwise <found>
           is set to NULL. <found> can be used by the
           caller to change the element.
           The idea of this function is to deliver the
           result of a search query in such a way that
           the element is identified without accessing
           all its members; this breaks down to provide
           an appropriate internal <equal> function.
**********************************************************/
{
  int Index, result;

  Index = evector_getIndexToElement(evector, element);

  if ((*evector->equal)(evector->data[Index], element) == 1) {
    result = 1;   /* found entry */
    *found = evector->data[Index];
  } else {
    result = 0;   /* entry not found */
    *found = NULL;
  }

  return result;
}

int evector_isIndexInRange(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an index.
  RETURNS: Whether <i> is a valid index for <evector> (1)
           or not (0): <i> is valid if is at least 0 and
           at most <1-number_of_elements>.
  CAUTION: Mostly used internal.
**********************************************************/
{
  if ((i<0) || (i>=evector->number_of_elements))
    return 0;
  else
    return 1;
}

void* evector_getNth(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an index.
  RETURNS: The <i>-th element of <evector>.
  CAUTION: The index <i> must be a valid index for
           <evector>.
**********************************************************/
{

#ifdef CHECK
  if ((i<0) || (i>=evector->number_of_elements))
    evector_my_error("evector_getNthwrong index.");
#endif

  return(evector->data[i]);
}

int evector_getNumberOfElements(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The current number of elements in <evector>.
**********************************************************/
{
  return evector->number_of_elements;
}

void evector_setNth(EVECTOR *evector, int i, void *element)
/*********************************************************
  INPUT:   An evecotr, an index, and an element.
  EFFECT:  Sets the <i>-th element of <evector> to
           <element>.
  CAUTION: The index <i> has to be valid for <evector>.
           Mostly thought for internal usage only.
**********************************************************/
{
#ifdef CHECK
  if ((i<0) || (i>=evector->number_of_elements))
    evector_my_error("evector_getNthwrong index.");
#endif

  /* sortedness not necessarily preserved */
  evector->is_sorted = 0;

  /* assign i-th slot */
  evector->data[i] = element;

  return;
}

int evector_getNthInt(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an index.
  RETURNS: The <i>-th element of <evector>.
  CAUTION: See <evector_getNth>.
           <evector> has to contains integers.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* again kind of unboxing */
  return(*((int*)evector->data[i]));
}

void evector_setNthInt(EVECTOR *evector, int i, int element)
/*********************************************************
  INPUT:   An evecotr, an index, and an integer.
  EFFECT:  Sets the <i>-th element of <evector> to
           <element> using <evector_setNth>.
  CAUTION: The evector must contain integers.
           Mostly thought for internal usage only.
**********************************************************/
{
  int *space;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of boxing */
  space = (int*) evector_mmm_malloc(sizeof(int));
  *space = element;

  evector_setNth(evector, i, space);

  return;
}

char* evector_getNthString(EVECTOR *evector, int i)
/*********************************************************
  INPUT:   An evector and an index.
  RETURNS: The <i>-th element of <evector>.
  CAUTION: See <evector_getNth>.
           <evector> has to contains strings.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* again kind of unboxing */
  return((char*)evector->data[i]);
}

void evector_setNthString(EVECTOR *evector, int i, const char *element)
/*********************************************************
  INPUT:   An evecotr, an index, and a string.
  EFFECT:  Sets the <i>-th element of <evector> to
           <element> using <evector_setNth>.
  CAUTION: The evector must contain strings.
           Mostly thought for internal usage only.
**********************************************************/
{
  char *space;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* kind of boxing */
  space = (char*) evector_mmm_malloc(sizeof(char)*(1+strlen(element)));
  strcpy(space, element);

  evector_setNth(evector, i, space);

  return;
}


/**************************************************************/
/* Pretty printing and conversion                             */
/**************************************************************/

void evector_printEVector(EVECTOR *evector, void (*print)(void*))
/*********************************************************
  INPUT:   An evector and a function pointer.
  EFFECT:  Prints out all elements stored in <evector>.
           For printing the single elements <print> is
           used.
**********************************************************/
{
  int i;

  printf("< ");
  for(i=0; i<evector->number_of_elements; ++i) {
    print(evector->data[i]);
    printf(" ");
  }
  printf("> ");

  return;
}

void evector_printEVectorInt(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector and a function pointer.
  EFFECT:  Prints out all integers stored in <evector>.
           Uses the function <evector_printEVector>.
  CAUTION: <evector> must contain integers.
**********************************************************/
{

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  evector_printEVector(evector, &evector_printInt);

  return;
}

void evector_printEVectorString(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector and a function pointer.
  EFFECT:  Prints out all strings stored in <evector>.
           Uses the function <evector_printEVector>.
  CAUTION: <evector> must contain strings.
**********************************************************/
{

#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  evector_printEVector(evector, &evector_printString);

  return;
}

char* evector_convertToString(EVECTOR *evector, char* (*convert_literally)(void*),
			      const char* seperator_string)
/*********************************************************
  INPUT:   An evector, a function pointer, and a
           string.
  RETURNS: The string representation of <evector>.
  EFFECT:  Single elements are converted to string using
           <convert_literally>. Their string
           representations are seperated by
           <seperator_string>.
  CAUTION: The string allocations done by
           <convert_literally> must be "tight" in the
           sense as explained in the description of the
           function <evector_destroyEvectorString>.
           Otherwise the memory management is confused.
**********************************************************/
{
  ESTRING *local;
  char *temp;
  int i;

  local = estring_EmptyString(256);

  if (evector->number_of_elements > 0) {
    for(i=0; i<evector_getNumberOfElements(evector); ++i) {
      temp = convert_literally(evector_getNth(evector, i));
      estring_AppendStringDelete(local, temp);

      if (i<evector_getNumberOfElements(evector)-1)
	estring_AppendString(local, seperator_string);
    }
  }

  return estring_ExtractString(local);
}

char* evector_convertToStringInt(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector of integers, a function pointer,
           and a string.
  RETURNS: The string representation of <evector>.
  EFFECT:  <evector> must contain integers.
           The function <evector_convertToString> is
           *NOT* used.
**********************************************************/
{
  char *result;
  int i;

#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  if (evector->number_of_elements == 0) {
    result = evector_mmm_malloc(sizeof(char) * 1);
    result[0] = '\0';
  } else {
    result = (char*) evector_mmm_malloc(sizeof(char) * (10*evector->number_of_elements+1));

    sprintf(result, "%s", "");
    for(i=0; i<evector->number_of_elements; ++i)
      sprintf(result, "%s%d%s", result, evector_getNthInt(evector, i),
	      ((i<evector->number_of_elements-1) ? "," : "" ));
  }

  return result;
}

char* evector_convertToStringString(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector of strings, a function pointer,
           and a string.
  RETURNS: The string representation of <evector>.
  EFFECT:  <evector> must contain strings.
           The function <evector_convertToString> is
           *NOT* used.
**********************************************************/
{
  ESTRING *local;
  int i;


#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  local = estring_EmptyString(256);

  if (evector->number_of_elements > 0) {
    for(i=0; i<evector_getNumberOfElements(evector); ++i) {
      estring_AppendString(local, "\"");
      estring_AppendString(local, evector_getNthString(evector, i));
      estring_AppendString(local, "\"");

      if (i<evector->number_of_elements-1)
	estring_AppendString(local, ", ");
    }
  }

  return estring_ExtractString(local);
}

/**************************************************************/
/* Type specific functions                                    */
/**************************************************************/

int evector_containsIntElements(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: 1 if <evector> is known to contains integers,
           0 otherwise.
**********************************************************/
{
  return(evector->contains_int_data);
}

int evector_containsStringElements(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: 1 if <evector> is known to contains strings,
           0 otherwise.
**********************************************************/
{
  return(evector->contains_string_data);
}


/**************************************************************/
/* Equality, minumum and maximum elements                     */
/**************************************************************/

int evector_equal(EVECTOR *v1, EVECTOR *v2)
/*********************************************************
  INPUT:   Two evectors.
  RETURNS: Whether <v1> and <v2> are equal (1) or not
           (0).
  EFFECT:  <v1> and <v2> are said to be equal, if they
           have the same number of elements and they
           are pairwisely equal. Internal data, such as
           the expansion factor or the currently
           allocated slots are not considered for
           comparison.
  CAUTION: Elementwise equality is decided using the
           internal <equal> function.
**********************************************************/
{
  int result;
  int i;

  result = 0;

  if (v1->number_of_elements == v2->number_of_elements) {
    result = 1;
    for(i=0; i<v1->number_of_elements; ++i)
      if ( ((*(v1->equal))(v1->data[i], v2->data[i]) != 1) ||
	   ((*(v2->equal))(v1->data[i], v2->data[i]) != 1) ) {
	result = 0;
	break;
      }
  }

  return result;
}


void* evector_getMinElement(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The minimal element of <evector> wrt. to the
           internal order.
  EFFECT:  If <evector> is sorted, the first element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
**********************************************************/
{
  void *result;
  int i;

#ifdef CHECK
  if (evector->number_of_elements <= 0)
    evector_my_error("evector is empty");
#endif
  if (evector->is_sorted == 1)
    result = evector->data[0];
  else {
    result = evector->data[0];
    for(i=1; i<evector->number_of_elements; ++i)
      if ( (*(evector->compare))(evector->data[i], result) )
	result = evector->data[i];
  }

  return result;
}

void* evector_getMaxElement(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The maximal element of <evector> wrt. to the
           internal order.
  EFFECT:  If <evector> is sorted, the last element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
**********************************************************/
{
  void *result;
  int i;

#ifdef CHECK
  if (evector->number_of_elements <= 0)
    evector_my_error("evector is empty");
#endif

  if (evector->is_sorted == 1)
    result = evector->data[evector->number_of_elements-1];
  else {
    result = evector->data[0];
    for(i=1; i<evector->number_of_elements; ++i)
      if ( (*(evector->compare))(result, evector->data[i]) )
	result = evector->data[i];
  }

  return result;
}

int evector_getMinElementInt(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The minimal integer contained in <evector>.
  EFFECT:  If <evector> is sorted, the first element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
           <evector> has to contains integers.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* unboxing */
  return( *((int*) evector_getMinElement(evector)));
}

int evector_getMaxElementInt(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The maximal integer contained in <evector>.
  EFFECT:  If <evector> is sorted, the first element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
           <evector> has to contains integers.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_int_data == 0)
    evector_my_error("vector doesn't contain int data");
#endif

  /* unboxing */
  return( *((int*) evector_getMaxElement(evector)));
}

char* evector_getMinElementString(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The minimal string contained in <evector> wrt.
           to the lexicographic order on strings.
  EFFECT:  If <evector> is sorted, the first element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
           <evector> has to contains strings.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* unboxing */
  return( ((char*) evector_getMinElement(evector)));
}

char* evector_getMaxElementString(EVECTOR *evector)
/*********************************************************
  INPUT:   An evector.
  RETURNS: The maximal string contained in <evector> wrt.
           to the lexicographic order on strings.
  EFFECT:  If <evector> is sorted, the last element is
           returned. Otherwise all elements are scanned.
  CAUTION: O(1) if sorted vs. O(n) for the non-sorted
           case!
           <evector> has to contains strings.
**********************************************************/
{
#ifdef CHECK
  if (evector->contains_string_data == 0)
    evector_my_error("vector doesn't contain string data");
#endif

  /* unboxing */
  return( ((char*) evector_getMaxElement(evector)));
}

