/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   EXPANDABLE VECTORS                   * */
/* *                                                        * */
/* *  $Module: evector                                      * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _EVECTOR_
#define _EVECTOR_

/**************************************************************/
/* Macros                                                     */
/**************************************************************/

#define evector_mmm_malloc(size) memory_Malloc(size)
#define evector_mmm_free(ptr, size) memory_Free(ptr, size)

#define evector_my_error(msg) { misc_StartErrorReport(); misc_ErrorReport("\n"msg"\n"); misc_FinishErrorReport(); }


/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

/* Notes:
 *
 *   (1) <size> contains the number of elements for which storage is already available,
 *       <number_of_elements> instead the number of elements already contained in the
 *       vector. Necessarily the invariant <number_of_elements <= size> has to be
 *       preserved.
 *   (2) The vector is expanded if all slots are filled. The expansion is done wrt. to the
         current size, not wrt. to the initial size.
 *   (3) Until now there is no collapsing mechanism, hence the vector can only grow.
 *
 */

typedef struct {
  int number_of_elements;
  int factor_expand;
  int size;

  int is_sorted;

  int contains_int_data;
  int contains_string_data;

  void **data;

  int (*compare)(void*, void*);
  int (*equal)(void*, void*);
} EVECTOR;


/**************************************************************/
/* Creation and deconstruction                                */
/**************************************************************/

EVECTOR* evector_createEvector(int base_size, int factor_expand,
				      int (*equal)(void*, void*), int (*compare)(void*, void*));
EVECTOR* evector_createEvectorInt(int base_size, int factor_expand);
EVECTOR* evector_createEvectorString(int base_size, int factor_expand);
EVECTOR* evector_createFromString(char* string, char* seperators,
				  void* (*convert_literally)(char*),
				  void (*destroy)(void*),
				  int (*equal)(void*, void*), int (*compare)(void*, void*));
EVECTOR* evector_createFromStringInt(char *string, char* seperators);

EVECTOR* evector_copyEvector(EVECTOR *source, void* (*copy_literally)(void*));
EVECTOR* evector_copyEvectorInt(EVECTOR *source);
EVECTOR* evector_copyEvectorString(EVECTOR *source);

void evector_destroyEvectorWithtoutElements(EVECTOR *evector);
void evector_deconstructEvector(EVECTOR *evector, void (*deconstruct)(void*));
void evector_destroyEvectorInt(EVECTOR *evector);
void evector_destroyEvectorString(EVECTOR *evector);

void evector_setSorted(EVECTOR *evector);

/**************************************************************/
/* Insertion and appending                                     */
/**************************************************************/

void evector_insertAppended(EVECTOR *evector, void *element);
void evector_insertAppendedInt(EVECTOR *evector, int i);
void evector_insertAppendedString(EVECTOR *evector, const char *s);

/* 0: element already there, 1: inserted */
int evector_insertSortedNonDuplicate(EVECTOR *evector, void *element);
/* 0: element already there, 1: inserted */
int evector_insertSortedNonDuplicateInt(EVECTOR *evector, int i);
/* 0: element already there, 1: inserted */
int evector_insertSortedNonDuplicateString(EVECTOR *evector, const char* s);


/**************************************************************/
/* Search, put and take functions                             */
/**************************************************************/

int evector_isMember(EVECTOR *evector, void *element);
int evector_isMemberInt(EVECTOR *evector, int i);
int evector_isMemberString(EVECTOR *evector, const char *s);
int evector_getMemberByKey(EVECTOR *evector, void *element, void **found);
int evector_getIndexToElement(EVECTOR *evector, void *element);
int evector_getIndexToElementInt(EVECTOR *evector, int i);
int evector_getIndexToElementString(EVECTOR *evector, char *s);
void* evector_getNth(EVECTOR *evector, int i);
void evector_setNth(EVECTOR *evector, int i, void *element);
int evector_getNumberOfElements(EVECTOR *evector);
int evector_getNthInt(EVECTOR *evector, int i);
void evector_setNthInt(EVECTOR *evector, int i, int element);
char* evector_getNthString(EVECTOR *evector, int i);
void evector_setNthString(EVECTOR *evector, int i, const char *element);
int evector_isIndexInRange(EVECTOR *evector, int i);


/**************************************************************/
/* Pretty printing and conversion                             */
/**************************************************************/

void evector_printEVector(EVECTOR *evector, void (*print)(void*));
void evector_printEVectorInt(EVECTOR *evector);
void evector_printEVectorString(EVECTOR *evector);

char* evector_convertToString(EVECTOR *evector, char* (*convert_literally)(void*),
			      const char* seperator_string);
char* evector_convertToStringInt(EVECTOR *evector);
char* evector_convertToStringString(EVECTOR *evector);


/**************************************************************/
/* Type specific functions                                    */
/**************************************************************/

int evector_containsIntElements(EVECTOR *evector);
int evector_containsStringElements(EVECTOR *evector);


/**************************************************************/
/* Equality, minumum and maximum elements                     */
/**************************************************************/

/* v1 and v2 are equal, if they have the same number of elements,
   and they are pairwisely equal */
int evector_equal(EVECTOR *v1, EVECTOR *v2);
void* evector_getMinElement(EVECTOR *evector);
void* evector_getMaxElement(EVECTOR *evector);
int evector_getMinElementInt(EVECTOR *evector);
int evector_getMaxElementInt(EVECTOR *evector);
char* evector_getMinElementString(EVECTOR *evector);
char* evector_getMaxElementString(EVECTOR *evector);


#endif
