/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                MISCELLANEOUS                           * */
/* *                                                        * */
/* *  $Module:   MISC                                       * */ 
/* *                                                        * */
/* *  Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001      * */
/* *  MPI fuer Informatik                                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.57 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 2005/10/04 13:08:31 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: misc.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "misc.h"
#include <errno.h>
#include <unistd.h>


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void misc_StartErrorReporting(const char* file, int line)  
{ 
  fflush(stdout);
  fprintf(misc_ERROROUT,"\n\tError in file %s at line %d\n",file, line);
}

void misc_FinishErrorReport(void) 
{ 
  fputs("\n Please report this error via email to spass@mpi-sb.mpg.de including\n the SPASS version, input problem, options, operating system.\n",misc_ERROROUT);
  misc_DumpCore();
}

void misc_ErrorReport(const char * Format, ...)
{
  va_list args;
  va_start(args,Format);
  vfprintf(misc_ERROROUT,Format,args);
  va_end(args);
}

void misc_UserErrorReport(const char * Format, ...)
{
  va_list args;
  va_start(args,Format);
  vfprintf(misc_USERERROROUT,Format,args);
  va_end(args);
}


void misc_PrintChar(NAT Number, char Character)
/**************************************************************
  INPUT:   A positive number and a character.
  RETURNS: Nothing.
  EFFECT:  Prints as many times the same character as the number 
           demands.
***************************************************************/
{
  NAT Counter;
  for (Counter = 1; Counter <= Number; Counter++)
    putchar(Character);
}


void misc_DumpCore(void)
/**************************************************************
  INPUT:   Nothing.
  RETURNS: Nothing.
  EFFECT:  Flushes the streams then dumps a core.
***************************************************************/
{
  fputs("\n\n", misc_ERROROUT);
  fflush(misc_ERROROUT);
  fflush(stdout);
  fflush(stderr);
  abort();
}

void misc_DumpCoreOut(const char* String)
/**************************************************************
  INPUT:   A  string.
  RETURNS: Nothing.
  EFFECT:  Prints <String> and then dumps a core.
***************************************************************/
{
  fprintf(stderr, "\n %s \n", String);
  misc_DumpCore();
}



int misc_ReturnValue(void)
{
  return 0;
}


int misc_Max(int a, int b)
{
  if (a > b)
    return a;
  else
    return b;
}

FILE* misc_OpenFile(const char* Name, const char* Mode)
/**************************************************************
  INPUT:   The name of a file and a string containing the mode
           for opening the file (see fopen(3)).
           Examples for Mode are "r" for reading and "w" for writing.
  RETURNS: The FILE pointer, if the file was successfully opened.
  EFFECT:  If it wasn't possible to open the file with the
           requested mode, an error message is printed and the
           program exits.
***************************************************************/
{
  FILE* File;

  File = fopen(Name,Mode);

  if (File == (FILE*)NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n\tError in opening file %s for %s !\n\n", Name, 
			 (Mode[0] == 'r' ? "reading" :
			  (Mode[0] == 'w' ? "writing" : "i/o operations")));
    misc_FinishUserErrorReport();
  }  

  return File;
}

void misc_CloseFile(FILE* File, const char* Name)
/**************************************************************
  INPUT:   A FILE and its name.
  RETURNS: Nothing.
  EFFECT:  Closes the file. If an error occurs, a message is
           printed and the program exits.
***************************************************************/
{
  int Result;

  Result = fclose(File);

  if (Result != 0) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n\tError in closing file %s !\n\n", Name);
    misc_FinishUserErrorReport();
  }  
}

/* SZ > */
int   misc_ExistsFile(const char* filename)
/**************************************************************
  INPUT:   A FILE and its name.
  RETURNS: Whether the file exists (1) or not (0).
***************************************************************/
{
  if (access(filename, F_OK) == 0)
    return 1;
  else {
    errno = 0;
    return 0;
  }
}
/* < SZ */

/* SZ > */
void  misc_CopyFile(const char *src_name, const char* dst_name)
/**************************************************************
  INPUT:   The filename of the source file, the filename of 
           the target file.
  RETURNS: Whether the file exists (1) or not (0).
  CAUTION: If already existing, file <dst_name> will be 
           overwritten.
***************************************************************/
{
  char buffer[4097];
  FILE *src, *dst;
  size_t read_bytes;  

#ifdef CHECK
  if (!misc_ExistsFile(src_name)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nIn misc_CopyFile: file \"%s\" does not exist.\n", src_name);
    misc_FinishUserErrorReport();
  }
#endif

  src = fopen(src_name, "r");
  dst = fopen(dst_name, "w");

  while((read_bytes=fread((void*) buffer, sizeof(char), 4096, src)))
    fwrite((void*) buffer, sizeof(char), read_bytes, dst);
  
  fclose(src);
  fclose(dst);
  
  /* ignore any errors */
  errno = 0;
}

int misc_pow_i(int a, int n)
/**************************************************************
  INPUT:   An integer base, an integer exponent.
  RETURNS: a^n
***************************************************************/
{
  int result;
  int i;
  
  result = 1.0;
  
  for(i=0; i<n; ++i)
    result *= a;
  
  return result;
}
/* < SZ */

float misc_pow_f(float a, int n)
/**************************************************************
  INPUT:   A floating point base, an integer exponent.
  RETURNS: a^n
***************************************************************/
{
  float result;
  int i;
  
  result = 1.0;
  
  for(i=0; i<n; ++i)
    result *= a;
  
  return result;
}
/* < SZ */
