/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                  SEARCH TREE POSITIONS                 * */
/* *                                                        * */
/* *  $Module:   POSITION                                   * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "position.h"
#include "memory.h"
#include "misc.h"


/**************************************************************/
/* Internal functions                                         */
/**************************************************************/

int position_getMostSignificantBit(POSITION_FIELD_ENTRY entry);
void position_setLeastSignificantBit(POSITION_FIELD_ENTRY *entry, int v);
int position_mustbeExpanded(POSITION *p);


int position_getMostSignificantBit(POSITION_FIELD_ENTRY component)
/*********************************************************
  INPUT:   An component of a position filed.
  RETURNS: The most significant bit of the component
           <component>.
**********************************************************/
{
  POSITION_FIELD_ENTRY value;

  value = component;
  value &= ((POSITION_FIELD_ENTRY)1) << ((sizeof(POSITION_FIELD_ENTRY)*CHAR_BIT)-1);
  value = value >> ((sizeof(POSITION_FIELD_ENTRY)*CHAR_BIT)-1);
  value &= ((POSITION_FIELD_ENTRY)1);

  return value;
}

void position_setLeastSignificantBit(POSITION_FIELD_ENTRY *component, int v)
/*********************************************************
  INPUT:  A component of a position field, a value.
  EFFECT: Sets the least significant bit of <component>
          to <v>.
**********************************************************/
{
#ifdef CHECK
  if ((v != 0) && (v != 1))
    position_my_error("position_setLeastSignificantBit: Non-bit value given.");
#endif

  if (v == 1)
    (*component) |= ((POSITION_FIELD_ENTRY)1);
  else
    (*component) &= (~((POSITION_FIELD_ENTRY)1));
}

int position_mustbeExpanded(POSITION *p)
/*********************************************************
  INPUT:   A position.
  RETURNS: Whether the position field of <p> is
           completely filled (1) and thus must be
           expanded the next time the position string is
           extended.
**********************************************************/
{
  if (position_getDepth(p) >=
      (CHAR_BIT * sizeof(POSITION_FIELD_ENTRY) * position_getFieldLength(p)))
    return 1;
  else
    return 0;
}

/**************************************************************/
/* Position Access Functions                                  */
/**************************************************************/

void position_set(POSITION *p, int i, int v)
/*********************************************************
  INPUT:   A position, an index and a value.
  EFFECT:  Sets the the <i>-th bit of <p> to <v>.
  CAUTION: <i> has to be a correct index for <p>.
**********************************************************/
{
  int field_index, bit;

#ifdef CHECK
  if ((v!=0) && (v!=1))
    position_my_error("position_set: Non-bit value given.");
#endif

  position_getIndex(i, &field_index, &bit);

#ifdef CHECK
  if (!(field_index < p->field_length))
    position_my_error("position_set: wrong index given.");
#endif

  if (v == 1)
    p->field[field_index] |= ((POSITION_FIELD_ENTRY)1 << bit);
  else
    p->field[field_index] &= (~((POSITION_FIELD_ENTRY)1 << bit));
}

int position_get(POSITION *p, int i)
/*********************************************************
  INPUT:   A position and an index.
  RETURNS: The value at the <i>-th bit of <p>.
  CAUTION: <i> has to be a correct index for <p>.
**********************************************************/
{
  int field_index, bit;
  POSITION_FIELD_ENTRY test;

  position_getIndex(i, &field_index, &bit);

#ifdef CHECK
  if (!(field_index <= p->field_length))
    position_my_error("position_get: wrong index given.");
#endif

  test = (p->field[field_index])>>(bit);
  test &= (POSITION_FIELD_ENTRY)1;

  if (test)
    return 1;
  else
    return 0;
}

void position_getIndex(int i, int *field_index, int *bit_index)
/*********************************************************
  INPUT:   An index, two integers used as return values.
  EFFECT:  <field_index> is set to the index of the field
           component that contains the <i>-th value and
           <bit_index> to the index of the bit within the
           <field_index>-th component.
**********************************************************/
{
  int n;

  *field_index = 0;
  n = sizeof(POSITION_FIELD_ENTRY) * CHAR_BIT;

  while (i >= n) {
    ++(*field_index);
    n += (sizeof(POSITION_FIELD_ENTRY) * CHAR_BIT);
  }

  *bit_index = i - (n-(sizeof(POSITION_FIELD_ENTRY) * CHAR_BIT));

/*   printf("........%d, %d, %d\n", i, *field_index, *bit_index); */
}


/**************************************************************/
/* Creation and disallocation                                 */
/**************************************************************/

POSITION* position_createPosition(void)
/*********************************************************
  RETURNS: A newly created position.
  EFFECT:  The field is set to NULL (i.e. contains
           nothing so far), the depth is set to zero.
**********************************************************/
{
  POSITION *local;

  local = (POSITION*) position_mmm_malloc(sizeof(POSITION));

  local->depth = 0;
  local->field_length = 0;
  local->field = NULL;

  return local;
}

POSITION* position_createRootPosition(void)
/*********************************************************
  RETURNS: A newly created position which represent the
           root position \epsilon.
**********************************************************/
{
  return(position_createPosition());
}

POSITION* position_createFromString(char *encoding)
/*********************************************************
  INPUT:   A string.
  RETURNS: The position that is represented by
           <encoding>.
  EFFECT:  The string representation of a position is
           either "e" or "" for the empty string and a
           non-enpty string over '0' and '1' for a
           non-root position.
**********************************************************/
{
  POSITION *local;
  int i;

  local = position_createRootPosition();

  /* somehow ambiguent, the empty position can be encoded using "e" and "" */
  if (strcmp(encoding, "e") == 0)
    return local;

  for(i=0; i<strlen(encoding); ++i) {
#ifdef CHECK
    if (!((encoding[i] == '1') || (encoding[i] == '0')))
      position_my_error("position_createFromString: non-bit-string encoding given.");
#endif
    if (encoding[i] == '0')
      position_append0(local);
    else if (encoding[i] == '1')
      position_append1(local);
  }

  return local;
}

POSITION* position_copyPosition(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: A deep copy of <p>.
  EFFECT:  The field data is copied, memory is allocated.
**********************************************************/
{
  POSITION *local;
  int i;

  local = (POSITION*) position_mmm_malloc(sizeof(POSITION));

  local->depth = p->depth;
  local->field_length = p->field_length;

  if (p->field_length > 0)
    local->field =
      (POSITION_FIELD_ENTRY*) (position_mmm_malloc(sizeof(POSITION_FIELD_ENTRY)*p->field_length));
  else
    local->field = NULL;

  for(i=0; i<p->field_length; ++i)
    local->field[i] = p->field[i];

  return local;
}

void position_destroyPosition(POSITION* p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Frees the memory eoccupied by <p>.
**********************************************************/
{
  if (position_getFieldLength(p) > 0)
    position_mmm_free(p->field, sizeof(POSITION_FIELD_ENTRY)*(position_getFieldLength(p)));

  p->field_length = p->depth = 0;
  position_mmm_free(p, sizeof(POSITION));
}


/**************************************************************/
/* Specials                                                   */
/**************************************************************/

void position_expandField(POSITION *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Reallocates the position field of <p> and
           enlarges it by one component. Data is copied
           from the original filed and its memory is
           freed.
  CAUTION: This function enlarges the field no matter
           whether the field is already full or has still
           space for further bits.
**********************************************************/
{
  POSITION_FIELD_ENTRY *local;
  int i;

  local = (POSITION_FIELD_ENTRY*) position_mmm_malloc(sizeof(POSITION_FIELD_ENTRY) * (position_getFieldLength(p)+1));

  for(i=0; i<position_getFieldLength(p); ++i)
    local[i] = p->field[i];
  local[i] = (POSITION_FIELD_ENTRY) 0;

  if (position_getFieldLength(p) > 0)
    position_mmm_free(p->field, sizeof(POSITION_FIELD_ENTRY) * position_getFieldLength(p));
  p->field = local;
  position_incFieldLength(p);
}

POSITION* position_getAbovePosition(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: A newly allocated position which represents
           the position above <p>.
  EFFECT:  Uses <position_above>.
  CAUTION: <p> must not be the root position.
**********************************************************/
{
  POSITION* local;

  local = position_copyPosition(p);

  position_above(local);

  return local;
}

void position_above(POSITION* p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Sets <p> to the position above <p>.
  CAUTION: <p> must not be the root position.
**********************************************************/
{
#ifdef CHECK
  if (position_isRootPosition(p))
    position_my_error("position_above: root position given.");
#endif

  position_set(p, position_getDepth(p)-1, 0);
  position_decDepth(p);
}

POSITION* position_get0Successor(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: A newly allocated position that is the
           0-successor of p, i.e. encodes the position of
           the left branch below <p>.
  EFFECT:  Uses function <position_append1>.
**********************************************************/
{
  POSITION *local;

  local = position_copyPosition(p);

  position_append0(local);

  return local;
}

POSITION* position_get1Successor(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: A newly allocated position that is the
           1-successor of p, i.e. encodes the position of
           the right branch below <p>.
  EFFECT:  Uses function <position_append1>.
**********************************************************/
{
  POSITION *local;

  local = position_copyPosition(p);

  position_append1(local);

  return local;
}

void position_append0(POSITION* p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Sets <p> to its 0-sucessor, i.e. the position
           of the left branch below <p>.
**********************************************************/
{
  if (position_mustbeExpanded(p))
    position_expandField(p);

  position_incDepth(p);
}

void position_append1(POSITION* p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Sets <p> to its 1-sucessor, i.e. the position
           of the right branch below <p>.
**********************************************************/
{
  position_append0(p);

  position_set(p, position_getDepth(p)-1, 1);
}

int position_isRootPosition(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: 1 if <p> represents the empty (root) position,
           0 otherwise.
**********************************************************/
{
  return (position_getDepth(p) == 0);
}

void position_enqueue0(POSITION *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Shifts the bits of <p> by 1 to the left and
           sets the 0-th bit of <p> to 0.
           This is used if the backtrack stack is
           traversed from the top to the bottom.
**********************************************************/
{
  int i, lmsb, lmsb_new;

  if (position_isRootPosition(p))
    position_append0(p);
  else {
    position_append0(p);

    lmsb = 0;
    for(i=0; i<position_getFieldLength(p); ++i) {
      lmsb_new = position_getMostSignificantBit(p->field[i]);
      p->field[i] = p->field[i] << 1;
      position_setLeastSignificantBit(&(p->field[i]), lmsb);
      lmsb = lmsb_new;
    }
  }
}

void position_enqueue1(POSITION *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Shifts the bits of <p> by 1 to the left and
           sets the 0-th bit of <p> to 1.
           This is used if the backtrack stack is
           traversed from the top to the bottom.
**********************************************************/
{
  position_enqueue0(p);
  position_set(p, 0, 1);
}


/**************************************************************/
/* Orderings                                                  */
/**************************************************************/

int position_equal(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> and <p2> are equal, 0 otherwise.
  EFFECT:  Two position are equal if they represent the
           same branch.
**********************************************************/
{
  int i;

  if (position_getDepth(p1) != position_getDepth(p2))
    return 0;

  for (i=0; i<position_getFieldLength(p1); ++i)
    if (p1->field[i] != p2->field[i])
      return 0;

  return 1;
}

int position_PrefixOrderStrict(POSITION *p1, POSITION *p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <_p <p2>, 0 otherwise, i.e. whether
           <p1> is a strict prefix of <p2>.
**********************************************************/
{
  int i;

  if (position_isRootPosition(p1))
    return (!position_isRootPosition(p2));

  if (position_isRootPosition(p2))
    return 0;

  if (position_getDepth(p1) >= position_getDepth(p2))
    return 0;

  for(i=0; i<position_getDepth(p1); ++i)
    if (position_get(p1, i) != position_get(p2, i))
      return 0;

  return 1;
}

int position_PrefixOrderEqual(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <_p <p2>, 0 otherwise, i.e. whether
           <p1> is a prefix of <p2>.
**********************************************************/
{
  if (position_equal(p1, p2))
    return 1;
  else
    return (position_PrefixOrderStrict(p1, p2));
}

int position_BranchOrderStrict(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <=_b <p2>, 0 otherwise, i.e. whether
           <p1> is strictly smaller then <p2> wrt. the
           lexicographic order for bitstrings.
**********************************************************/
{
  int i, max;

  if (position_isRootPosition(p2))
    return 0;

  if (position_isRootPosition(p1))
    return 1;

  if (position_getDepth(p1) < position_getDepth(p2))
    max = position_getDepth(p1);
  else
    max = position_getDepth(p2);

  for(i=0; i<max; ++i)
    if ((position_get(p1, i) == 0) && (position_get(p2, i) == 1))
      return 1;
    else if ((position_get(p1, i) == 1) && (position_get(p2, i) == 0))
      return 0;

  return 0;
}

int position_BranchOrderEqual(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <_b <p2>, 0 otherwise, i.e. whether
           <p1> is smaller then <p2> wrt. the
           lexicographic order for bitstrings.
**********************************************************/
{
  if (position_equal(p1, p2))
    return 1;
  else
    return (position_BranchOrderStrict(p1, p2));
}

int position_TemporalVisitingOrderStrict(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <_t <p2>, 0 otherwise, i.e. whether
           <p2> is a strict temporal successor of <p1>.
**********************************************************/
{
  if (position_PrefixOrderStrict(p1, p2))
    return 1;
  else
    return (position_BranchOrderStrict(p1, p2));
}

int position_TemporalVisitingOrderEqual(POSITION* p1, POSITION* p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: 1 if <p1> <_t <p2>, 0 otherwise, i.e. whether
           <p2> is a temporal successor of <p1>.
**********************************************************/
{
  if (position_equal(p1, p2))
    return 1;
  else
    return (position_TemporalVisitingOrderStrict(p1, p2));
}


/**************************************************************/
/* Pretty printing and conversion                             */
/**************************************************************/

void position_print(POSITION* p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Prints <p> to STDOUT.
**********************************************************/
{
  char *temp;
  int i;

  temp = position_convertToString(p);

  printf("Position (depth: %d, field length: %d): '%s'; ",
	 position_getDepth(p), position_getFieldLength(p), temp);

  printf("Field: ");
  for(i=0; i<position_getFieldLength(p); ++i)
    printf("%ld ", p->field[i]);
  printf("\n");

  position_mmm_free(temp, strlen(temp)+1);
}

char* position_convertToString(POSITION* p)
/*********************************************************
  INPUT:   A position <p>.
  RETURNS: A newly allocated string representing <p>.
  EFFECT:  The root position is converted to "e", all
           other position by strings over '0' and '1'.
           Memory is allocated for the string.
**********************************************************/
{
  char *local;
  int i;

  if (position_isRootPosition(p)) {
    local = position_mmm_malloc(sizeof(char)*2);
    strcpy(local, "e");
    return local;
  }

  local = position_mmm_malloc(sizeof(char) * (1+position_getDepth(p)));

  for(i=0; i<position_getDepth(p); ++i)
    if (position_get(p,  i) == 0)
      local[i] = '0';
    else
      local[i] = '1';
  local[i] = '\0';

  return local;
}


/**************************************************************/
/* Generic evector functions                                  */
/**************************************************************/

int position_evector_tree_order(void *p1, void *p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: Whether <p2> is a strict temporal sucessor of
           <p1>.
  EFFECT:  Uses the function
           <position_TemporalVisitingOrderStrict>.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  return (position_TemporalVisitingOrderStrict((POSITION*) p1, (POSITION*) p2));
}

int position_evector_equal(void *p1, void *p2)
/*********************************************************
  INPUT:   Two positions.
  RETURNS: Whether <p1> is equal to <p2>.
  EFFECT:  Uses the function <position_equal>.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  return (position_equal((POSITION*) p1, (POSITION*) p2));
}

void position_evector_print(void *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Uses the function <position_print> to print
           <p> to STDOUT.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  position_print((POSITION*) p);
}

char* position_evector_convertToString(void *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Uses the function <position_convertToString>
           to convert <p> to a string.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  return position_convertToString((POSITION*) p);
}

void* position_evector_copyLiterally(void *p)
/*********************************************************
  INPUT:   A position.
  RETURNS: A deep copy of <p>.
  EFFECT:  Uses the function <position_copyPosition>
           to create a copy.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  return((POSITION*) position_copyPosition((POSITION*)p));
}

void position_evector_destroyPosition(void *p)
/*********************************************************
  INPUT:   A position.
  EFFECT:  Uses the function <position_destroyPosition>
           to free the memory occupied by <p>.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally.
**********************************************************/
{
  position_destroyPosition( (POSITION*) p);
}

void* position_evector_createFromString(char *encoding)
/*********************************************************
  INPUT:   A string encoding a position.
  RETURNS: A newly allocated position that is represented
           by <encoding>.
  EFFECT:  Uses the function <position_createFromString>
           to create a position from <encoding>.
  CAUTION: This function is intended to be used within
           the evector module, therefore it takes two
           void pointers which are recasted internally;
           also the return value is given as a void
           pointer.
**********************************************************/
{
  return ((void*) position_createFromString(encoding));
}

