/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                  SEARCH TREE POSITIONS                 * */
/* *                                                        * */
/* *  $Module:   POSITION                                   * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifndef _POSITION_
#define _POSITION_


/**************************************************************/
/* Macros                                                     */
/**************************************************************/

#define position_mmm_malloc(size) memory_Malloc(size)
#define position_mmm_free(ptr, size) memory_Free(ptr, size)

#define position_my_error(msg) { misc_StartErrorReport(); misc_ErrorReport("\n"msg"\n"); misc_FinishErrorReport(); }


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <limits.h>


/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

typedef unsigned long POSITION_FIELD_ENTRY;

typedef struct
{
  int depth;

  int field_length;
  POSITION_FIELD_ENTRY *field;
} POSITION;


/**************************************************************/
/* Position Access Functions                                  */
/**************************************************************/

void position_set(POSITION*, int, int);
int position_get(POSITION*, int);
void position_getIndex(int, int*, int*);

static __inline__ int position_getFieldLength(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: The length of the position field of <p>.
**********************************************************/
{
  return p->field_length;
}

static __inline__ int position_getDepth(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: The depth of <p>. Empty position have depth 0.
**********************************************************/
{
  return p->depth;
}

static __inline__ int position_incDepth(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: Increases the depth of <p> and returns the
           increased depth.
**********************************************************/
{
  return ++(p->depth);
}

static __inline__ int position_decDepth(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: Decreases the depth of <p> and returns the
           decreased depth.
**********************************************************/
{
  return --(p->depth);
}

static __inline__ int position_incFieldLength(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: Increases the field length of <p> and returns
           the increased length.
  CAUTION: The memory for the field is not extended, this
           has to be done using other functions.
**********************************************************/
{
  return ++(p->field_length);
}

static __inline__ int position_decFieldLength(POSITION* p)
/*********************************************************
  INPUT:   A position.
  RETURNS: Decreases the field length of <p> and returns
           the decreased length.
**********************************************************/
{
  return --(p->field_length);

}


/**************************************************************/
/* Creation and disallocation                                 */
/**************************************************************/

POSITION* position_createPosition(void);
POSITION* position_createRootPosition(void);
POSITION* position_createFromString(char*);
POSITION* position_copyPosition(POSITION*);
void position_destroyPosition(POSITION*);


/**************************************************************/
/* Specials                                                   */
/**************************************************************/

void position_expandField(POSITION*);
POSITION* position_getAbovePosition(POSITION*);
void position_above(POSITION*);
POSITION* position_get0Successor(POSITION*);
POSITION* position_get1Successor(POSITION*);
void position_append0(POSITION*);
void position_append1(POSITION*);
void position_enqueue0(POSITION*);
void position_enqueue1(POSITION*);
int position_isRootPosition(POSITION*);


/**************************************************************/
/* Orderings                                                  */
/*    Note that all the functions refer to the "<" order.     */
/**************************************************************/

int position_equal(POSITION*, POSITION*);
int position_PrefixOrderStrict(POSITION*, POSITION*);
int position_PrefixOrderEqual(POSITION*, POSITION*);
int position_BranchOrderStrict(POSITION*, POSITION*);
int position_BranchOrderEqual(POSITION*, POSITION*);
int position_TemporalVisitingOrderStrict(POSITION*, POSITION*);
int position_TemporalVisitingOrderEqual(POSITION*, POSITION*);


/**************************************************************/
/* Pretty printing and conversion                             */
/**************************************************************/

void position_print(POSITION*);
char* position_convertToString(POSITION*);


/**************************************************************/
/* Generic evector functions                                  */
/**************************************************************/

int position_evector_equal(void*, void*);
int position_evector_tree_order(void*, void*);
void position_evector_print(void*);
char* position_evector_convertToString(void*);
void* position_evector_copyLiterally(void*);
void position_evector_destroyPosition(void*);
void* position_evector_createFromString(char*);

#endif
