/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                CONNECTION TO CVC3                      * */
/* *                                                        * */
/* *  $Module: smt_cvc3                                     * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "smt_cvc3.h"
#include "search.h"
#include <errno.h>


/**************************************************************/
/* Module internal variables                                  */
/**************************************************************/

#define _SMTCVC3_BIGSTRING_SIZE 8192
#define _SMTCVC3_IDPREFIX "wq"
#define _SMTCVC3_IDNEGPREFIX "wz"
char smt_cvc3_big_string[_SMTCVC3_BIGSTRING_SIZE];
LIST smt_cvc3_keywords;
LIST smt_cvc3_error_return_codes;


/**************************************************************/
/* Internal Functions                                         */
/**************************************************************/

/*********************************************************
   The name of the auxilliary variable needed for proof
   documentation.
**********************************************************/
#define smt_cvc3_getNameOfProofDocVariable "__cno"


static int smt_cvc3_extractClauseNumberOutOfProofLine(char *line)
/*********************************************************
  INPUT:   A line of the CVC3 proof.
  RETURNS: If the proof contains a part like
           "(__cno = 144)" indicating that this formulae
           is the 144-th exported one. In this case, 144
           is returned, otherwise -1 is returned.
           More generally, the name of the auxilliary
           introduced function symbol is specified in the
           macro <smt_cvc3_getNameOfProofDocVariable>.
**********************************************************/
{
  char *substring;
  int number;

  substring=strstr(line, "("smt_cvc3_getNameOfProofDocVariable" = ");

  if (substring == NULL)
    return -1;

  if (sscanf(substring, "("smt_cvc3_getNameOfProofDocVariable" = %d)", &number) != 1)
    return -1;

  if (number == 0)
    /* this is our auxilliary constraint; it's not SPASS clause and thus not
       a formula we consider as part of our proof */
    return -1;

#ifdef CHECK
  if (number < 0) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_extractClauseNumberOutOfProofLine:");
    misc_ErrorReport("\n Got a negative clause number.\n");
    misc_FinishErrorReport();
  }
#endif

  return number;
}

SMT_INTERFACE* smt_cvc3_getInterface(void)
/*********************************************************
  RETURNS: The created instance of the interface for
           addressing the SMT functionalities from
           dpconnect.
  EFFECT:  Makes the internal functionality usable for
           dpconnect by establishing an instance of
           the datatype SMT_INTERFACE.
**********************************************************/
{
  SMT_INTERFACE *local;

  /* what are the capabilites of cvc3? */
  local = smt_interface_Create("CVC3",   /* name */
			       1,        /* unsatf. cores supp. */
			       1         /* non-gr. export supp. */
			       );

  /* provide pointers to the implementation of the functions that
     are required by the interface */
  local->smt_InitProver = &smt_cvc3_Init;
  local->smt_FreeProver = &smt_cvc3_Free;
  local->smt_clauseToString = &smt_cvc3_Cvc3ClauseToString;
  local->smt_getCommandLineArguments = &smt_cvc3_getCommandLineArguments;
  local->smt_parseSMTOutput = &smt_cvc3_parseSMTOutput;
  local->smt_addAssertion = &smt_cvc3_printAssertion;
  local->smt_printCheck = &smt_cvc3_printCheck;
  local->smt_printSPASSSignature = &smt_cvc3_printSPASSSignature;
  local->smt_printComment = &smt_cvc3_printComment;
  local->smt_prepareNonFilteredClauseInput = &smt_cvc3_prepareNonFilteredClauseInput;
  local->smt_SigToString = &smt_cvc3_Cvc3SigToString;

  return local;
}

void smt_cvc3_Init(void)
/*********************************************************
  EFFECT:  Initializes the module, in particular sets up
           the list of forbidden variable names for the
           SMT procedure.
**********************************************************/
{
  /* build the list of CVC3 keywords */
  smt_cvc3_keywords = list_Nil();
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("ASSERT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("QUERY"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("CHECKSAT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("WHERE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("COUNTEREXAMPLE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("COUNTERMODEL"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("CONTINUE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("RESTART"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("PUSH"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("POP"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("POPTO"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("TRANSFORM"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("PRINT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("OPTION"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("TRACE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("UNTRACE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("ECHO"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("INCLUDE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("REAL"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BITVECTOR"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("INT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("OF"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("ARRAY"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("DATATYPE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("END"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("LET"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("IN"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("LAMBDA"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("FORALL"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("IF"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("ELSEIF"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("ELSE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("SX"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("AND"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("OR"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("NOT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("XOR"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("NOR"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("NAND"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("NXOR"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVPLUS"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVMULT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVUMINUS"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVSUB"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVLT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVLE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVGT"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("BVGE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("FALSE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("TRUE"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("WITH"),
				 smt_cvc3_keywords);
  smt_cvc3_keywords = list_Push((POINTER)string_StringCopy("SUBTYPE"),
				 smt_cvc3_keywords);

  /* TO DO: add errorneous return codes later on */
  smt_cvc3_error_return_codes = list_Nil();

}

void smt_cvc3_Free(void)
/*********************************************************
  EFFECT:  Frees the module.
**********************************************************/
{
  LIST Scan;

  for(Scan = smt_cvc3_keywords; !list_Empty(Scan); Scan = list_Cdr(Scan))
    string_StringFree(list_Car(Scan));

  list_Delete(smt_cvc3_keywords);

  list_Delete(smt_cvc3_error_return_codes);
}

EVECTOR* smt_cvc3_getCommandLineArguments(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A proofsearch object.
  RETURNS: An expandable vector of command line arguments
           suitable for the current configuration as
           specified by <Search>.
  NOTE:    cf. "http://cs.nyu.edu/pipermail/cvc-users/2007/000006.html"
           (2007 July 5): program must be invoked with
           options "+proofs -sat fast"
**********************************************************/
{
  EVECTOR *local;
  FLAGSTORE store;
  char temp[30];

  /* general remark: CVC3's (CVCL's ?) command line processing seems to be quite
     weird: it takes options with additional arguments as two options: instead of
     >>"-timeout 2"<< one has to give >>"-timeout" "2"<< */

  /* abbreviation */
  store = prfs_Store(Search);

  local = evector_createEvectorString(15, 2);

  /* here: fine tuning in the future */
  evector_insertAppendedString(local, "-max-quant-inst");
  sprintf(temp, "%d", 200);
  evector_insertAppendedString(local, temp);

  evector_insertAppendedString(local, "-quant-max-inst");
  sprintf(temp, "%d", 600);
  evector_insertAppendedString(local, temp);

  evector_insertAppendedString(local, "-quant-naive-num");
  sprintf(temp, "%d", 100);
  evector_insertAppendedString(local, temp);

  /* if needed, turn generation of unsatisfiale cores on;
     cf. "http://cs.nyu.edu/pipermail/cvc-users/2007/000006.html" (2007 July 5)
     program must be invoked with options "+proofs -sat fast" */
  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCEVIDENCE) == flag_ON) {
    evector_insertAppendedString(local, "+proofs");
    evector_insertAppendedString(local, "-sat");
    evector_insertAppendedString(local, "fast");
  }

  /* name of input file as last argument */
  evector_insertAppendedString(local, (prfs_DpConnect(Search))->smt_input_filename);

  return local;
}

int smt_cvc3_parseSMTOutput(struct PROOFSEARCH_HELP* Search,
			    char *output_filename, EVECTOR** explanation,
			    EVECTOR** dep_positions, int smt_exitstatus)
/*********************************************************
  INPUT:   A proofsearch object, the filename where
           output was written to, two pointers to a
           expandable vectors used as return values, and
           an integer giving the exit status of the SMT
           procedure.
  RETURNS: 1: unsat, 2: sat, 3: unknown, 4: error or
           unrecognized; depending on the flag
           <DpcEvidence> <explanation> is filled with a
           a list of SPASS clauses that are the reason
           for the empty clause.
  EFFECT:  Analyses the output of the SMT procedure.
  CAUTION: *NEVER* call this routine directly, instead
           use <dpconnect_checkForExternalSMTResult>.
**********************************************************/
{
  FILE *output;
  int result, i, j, current_index, start_of_proof, current_no;
  CLAUSE_DICTIONARY *dictionary;
  EVECTOR *unsat_core;
#ifdef CHECK
  int max_explanation_id;
#endif

  /* initialization */
  *explanation = (EVECTOR*) NULL;
  *dep_positions = (EVECTOR*) NULL;

  /* abbreviation */
  dictionary = (prfs_DpConnect(Search))->clause_dictionary;

#ifdef CHECK
  /* check whether file with output exists */
  /* (prfs_DpConnect(Search))->smt_output_filename*/
  if (misc_ExistsFile(output_filename) == 0) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In dpconnect_parseSMTOutput:");
    misc_ErrorReport("\n No file containing SMT output existing.\n");
    misc_FinishErrorReport();
  }
#endif

  /* first of all, did we get a correct exit status? */
  if (list_PointerMember(smt_cvc3_error_return_codes, (POINTER) smt_exitstatus)) {
    /* cvc3 tends to crash for unknown reasons occasionally, so we return "unknown" instead of "error" */
    return 3;
    /* return 4; */
  }

  /* try to inspect output */
  output = fopen(output_filename, "r");
  if (output == NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_parseSMTOutput:");
    misc_ErrorReport("\n Could not open file for reading SMT output.\n");
    misc_FinishErrorReport();
  }

  result = -1;
  while (NULL != fgets(smt_cvc3_big_string, _SMTCVC3_BIGSTRING_SIZE, output)) {
    if (strstr(smt_cvc3_big_string, "Valid.\n") != NULL) {
      result = 1;
    } else if (strstr(smt_cvc3_big_string, "Invalid.\n") != NULL) {
      result = 2;
    } else if (strstr(smt_cvc3_big_string, "Unknown.\n") != NULL) {
      result = 3;
    } else if (strstr(smt_cvc3_big_string, "error\n") != NULL) {
      /* cvc3 tends to crash for unknown reasons occasionally, so we return "unknown" instead of "error" */
      result = 3;
      /* result = 4; */
    }

    if (result != -1)
      /* we got what we needed */
      break;
  }

  /* if proof documentation is wished and the result was "unsat" (for cvc "Valid")
     we try to extract the proof out of the output */
  if ((result == 1) &&
      (flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) == flag_ON)) {

    /* prepare vector for taking CVC3 ids of clauses required for the proof */
    unsat_core = evector_createEvectorInt(100, 2);

    start_of_proof = 0;
    while (NULL != fgets(smt_cvc3_big_string, _SMTCVC3_BIGSTRING_SIZE, output)) {
      if (start_of_proof == 0)
	if (strstr(smt_cvc3_big_string, "% Relevant assumptions:") != NULL)
	  start_of_proof = 1;

      if (start_of_proof == 1) {
	current_no = smt_cvc3_extractClauseNumberOutOfProofLine(smt_cvc3_big_string);
	if (current_no > 0)
	  /* we found a number */
	  evector_insertAppendedInt(unsat_core, current_no);
      }
    }
    /* unsat_core contains now all "CVC3 ids" */

#ifdef CHECK
    if (start_of_proof == 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_cvc3_parseSMTOutput:");
      misc_ErrorReport("\n An explanation was supposed, but we did not find one.\n");
      misc_FinishErrorReport();
    }

    if (evector_getNumberOfElements(unsat_core) == 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_cvc3_parseSMTOutput:");
      misc_ErrorReport("\n Got an empty explanation.\n");
      misc_FinishErrorReport();
    }
#endif

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("We got the following explanation from CVC3:\n");
    evector_printEVectorInt(unsat_core);
    printf("\n");
    smt_debug_end();
#endif

#ifdef CHECK
    max_explanation_id = evector_getMaxElementInt(unsat_core);
    if (max_explanation_id >
	clause_dictionary_getNumberOfElements(dictionary)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_cvc3_parseSMTOutput:");
      misc_ErrorReport("\n Inconsistent datastructures.\n");
      misc_FinishErrorReport();
    }
#endif

    /* prepare datastructure to take the explanation */
    *dep_positions =
      evector_createEvector(clause_dictionary_getNumberOfElements(dictionary),
			    2, &position_evector_equal, &position_evector_tree_order);
    *explanation =
      evector_createEvectorInt(10+clause_dictionary_getNumberOfElements(dictionary),
			       2);

    /* create explanation:
       (1) look in dictionary for corresponding SPASS clause numbers
       (2) look in dictionary for dependent positions and accumulate them */
    for (i=0; i<evector_getNumberOfElements(unsat_core); ++i) {
      current_index = evector_getNthInt(unsat_core, i)-1;

      /* (1) insert SPASS clause number */
      evector_insertAppendedInt(*explanation,
				clause_dictionary_getNthClauseNo(dictionary, current_index));

      /* (2) accumulate dependent positions */
      for(j=0;
	  j<evector_getNumberOfElements(clause_dictionary_getNthDependentPositions(dictionary,
										   current_index));
	  ++j) {
	if (!evector_isMember(*dep_positions,
			      (POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)))
	  evector_insertAppended(*dep_positions,
				 position_copyPosition((POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)));
      }
    }
    /* free memory for temporary allocated evector */
    evector_destroyEvectorInt(unsat_core);
#ifdef CHECK
    if (evector_getNumberOfElements(*explanation) == 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_cvc3_parseSMTOutput:");
      misc_ErrorReport("\n Got an empty explanation.\n");
      misc_FinishErrorReport();
    }
#endif

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("This yields the following SPASS explanation:\n");
    evector_printEVectorInt(*explanation);
    printf("With the following positions the empty clause depends on:\n");
    evector_printEVector(*dep_positions, &position_evector_print);
    printf("\n");
    smt_debug_end();
#endif

  } /* we got "unsat" and were seeking for an explanation in the output */

  /* close file containing the input */
  fclose(output);
  /* relax, ignore any errors */
  errno = 0;

  if (result == -1) {
    /* read all output, but could not match any of the patterns */
    /* cvc3 tends to crash for unknown reasons occasionally, so we return "unknown" instead of "error" */
    result = 3;
    /* result = 4; */
  }

#ifdef SMT_DEBUG
  smt_debug_start();
  printf("In smt_cvc3_parseSMTOutput: Got the following result from cvc3:\n%s\n",
	 smt_cvc3_big_string);
  smt_debug_end();
#endif

  return result;
}

void smt_cvc3_printAssertion(struct PROOFSEARCH_HELP* Search, FILE *export_file, char* formula, int spass_id, int position)
/*********************************************************
  INPUT:   A proofsearch object, a pointer to a file
           stream, a string representing the formula, the
           SPASS clause id and the number indicating how
           many clauses the SMT procedure has seen so
           far.
  EFFECT:  Appends a cvc instruction to the current
           cvc input (in file <export_file>)
           which asserts the formula given by
           the string <formula>. Depending on the value of
           <DpcEvidence> possibly the auxilliary function
           symbol for proof documentation is introduced
           using the <position> information.
  CAUTION: The file has to be opened before.
**********************************************************/
{
#ifdef CHECK
  if (export_file == (FILE*) NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_printAssertion:");
    misc_ErrorReport("\n Error providing cvc3 output.\n");
    misc_FinishErrorReport();
  }
#endif

  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCEVIDENCE) == flag_ON)
    fprintf(export_file, "ASSERT(%s=%d OR %s); %s SPASS clause ID is %d\n",
	    smt_cvc3_getNameOfProofDocVariable, position, formula, "%%", spass_id);
  else
    fprintf(export_file, "ASSERT(%s); %s SPASS clause ID is %d\n", formula, "%%", spass_id);
}

void smt_cvc3_printCheck(struct PROOFSEARCH_HELP* Search, FILE *export_file)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream.
  EFFECT:  Appends a cvc3 instruction to the current
           cvc3 input (in file <export_file>)
           which asks the DP for a check.
**********************************************************/
{
#ifdef CHECK
  if (export_file ==  NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_printAssertion:");
    misc_ErrorReport("\n Error providing cvc3 output.\n");
    misc_FinishErrorReport();
  }
#endif

  fprintf(export_file, "\n\n%s check the context\nQUERY(FALSE);\n", "%%");
  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCEVIDENCE) == flag_ON)
    fprintf(export_file, "\n\n%s ask for an unsatisfiable core\nDUMP_ASSUMPTIONS;\n", "%%");
}

void smt_cvc3_printComment(struct PROOFSEARCH_HELP* Search, FILE *export_file, char *comment)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream and a string.
  EFFECT:  Appends a CVC3 comment to the current CVC3
           input (in file <export_file>).
**********************************************************/
{
  fprintf(export_file, "%s %s\n", "%%", comment);
}

void smt_cvc3_prepareNonFilteredClauseInput(struct PROOFSEARCH_HELP *Search, FILE *f)
/*********************************************************
  INPUT:   A proofsearch object and a pointer to a file
           stream.
  EFFECT:  Inserts some default stuff to the non-filtered
           input of the SMT procedure.
**********************************************************/
{
  /* we insert an auxilliary equation needed for proof documentation */
  fprintf(f, "%s This is an auxialliary assertion needed for proof documentation\n",
	  "%%");
  fprintf(f, "ASSERT (%s = 0);\n\n", smt_cvc3_getNameOfProofDocVariable);

}

void smt_cvc3_printSPASSSignature(struct PROOFSEARCH_HELP* Search, FILE* export_file)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream.
  EFFECT:  Appends a comment to the cvc3 input (in file
           <export_file>).
**********************************************************/
{
#ifdef CHECK
  if (export_file ==  NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_printAssertion:");
    misc_ErrorReport("\n Error providing cvc3 output.\n");
    misc_FinishErrorReport();
  }
#endif

  fprintf(export_file, "%s", "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
  fprintf(export_file, "%s This CVC3 input was automatically created  %s\n", "%%", "%%");
  fprintf(export_file, "%s by SPASS+T.                                %s\n", "%%", "%%");
  fprintf(export_file, "%s", "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n");
}


/**************************************************************/
/* Conversion Functions                                       */
/**************************************************************/

BOOL smt_cvc3_VariableIDsEqual(POINTER p1, POINTER p2)
{
  return (*((NAT*) p1) == *((NAT*) p2));
}

ESTRING* smt_cvc3_Cvc3QuantifierPrefixToString(CLAUSE Clause, char *type)
/**************************************************************
  INPUT:   A clause, the type of all variables.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts the quantifier prefix into string, using
           <type> for all variables.
***************************************************************/
{
  ESTRING *estring;
  char convert[50];
  LIST vars, iter;

  estring = estring_EmptyString(200);
  estring_AppendString(estring, "FORALL ( ");

  vars =  clause_ListOfVariables(Clause);
  vars = list_DeleteDuplicates(vars, &smt_cvc3_VariableIDsEqual);

  for (iter=vars; !list_Empty(iter); iter=list_Cdr(iter)) {
    strcpy(convert, "");
    sprintf(convert, "%d", *((NAT *)list_Car(iter)));

    estring_AppendString(estring, " U_");
    estring_AppendString(estring, convert);
    estring_AppendString(estring, " : ");
    if ((strcmp(type, "int") == 0) || (strcmp(type, "INT") == 0)) {
      estring_AppendString(estring, "INT");
    } else if ((strcmp(type, "real") == 0) || (strcmp(type, "REAL") == 0)) {
      estring_AppendString(estring, "REAL");
    } else {
      estring_AppendString(estring, type);
    }
    if (!list_Empty(list_Cdr(iter)))
      estring_AppendString(estring, ", ");
  }

  estring_AppendString(estring, " ): ");

  list_Delete(vars);

  return estring;
}

ESTRING* smt_cvc3_Cvc3ClauseToString(CLAUSE Clause, char *type)
/**************************************************************
  INPUT:   A clause, a type.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Clause> into the format of the external
           SMT procedure; variables are assumed to be of type
           <type>.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING *estring, *qprefix_estring;

  int     n,j;
  LITERAL Lit;
  TERM    Atom;
  BOOL    numericTheory;

  numericTheory = (strcmp(type, "data") != 0);

  n = clause_Length(Clause);

  estring = estring_EmptyString(150);

  /* > SZ < */

  estring_AppendString(estring, "(");
  /* do we have to add a quantifier prefix? */
  if (clause_ContainsVariables(Clause)) {
    qprefix_estring = smt_cvc3_Cvc3QuantifierPrefixToString(Clause, type);
    estring_AppendEStringDelete(estring, qprefix_estring);
  }

  if (n == 0) {
    estring_AppendString(estring, "FALSE");
  } else {
    for (j = 0; j < n; j++) {
      Lit  = clause_GetLiteral(Clause,j);
      Atom = clause_LiteralAtom(Lit);
      if (clause_LiteralIsNegative(Lit)) {
        estring_AppendString(estring, "NOT(");
        estring_AppendEStringDelete(estring, smt_cvc3_Cvc3AtomToString(Atom, numericTheory));
        estring_AppendString(estring, ")");
      } else {
	estring_AppendEStringDelete(estring, smt_cvc3_Cvc3AtomToString(Atom, numericTheory));
      }
      if (j <= (n-2)) {
	estring_AppendString(estring, " OR ");
      }
    }
  }
  estring_AppendString(estring, ")");

  return estring;
}


ESTRING* smt_cvc3_Cvc3AtomToString(TERM Atom, BOOL numericTheory)
/**************************************************************
  INPUT:   An atom and a boolean indicating whether some numeric theory is in effect.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Atom> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  SYMBOL TopSymbol;
  ESTRING *estring;

  estring = estring_EmptyString(150);

  TopSymbol = term_TopSymbol(Atom);

  if (TopSymbol == fol_True())
    estring_AppendString(estring, "TRUE");
  else if (TopSymbol == fol_False())
    estring_AppendString(estring, "FALSE");
  else if (term_IsComplex(Atom)) {
    if (TopSymbol == fol_Equality()) {
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Atom), numericTheory));
      estring_AppendString(estring, " = ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Atom), numericTheory));
    } else if (! numericTheory) {
      estring_AppendString(estring, _SMTCVC3_IDPREFIX);
      estring_AppendStringDelete(estring, symbol_FPrintToString(TopSymbol));
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermListToString(term_ArgumentList(Atom), numericTheory));
      estring_AppendString(estring, ")");
    } else if (TopSymbol == fol_ThLess()) {
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Atom), numericTheory));
      estring_AppendString(estring, " < ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Atom), numericTheory));
    } else if (TopSymbol == fol_ThLesseq()) {
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Atom), numericTheory));
      estring_AppendString(estring, " <= ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Atom), numericTheory));
    } else if (TopSymbol == fol_ThGreater()) {
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Atom), numericTheory));
      estring_AppendString(estring, " > ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Atom), numericTheory));
    } else if (TopSymbol == fol_ThGreatereq()) {
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Atom), numericTheory));
      estring_AppendString(estring, " >= ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Atom), numericTheory));
    } else {
      estring_AppendString(estring, _SMTCVC3_IDPREFIX);
      estring_AppendStringDelete(estring, symbol_FPrintToString(TopSymbol));
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermListToString(term_ArgumentList(Atom), numericTheory));
      estring_AppendString(estring, ")");
    }
  } else {
    /* term_IsConstant(Atom) */
    estring_AppendString(estring, _SMTCVC3_IDPREFIX);
    estring_AppendStringDelete(estring, symbol_FPrintToString(TopSymbol));
  }

  return estring;
}

ESTRING* smt_cvc3_Cvc3TermToString(TERM Term, BOOL numericTheory)
/**************************************************************
  INPUT:   A term and a boolean indicating whether some numeric theory is in effect.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Term> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING *estring;
  SYMBOL TopSymbol;
  char convert[20];

  estring = estring_EmptyString(150);

  TopSymbol = term_TopSymbol(Term);
  if (term_IsComplex(Term)) {
    if (! numericTheory) {
      estring_AppendString(estring, _SMTCVC3_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring,
				  smt_cvc3_Cvc3TermListToString(term_ArgumentList(Term), numericTheory));
      estring_AppendString(estring, ")");
    } else if (TopSymbol == fol_ThSucc()) {
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " + 1)");
    } else if (TopSymbol == fol_ThPred()) {
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " - 1)");
    } else if (TopSymbol == fol_ThPlus()) {
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " + ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Term), numericTheory));
      estring_AppendString(estring, ")");
    } else if (TopSymbol == fol_ThMinus()) {
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " - ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Term), numericTheory));
      estring_AppendString(estring, ")");
    } else if (TopSymbol == fol_ThUminus()) {
      estring_AppendString(estring, "(0 - ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, ")");
    } else if (TopSymbol == fol_ThTimes()
               && (term_IsNumeric(term_FirstArgument(Term))
                   || term_IsNumeric(term_SecondArgument(Term)))) {
      TERM t1, t2;
      if (term_IsNumeric(term_FirstArgument(Term))) {
        t1 = term_FirstArgument(Term);
        t2 = term_SecondArgument(Term);
      } else {
        t1 = term_SecondArgument(Term);
        t2 = term_FirstArgument(Term);
      }
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(t1, numericTheory));
      estring_AppendString(estring, " * ");
      estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(t2, numericTheory));
      estring_AppendString(estring, ")");
/*     } else if (((TopSymbol == fol_ThDivide())/\*  || (TopSymbol == fol_ThDivideNat()) *\/) && */
/* 	       (term_IsNumeric(term_FirstArgument(Term)) &&  */
/* 		term_IsNumeric(term_SecondArgument(Term)))) { */
/*       /\* we directly export a term of the form "divide\(_nat\)(3,4)" (in any case this is  */
/* 	 done later after arithmetic simplification) *\/ */
/*       estring_AppendString(estring, "("); */
/*       estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_FirstArgument(Term), numericTheory)); */
/*       estring_AppendString(estring, " / "); */
/*       estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(term_SecondArgument(Term), numericTheory)); */
/*       estring_AppendString(estring, ")"); */
    } else {
      estring_AppendString(estring, _SMTCVC3_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
      estring_AppendString(estring, "(");
      estring_AppendEStringDelete(estring,
				  smt_cvc3_Cvc3TermListToString(term_ArgumentList(Term), numericTheory));
      estring_AppendString(estring, ")");
    }

  } else if (term_IsVariable(Term)) {
    sprintf(convert, "U_%d", *((NAT*) Term));
    estring_AppendString(estring, convert);
  } else {

    /* term_IsConstant(Term) */
    if (numericTheory &&
        (symbol_denotesInt(term_TopSymbol(Term)) ||
	 symbol_denotesRat(term_TopSymbol(Term)))) {
      estring_AppendEStringDelete(estring,
				  smt_cvc3_Cvc3NumericSymbolToString(term_TopSymbol(Term)));
    } else {
      estring_AppendString(estring, _SMTCVC3_IDPREFIX);
      estring_AppendStringDelete(estring, symbol_FPrintToString(term_TopSymbol(Term)));
    }
  }


  return estring;
}


ESTRING* smt_cvc3_Cvc3TermListToString(LIST List, BOOL numericTheory)
/**************************************************************
  INPUT:   A list of terms and a boolean indicating whether some numeric theory is in effect.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <List> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING *estring;

  estring = estring_EmptyString(150);

  for (; !list_Empty(List); List=list_Cdr(List)) {
    estring_AppendEStringDelete(estring, smt_cvc3_Cvc3TermToString(list_Car(List), numericTheory));
    if (!list_Empty(list_Cdr(List)))
      estring_AppendString(estring, ",");
  }

  return estring;
}

void     smt_cvc3_checkIdentifier(char *Name)
/**************************************************************
  INPUT:   A name intended to be used for an identifier.
  EFFECT:  Nothing if <Name> can be used for the name of a
           signature symbol, otherwise, if it is a reserved
           keyword for CVC3, an error is reported and the
           program terminates.
  CAUTION: Function "smt_cvc3_Init" must have been called
           before.
***************************************************************/
{
  LIST Scan;

  for(Scan = smt_cvc3_keywords; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (string_Equal((char*) list_Car(Scan), Name)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Error, the name \"%s\" is a reserved %s keyword.\n",
			   Name, "CVC3");
      misc_FinishUserErrorReport();
    }
  }
}

ESTRING* smt_cvc3_Cvc3SigToString(const char* theory)
/**************************************************************
  INPUT:   A theory type specification.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts the signature to the format of the
           external SMT procedure. Uses <Type> as base type
           for free function symbols and predicates.
***************************************************************/
{
  ESTRING *estring;
  ESTRING *auxstring;
  char Type[50];
  BOOL numericTheory;

  /* set type appropriately */
  if (strcmp(theory, "data") == 0) {
    strcpy(Type, "data");
    numericTheory = FALSE;
  } else if (strcmp(theory, "int") == 0) {
    strcpy(Type, "INT");
    numericTheory = TRUE;
  } else if (strcmp(theory, "real") == 0) {
    strcpy(Type, "REAL");
    numericTheory = TRUE;
  }
#ifdef CHECK
  else {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_Cvc3SigToString:");
    misc_ErrorReport("\n Wrong theory type specification given.\n");
    misc_FinishErrorReport();
  }
#endif

  estring = estring_EmptyString(150);

  estring_AppendString(estring, "%% signature \n\n");

  if (! numericTheory) {
    estring_AppendString(estring, "data: TYPE;\n\n");
  }

  /* add signature of a designated variable for easy extraction
     of unsatisfiable cores */
  estring_AppendString(estring, "%% auxilliary function symbol for proof documentation\n");
  estring_AppendString(estring, smt_cvc3_getNameOfProofDocVariable);
  estring_AppendString(estring, ": INT;\n\n");

  if (symbol_SignatureExists()) {
    int          i, Index;
    SYMBOL       Info;
    SIGNATURE    Entry;


    for (Index = 1 ; Index < symbol_ACTINDEX ; Index++) {
      Entry  = symbol_Signature(Index);
      if (Entry != NULL) {
	Info = Entry->info;

	switch (symbol_Type(Info)) {
	case symbol_CONSTANT:
          if (symbol_denotesInt(Info) || symbol_denotesRat(Info)) {
            if (! numericTheory) {
              auxstring = smt_cvc3_Cvc3NumericSymbolToString(Info);
              if ((estring_GetString(auxstring))[0] == '-') {
                estring_AppendString(estring, _SMTCVC3_IDNEGPREFIX);
                estring_AppendString(estring, estring_GetString(auxstring)+1);
	      } else {
                estring_AppendString(estring, _SMTCVC3_IDPREFIX);
                estring_AppendString(estring, estring_GetString(auxstring));
	      }
              estring_StringFree(auxstring);
              estring_AppendString(estring, ": ");
	      estring_AppendString(estring, Type);
	      estring_AppendString(estring, ";\n");
	    } else {
              ;
	    }
	  } else {
            estring_AppendString(estring, _SMTCVC3_IDPREFIX);
	    estring_AppendString(estring, Entry->name);
	    estring_AppendString(estring, ": ");
	    estring_AppendString(estring, Type);
	    estring_AppendString(estring, ";\n");
	  }
	  break;
	case symbol_FUNCTION:
          if (numericTheory && (
                 symbol_Equal(Info, fol_ThSucc())
              || symbol_Equal(Info, fol_ThPred())
              || symbol_Equal(Info, fol_ThPlus())
              || symbol_Equal(Info, fol_ThMinus())
              || symbol_Equal(Info, fol_ThUminus())
              /* we declare "times" in addition to the built-in "*",
                 so that we can use it in non-linear contexts */
              /* || symbol_Equal(Info, fol_ThTimes()) */
	      /* we declare "divide" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
              /* || symbol_Equal(Info, fol_ThDivide()) */
              /* we declare "divide_nat" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
	      /* || symbol_Equal(Info, fol_ThDivideNat()) */
              /* || symbol_Equal(Info, fol_ThRemainder()))  */
	      )) {
            continue;
	  }
          estring_AppendString(estring, _SMTCVC3_IDPREFIX);
          estring_AppendString(estring, Entry->name);
          estring_AppendString(estring, ": ");
          if (Entry->arity == 1) {
	    estring_AppendString(estring, Type);
	    estring_AppendString(estring, " -> ");
	  } else {
	    estring_AppendString(estring, "(");
            for (i = Entry->arity; i > 0; i--) {
              estring_AppendString(estring, Type);
	      estring_AppendString(estring, i == 1 ? ") -> " : ", ");
	    }
	  }
	  estring_AppendString(estring, Type);
	  estring_AppendString(estring, ";\n");
	  break;
        case symbol_PREDICATE:
          if (symbol_Equal(Info, fol_Equality())) {
            continue;
	  }
          if (numericTheory && (
                 symbol_Equal(Info, fol_ThLess())
              || symbol_Equal(Info, fol_ThLesseq())
              || symbol_Equal(Info, fol_ThGreater())
              || symbol_Equal(Info, fol_ThGreatereq())
              || symbol_Equal(Info, fol_ThGround())
              || symbol_Equal(Info, fol_True())
              || symbol_Equal(Info, fol_False()))) {
            continue;
	  }
          estring_AppendString(estring, _SMTCVC3_IDPREFIX);
          estring_AppendString(estring, Entry->name);
	  estring_AppendString(estring, ": ");
          if (Entry->arity == 0) {
            ;
	  } else if (Entry->arity == 1) {
	    estring_AppendString(estring, Type);
	    estring_AppendString(estring, " -> ");
	  } else {
	    estring_AppendString(estring, "(");
            for (i = Entry->arity; i > 0; i--) {
	      estring_AppendString(estring, Type);
	      estring_AppendString(estring, i == 1 ? ") -> " : ", ");
	    }
	  }
	  estring_AppendString(estring, "BOOLEAN;\n");
          break;
	case symbol_JUNCTOR:
	  ;
	}
      }
    }
  }

  return estring;
}

ESTRING* smt_cvc3_Cvc3NumericSymbolToString(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: The string representation of the numeric symbol
           <Symbol>.
  SUMMARY: If symbol denotes an integer number, just <Symbol>,
           if it encodes a rational number <n/m> then <(n/m)>,
           otherwise an error.
  CAUTION: The function only accepts arithmetic symbols as
           input.
***************************************************************/
{
  ESTRING *result;
  MPN *r;

  result = estring_EmptyString(10);

  r = symbol_getValue(Symbol);

  if (symbol_denotesInt(Symbol)) {
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_numref(*r)));
  } else if (symbol_denotesRat(Symbol)) {
    estring_AppendString(result, "(");
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_numref(*r)));
    estring_AppendString(result, " / ");
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_denref(*r)));
    estring_AppendString(result, ")");
  } else {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_cvc3_Cvc3NumericSymbolToString:");
    misc_ErrorReport("\n Argument is not a numeric symbol.\n");
    misc_FinishErrorReport();
  }

  return result;
}

ESTRING* smt_cvc3_Cvc3FormulaToString(TERM Formula, char *mtype, BOOL numericTheory)
/**************************************************************
  INPUT:   A formula (as TERM), a type, and a boolean indicating whether some numeric theory is in effect
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Formula> into the format of the external
           SMT procedure; variables are assumed to be of type
           <type>.
           Uses the other converting functions.
  CAUTION: <Formula> can be an arbitrary formula (probably
           most of the times not in CNF, since FLOTTER was not
           invoked so far).
***************************************************************/
{
  SYMBOL Top;
  ESTRING *result;
  char convert[20];
  char type[20];

  if ((strcmp(mtype, "real")==0) || (strcmp(mtype, "REAL")==0)) {
    strcpy(type, "REAL");
  } else if ((strcmp(mtype, "int")==0) || (strcmp(mtype, "INT")==0)) {
    strcpy(type, "INT");
  } else {
    strcpy(type, mtype);
  }

  result = estring_EmptyString(100);

  Top = term_TopSymbol(Formula);

  if (symbol_Equal(Top, fol_And()) || symbol_Equal(Top, fol_Or()) ||
      symbol_Equal(Top, fol_Equiv()) || symbol_Equal(Top, fol_Implies()) ||
      symbol_Equal(Top, fol_Implied())) {

    if (list_Length(term_ArgumentList(Formula)) == 1) {
      /* formulae as "or(P(x))" are accepted by DFG parser */
      estring_AppendEStringDelete(result,
				  smt_cvc3_Cvc3FormulaToString(term_FirstArgument(Formula),
							       type, numericTheory));
    } else {
      if (symbol_Equal(Top, fol_Implied()))
	estring_AppendEStringDelete(result,
				    smt_cvc3_Cvc3FormulaToString(term_SecondArgument(Formula),
								 type, numericTheory));
      else
	estring_AppendEStringDelete(result,
				    smt_cvc3_Cvc3FormulaToString(term_FirstArgument(Formula),
								 type, numericTheory));

      if (symbol_Equal(Top, fol_And()))
	estring_AppendString(result, " AND ");
      else if (symbol_Equal(Top, fol_Or()))
	estring_AppendString(result, " OR ");
      else if (symbol_Equal(Top, fol_Equiv()))
	estring_AppendString(result, " <=> ");
      else /* (symbol_Equal(Top, fol_Implies())) || (symbol_Equal(Top, fol_Implied())) */
	estring_AppendString(result, " => ");

      if (symbol_Equal(Top, fol_Implied()))
	estring_AppendEStringDelete(result,
				    smt_cvc3_Cvc3FormulaToString(term_FirstArgument(Formula), type, numericTheory));
      else
	estring_AppendEStringDelete(result,
				    smt_cvc3_Cvc3FormulaToString(term_SecondArgument(Formula), type, numericTheory));
    }
  } else if (symbol_Equal(Top, fol_Not())) {
    estring_AppendString(result, "NOT( ");

/*     for (Scan=term_ArgumentList(Formula);!list_Empty(Scan);Scan=list_Cdr(Scan)) { */
    estring_AppendEStringDelete(result,
				smt_cvc3_Cvc3FormulaToString(term_FirstArgument(Formula),
							       type, numericTheory));
    estring_AppendString(result, ")");
  } else if (symbol_Equal(Top, fol_True())) {
    estring_AppendString(result, "TRUE");
  } else if (symbol_Equal(Top, fol_False())) {
    estring_AppendString(result, "FALSE");
  } else if (fol_IsQuantifier(Top)) {
    LIST Scan;

    if (symbol_Equal(Top, fol_All()))
      estring_AppendString(result, "FORALL (");
    else
      estring_AppendString(result, "EXISTS (");
    for (Scan=fol_QuantifierVariables(Formula); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      if (term_IsVariable((TERM) list_Car(Scan))) {
	sprintf(convert, "U_%d", *((NAT*) list_Car(Scan)));
	estring_AppendString(result, convert);
	estring_AppendString(result, ": ");
	estring_AppendString(result, type);

	if (!list_Empty(list_Cdr(Scan)))
	  estring_AppendString(result, ", ");
	else
	  estring_AppendString(result, " ");
      }
#ifdef CHECK
      else {
	misc_StartErrorReport();
	misc_ErrorReport("\n In smt_cvc3_Cvc3FormulaToString:");
	misc_ErrorReport("\n Non-variable argument under quantifier.\n");
	misc_FinishErrorReport();
      }
#endif
    }
    estring_AppendString(result, "): ");
    estring_AppendEStringDelete(result,
				smt_cvc3_Cvc3FormulaToString((TERM) term_SecondArgument(Formula), type, numericTheory));
  } else {
    estring_AppendEStringDelete(result, smt_cvc3_Cvc3AtomToString(Formula, numericTheory));
  }

  return result;
}

void smt_cvc3_FPrintCnfCvc3(FILE* File, LIST Clauses, char *type, int evidence)
/**************************************************************
  INPUT:   A file, a list of clauses and a flag store, and a
           flag whether unsatisfiable cores are wanted.
  RETURNS: Nothing.
  SUMMARY: Prints the clauses to <File> in a format readable by Otter.
***************************************************************/
{
  LIST   scan;
  ESTRING *result;

  result = estring_EmptyString(10);

/*   fol_FPrintOtterOptions(File, Equality, */
/* 			 flag_GetFlagValue(Flags, flag_TDFG2OTTEROPTIONS)); */

  if (Clauses)
    for (scan=Clauses;!list_Empty(scan);scan=list_Cdr(scan)) {
      estring_AppendString(result, "ASSERT ");
      estring_AppendEStringDelete(result,
				  smt_cvc3_Cvc3ClauseToString(list_Car(scan),
								type));
      estring_AppendString(result, ";\n");
    }

  fprintf(File, "%s", estring_GetString(result));

  estring_StringFree(result);
}

void smt_cvc3_FPrintCvc3(FILE* File, LIST Formulae, FLAG_TDFG2OTTEROPTIONSTYPE Option,
			 char *type, int evidence)
/**************************************************************
  INPUT:   A file, a list of pairs (label.formula), an option
           flag, and a flag whether unsatisfiable cores are
           wanted.
  RETURNS: Nothing.
  SUMMARY: Prints a  the respective formulae in CVC3 format to <File>.
***************************************************************/
{
  LIST   Scan;
  ESTRING *result;
  BOOL   numericTheory;

  /* fol_FPrintOtterOptions(File, Equality, Option); */

  numericTheory = (strcmp(type, "data") != 0);

  result = estring_EmptyString(10);

  if (!list_Empty(Formulae)) {
    for (Scan=Formulae;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      estring_AppendString(result, "ASSERT ");
      estring_AppendEStringDelete(result,
				  smt_cvc3_Cvc3FormulaToString(list_PairSecond(list_Car(Scan)),
								 type, numericTheory));
      estring_AppendString(result, ";\n");
    }
  }

  fprintf(File, "%s", estring_GetString(result));
  fprintf(File, "\n");

  estring_StringFree(result);
}
