/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          INTERFACE TO EXTERNAL SMT PROCEDURES          * */
/* *                                                        * */
/* *  $Module:   smt_interface                              * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "smt_interface.h"
#include "dpconnect.h"
#include "search.h"

/* insert here all headers for libraries for new solvers */
#include "smt_yices.h"
#include "smt_cvc3.h"


/**************************************************************/
/* Initialiazation                                            */
/**************************************************************/

SMT_INTERFACE* smt_interface_Init(struct PROOFSEARCH_HELP *Search)
/*********************************************************
  INPUT:   A Proofsearch object.
  EFFECT:  Delegates the work to the corresponding <Init>
           procedure of the SMT procedure that is chosen
           by flag <DpcProver>. Afterward the
           initialization function of the prover is
           invoked automatically. Thus, the SMT_INTERFACE
           structure is readily attached and can be used.
**********************************************************/
{
  SMT_INTERFACE *local;

  /* insert here stuff for libraries for new solvers */

  /* take the right interface form the specialized /smt_.+/ library */
  switch(flag_GetFlagValue(prfs_Store(Search), flag_DPCPROVER)) {
  case flag_DPCYICES:
    local = smt_yices_getInterface();
    break;
  case flag_DPCCVC3:
    local = smt_cvc3_getInterface();
    break;
  default:
    misc_StartUserErrorReport();
    misc_UserErrorReport("In smt_interface_Init:\n");
    misc_UserErrorReport("Unsupported SMT procedure!\n");
    misc_FinishUserErrorReport();
  }

  smt_interface_InitProver(local);

  return local;
}


/**************************************************************/
/* Allocation and deconstruction                              */
/**************************************************************/

SMT_INTERFACE* smt_interface_Create(const char *name, int unsatisfiable_cores,
				    int nonground_export)
/*********************************************************
  INPUT:   The name of the SMT procedure, two integers
           whether it supports unsatisfiable cores and
           quantifier reasoning respectively.
  EFFECT:  Allocated memory for an SMT_INTERFACE
           structure and fills it with the given data.
**********************************************************/
{
  SMT_INTERFACE *temp;

  temp = (SMT_INTERFACE*) memory_Malloc(sizeof(SMT_INTERFACE));

  temp->name = memory_Malloc(sizeof(char)*(1+strlen(name)));
  strcpy(temp->name, name);

  temp->unsatisfiable_cores = unsatisfiable_cores;
  temp->nonground_export = nonground_export;

  return temp;
}

void smt_interface_Disalloc(SMT_INTERFACE* smt_interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  EFFECT:  Free the memory that was occupied by the
           structure.
**********************************************************/
{
  memory_Free(smt_interface->name, sizeof(char)*(strlen(smt_interface->name)+1));
  memory_Free(smt_interface, sizeof(SMT_INTERFACE));

  return;
}
