/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          INTERFACE TO EXTERNAL SMT PROCEDURES          * */
/* *                                                        * */
/* *  $Module:   smt_interface                              * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/*                                                            */
/*  How to plug in a new decision procedure into SPASS+T?     */
/*                                                            */
/*                                                            */
/* Say you want to plug in a new decision procedured named by */
/* "my_procedure". Then you have to do the following:         */
/*                                                            */
/*    1) Provide a library "smt_my_procedure" that implements */
/*       the functions that are specified in SMT_INTERFACE    */
/*       structure; for any comments see the descriptions of  */
/*       the corresponding wrapper functions below            */
/*    2) Implement a function "smt_my_procedure_getInterface" */
/*       which returns a SMT_INTERFACE object by calling      */
/*         smt_interface_Create("my_procedure",               */
/*            (my_procedure supports unsatf. cores ? 1 : 0),  */
/*            (my_procedure supports quantifiers ? 1 : 0))    */
/*       and afterward setting the function points            */
/*       appropriately to the implementations                 */
/*    3) In flags.h: extend FLAG_DPCPROVERTYPE by a value     */
/*       flag_DPCMY_PROCEDURE                                 */
/*    4) Extend the function "smt_interface_Init" in the      */
/*       obvious way                                          */
/*                                                            */
/* For any details more details see the modules "smt_yices"   */
/* and "smt_cvc3".                                            */
/*                                                            */
/**************************************************************/


#ifndef _SMTINTERFACE_
#define _SMTINTERFACE_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "estrings.h"
#include "evector.h"
#include "clause.h"


/**************************************************************/
/* Datastructures                                             */
/**************************************************************/

struct PROOFSEARCH_HELP;

typedef struct SMT_INTERFACE_HELP {
  /* name of the dp */
  char *name;
  /* dp provides unsatisfiable cores */
  int unsatisfiable_cores;
  /* dp supports non-ground export */
  int nonground_export;

  /* functions that have to be implemented by a module for a certain dp;
     see below for comments */
  void (*smt_InitProver)(void);
  void (*smt_FreeProver)(void);
  ESTRING* (*smt_clauseToString)(CLAUSE, char*);
  EVECTOR* (*smt_getCommandLineArguments)(struct PROOFSEARCH_HELP*);
  int (*smt_parseSMTOutput)(struct PROOFSEARCH_HELP*, char*, EVECTOR**, EVECTOR**, int);
  void (*smt_addAssertion)(struct PROOFSEARCH_HELP*, FILE*, char*, int, int);
  void (*smt_printCheck)(struct PROOFSEARCH_HELP*, FILE*);
  void (*smt_printSPASSSignature)(struct PROOFSEARCH_HELP*, FILE*);
  void (*smt_printComment)(struct PROOFSEARCH_HELP*, FILE*, char*);
  void (*smt_prepareNonFilteredClauseInput)(struct PROOFSEARCH_HELP*, FILE*);
  ESTRING* (*smt_SigToString)(const char*);
} SMT_INTERFACE;

/**************************************************************/
/* Initialiazation                                            */
/**************************************************************/

SMT_INTERFACE *smt_interface_Init(struct PROOFSEARCH_HELP*);


/**************************************************************/
/* Allocation and deconstruction                              */
/**************************************************************/

SMT_INTERFACE *smt_interface_Create(const char*, int, int);
void smt_interface_Disalloc(SMT_INTERFACE*);


/**************************************************************/
/* Access Functions                                           */
/**************************************************************/

static __inline__
char* smt_interface_getName(SMT_INTERFACE *interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  RETURNS: The name of the decision procedure.
**********************************************************/
{
  return interface->name;
}

static __inline__
int smt_interface_supportsUnsatisfiableCores(SMT_INTERFACE *interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  RETURNS: 1 if unsatisfiable cores can be used, 0
           otherwise.
**********************************************************/
{
  return interface->unsatisfiable_cores;
}

static __inline__
int smt_interface_supportsNonGroundExport(SMT_INTERFACE *interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  RETURNS: 1 if quantifier reasoning is supported, 0
           otherwise.
**********************************************************/
{
  return interface->nonground_export;
}

static __inline__
void smt_interface_InitProver(SMT_INTERFACE *interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  EFFECT:  Calls the initializer of the dp module.
**********************************************************/
{
  (*(interface->smt_InitProver))();
}

static __inline__
void smt_interface_FreeProver(SMT_INTERFACE *interface)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure.
  EFFECT:  Calls the finalizer of the dp module.
**********************************************************/
{
  (*(interface->smt_FreeProver))();
}

static __inline__
ESTRING* smt_interface_clauseToString(SMT_INTERFACE *interface,
			    CLAUSE Clause, char *type)
{
  return (*(interface->smt_clauseToString))(Clause, type);
}

static __inline__
EVECTOR* smt_interface_getCommandLineArguments(SMT_INTERFACE *interface,
					       struct PROOFSEARCH_HELP *Search)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure, a proofsearch
           object.
  RETURNS: An expandable vector of command line arguments
           suitable for the current configuration as
           specified by <Search>.
**********************************************************/
{
  return (*(interface->smt_getCommandLineArguments))(Search);
}

static __inline__
int smt_interface_parseSMTOutput(SMT_INTERFACE *interface,
				 struct PROOFSEARCH_HELP *Search,
				 char *filename,
				 EVECTOR **explanation,
				 EVECTOR **dep_positions,
				 int smt_exitstatus)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure, a proofsearch
           object, a file name, two pointers to
           expandable vectors used as return value, and
           an integer giving the exit status of the SMT
           procedure.
  RETURNS: 1: unsat, 2: sat, 3: unknown, 4: error or
           unrecognized; depending on the flag <DocProof>
           <explanation> is filled with a list of SPASS
           clauses that are the reason the empty clause
  EFFECT:  Analyses the output of the SMT procedure.
  CAUTION: *NEVER* call this routine directly, instead
           use <dpconnect_checkForExternalSMTResult>.
**********************************************************/
{
  return (*(interface->smt_parseSMTOutput))(Search, filename, explanation, dep_positions, smt_exitstatus);
}

static __inline__
void smt_interface_addAssertion(SMT_INTERFACE *interface,
				struct PROOFSEARCH_HELP *Search,
				FILE *export_file,
				char *formula,
				int spass_id,
				int position)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure, a proofsearch
           object, a pointer to a file stream, a string
           representing the formula, the SPASS clause id
           and the number indicating how many clauses the
           SMT procedure has seen so far.
  EFFECT:  Appends a yices instruction to the current
           dp input (in file <export_file>)
           which asserts the formula given by
           the string <formula>. Note that this is
           probably different if unsatisfiable cores are
           delivered or not.
  CAUTION: The file has to be opened before.
**********************************************************/
{
  (*(interface->smt_addAssertion))(Search, export_file, formula, spass_id, position);
}

static __inline__
void smt_interface_printCheck(SMT_INTERFACE *interface,
			      struct PROOFSEARCH_HELP *Search,
			      FILE *export_file)
/*********************************************************
  INPUT:   An SMT_INTERFACE object, a proofsearch object,
           and a pointer to a file stream.
  EFFECT:  Appends a dp instruction to the current
           dp input (in file <export_file>)
           which asks the DP for a check.
**********************************************************/
{
  (*(interface->smt_printCheck))(Search, export_file);
}

static __inline__
void smt_interface_printSPASSSignature(SMT_INTERFACE *interface,
				       struct PROOFSEARCH_HELP *Search,
				       FILE *export_file)
/*********************************************************
  INPUT:   An SMT_INTERFACE object, a proofsearch object,
           and a pointer to a file stream.
  EFFECT:  Appends a comment to the dp input (in file
           <export_file>). Can be left empty.
**********************************************************/
{
  (*(interface->smt_printSPASSSignature))(Search, export_file);
}

static __inline__
void smt_interface_printComment(SMT_INTERFACE *interface,
				struct PROOFSEARCH_HELP *Search,
				FILE *file,
				char *comment)
/*********************************************************
  INPUT:   An SMT_INTERFACE object, a proofsearch object,
           and a pointer to a file stream and a string.
  EFFECT:  Appends a dp comment (with content <comment>)
           to the current dp input (in file
           <export_file>).
**********************************************************/
{
  (*(interface->smt_printComment))(Search, file, comment);
}

static __inline__
void smt_interface_prepareNonFilteredClauseInput(SMT_INTERFACE *interface,
						 struct PROOFSEARCH_HELP *Search,
						 FILE *file)
/*********************************************************
  INPUT:   An SMT_INTERFACE structure, a proofsearch
           object and a pointer to a file stream.
  EFFECT:  Inserts some default stuff to the non-filtered
           input for the SMT procedure.
**********************************************************/
{
  (*(interface->smt_prepareNonFilteredClauseInput))(Search, file);
}

static __inline__
ESTRING* smt_interface_SigToString(SMT_INTERFACE *interface,
				   const char *type)
/**************************************************************
  INPUT:   An SMT_INTERFACE structure, a theory type
           specification.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts the signature to the format of the
           external SMT procedure. Uses <Type> as base type
           for free function symbols and predicates.
***************************************************************/
{
  return (*(interface->smt_SigToString))(type);
}

#endif
