/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              RESULTS OF THE SMT PROCEDURE              * */
/* *                                                        * */
/* *  $Module:   smt_result                                 * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "smt_result.h"
#include "memory.h"
#include "search.h"


/**************************************************************/
/* Implementations                                            */
/**************************************************************/

SMT_RESULT* smt_result_Create(int status, EVECTOR *explanation,
			      EVECTOR *dependent_positions)
/**************************************************************
  INPUT:   A status, an explanation (which might be <NULL>),
           an evector of dependent_positions (which might be
           <NULL>), a position.
  RETURNS: An initialized <SMT_RESULT> object.
  EFFECT:  Allocated memory for the encapsulating structure
           and initializes their values.
  CAUTION: None of the given arguments is copied deeply, they
           must not be freed except using the function
           <smt_result_Free>!
***************************************************************/
{
  SMT_RESULT* local;

  local = (SMT_RESULT*) memory_Malloc(sizeof(SMT_RESULT));

  local->status = status;
  local->explanation = explanation;
  local->dependent_positions = dependent_positions;
  local->seen = 0;

  return local;
}

SMT_RESULT* smt_result_CreateUnsatWithoutEvidence(CLAUSE_DICTIONARY *dictionary)
/**************************************************************
  INPUT:   A clause dictionary.
  RETURNS: An initialized <SMT_RESULT> object
           with status "unsatisfiable without explanation".
  EFFECT:  Generates an overapproximation as explanation: takes
           the set of all clauses that were given to the SMT
           procedure as unsatisfiable cores. All its numbers
           and dependent positions are taken from
           <dictionary>.
           Uses function <smt_result_Create> with appropriate
           status.
***************************************************************/
{
  /* return over-approximation as unsatisfiable core */
  return
    smt_result_Create(1,
		      evector_copyEvectorInt(clause_dictionary_getVectorOfClauseNumbers(dictionary)),
		      clause_dictionary_getAllDependentPositionsCopy(dictionary));
}

SMT_RESULT* smt_result_CreateUnsatWithEvidence(EVECTOR *explanation,
					       EVECTOR *dependent_positions)
/**************************************************************
  INPUT:   An explanation, an evector of dependent_positions.
  RETURNS: An initialized <SMT_RESULT> object
           with status "unsatisfiable with explanation".
  EFFECT:  Uses function <smt_result_Create> with appropriate
           status.
  CAUTION: None of the given arguments is copied deeply, they
           must not be freed except using the function
           <smt_result_Free>!
***************************************************************/
{
  return(smt_result_Create(2, explanation, dependent_positions));
}

SMT_RESULT* smt_result_CreateSat(EVECTOR *dependent_positions)
/**************************************************************
  INPUT:   An evector of branches that have been considered.
  RETURNS: An initialized <SMT_RESULT> object
           with status "satisfiable".
  EFFECT:  Uses function <smt_result_Create> with appropriate
           status.
  CAUTION: None of the given arguments is copied deeply, they
           must not be freed except using the function
           <smt_result_Free>!
***************************************************************/
{
  return(smt_result_Create(3, NULL, dependent_positions));
}

SMT_RESULT* smt_result_CreateUnknown(EVECTOR *dependent_positions)
/**************************************************************
  INPUT:   An evector of branches that have been considered.
  RETURNS: An initialized <SMT_RESULT> object
           with status "unknown".
  EFFECT:  Uses function <smt_result_Create> with appropriate
           status.
  CAUTION: None of the given arguments is copied deeply, they
           must not be freed except using the function
           <smt_result_Free>!
***************************************************************/
{
  return(smt_result_Create(4, NULL, dependent_positions));
}

void smt_result_Free(SMT_RESULT* result)
/**************************************************************
  INPIT:   An initialized <SMT_RESULT> object.
  EFFECT:  Frees memory for the encapsulating structure
           and all their values by invoking the suitabes
           functions.
***************************************************************/
{
  if (result->explanation != NULL)
    evector_destroyEvectorInt(result->explanation);

  if (result->dependent_positions != NULL)
    evector_deconstructEvector(result->dependent_positions,
			       &position_evector_destroyPosition);

  memory_Free(result, sizeof(SMT_RESULT));
}

BOOL smt_result_isRelevant(SMT_RESULT *smt_result, struct PROOFSEARCH_HELP *Search,
			   POSITION *p_spass)
/**************************************************************
  INPUT:   An SMT_RESULT object, a Proof Search object, and
           the current working position of SPASS.
  RETURNS: If the result is relevant wrt. to the current
           SPASS configuration, 1. Otherwise 0.
  EFFECT:  A result is relevant if
             (a) is is "unsat" or "unknwon"
                 and p_result <=_p p_spass and
                 took all condensations into account
             (b) is "sat" and p_result <=_p p_spass and all
                 non-ground formulae have been exported and
                 the SMT procedure already saw all input
                 clauses
  CAUTION: Note, that a relevant result might not necessarily
           contribute to the SPASS proof.
***************************************************************/
{
  BOOL result;
  DPCONNECT dpc;

  /* abbreviation */
  dpc = prfs_DpConnect(Search);

  result = FALSE;

  if (smt_result_isSat(smt_result)) {
    /* if ((flag_GetFlagValue(prfs_Store(Search), flag_DPCNONGROUNDEXPORT) != flag_OFF) &&
	(dpc->got_all_problem_input_clauses == 1) &&
	(position_PrefixOrderEqual(smt_result_getHighestPosition(smt_result),
				     p_spass))) */
        /* was: !position_BranchOrderStrict(...) */
      result = FALSE;
        /* was: TRUE */
  } else if (smt_result_isUnsat(smt_result)) {
    if ((position_PrefixOrderEqual(smt_result_getHighestPosition(smt_result),
				     p_spass)) &&
	(! condensationscope_dependsOnCondensedBranch(prfs_CondensationScope(Search),
						      smt_result_getDependentBranches(smt_result))))
        /* was: !position_BranchOrderStrict(...) */
      result = TRUE;
  } else if (smt_result_isUnknown(smt_result)) {
    /* if ((position_PrefixOrderEqual(smt_result_getHighestPosition(smt_result),
				     p_spass)) &&
	(! condensationscope_dependsOnCondensedBranch(prfs_CondensationScope(Search),
						      smt_result_getDependentBranches(smt_result)))) */
        /* was: !position_BranchOrderStrict(...) */
      result = FALSE;
        /* was: TRUE */
  }

  return result;
}

CLAUSE smt_result_generateEmptyClause(struct PROOFSEARCH_HELP *Search, SMT_RESULT *result)
/**************************************************************
  INPUT:   A proofsearch object and an SMT_RESULT object.
  RETURNS: If <result> is unsatisfiable then an empty will be
           returned, otherwise NULL will be returned.
  EFFECT:  If <result> encodes an unsatisfiability, then the
           empty clause that corresponds to this
           unsatisfiabiltiy is created:
               (a) Its set of parent clause numbers are set to
                   the numbers in result->explanation.
               (b) The i-th entry of the splitfield will be
                   set to 1 if a position with length i is a
                   member of result->dependent_positions,
                   and otherwise to 0
  CAUTION: In case of an unsatisfiability result that was
           overapproximated, the empty clause that is
           generated will be overapproximated too.
***************************************************************/
{
  CLAUSE EmptyClause;
  int i;

  EmptyClause = (CLAUSE) NULL;

  if (smt_result_isUnsat(result)) {
    /* generate clause of length 0 */
    EmptyClause = clause_CreateBody(0);

    /* clause origin is "theory inconsistency" */
    clause_SetFromTheoryInconsistency(EmptyClause);

    /* set up list of parent clauses */
    for(i=0; i<evector_getNumberOfElements(result->explanation); ++i) {
      clause_AddParentClause(EmptyClause,
			     evector_getNthInt(result->explanation, i));
      /* we don't have any information about which literal was used */
      clause_AddParentLiteral(EmptyClause, -1);
    }

    /* set split data of empty clause induced by the dependent positions */
    clause_SetSplitLevel(EmptyClause, position_getDepth(smt_result_getHighestPosition(result)));
    for(i=0; i<evector_getNumberOfElements(result->dependent_positions); ++i) {
      if (!position_isRootPosition((POSITION*)evector_getNth(result->dependent_positions, i)))
	clause_SetSplitFieldBit(EmptyClause,
				position_getDepth((POSITION*)
						  evector_getNth(result->dependent_positions, i)));
    }

    /* do rest of initializations necessary for new clauses */
    clause_OrientAndReInit(EmptyClause, prfs_Store(Search), prfs_Precedence(Search));
    clause_RemoveFlag(EmptyClause, CONCLAUSE);
    clause_IsFromTheoryInconsistency(EmptyClause);
  }

  return EmptyClause;
}
