/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              RESULTS OF THE SMT PROCEDURE              * */
/* *                                                        * */
/* *  $Module:   smt_result                                 * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


#ifndef _SMTRESULT_
#define _SMTRESULT_


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "position.h"
#include "clause_dictionary.h"
#include "evector.h"
#include "misc.h"
#include "clause.h"


/**************************************************************/
/* Datastructures                                             */
/**************************************************************/

struct PROOFSEARCH_HELP;

typedef struct {
  /* 1: unsat without unsatf. core, 2: unsat with unsatf. core,
     3: sat, 4: unknown, 5: error, or unrecognized */
  int status;

  /* unsatisfiable core */
  EVECTOR *explanation;

  /* list of positions of all branches the clauses that have been considered
     (or are contained in <explanation>) depend on */
  EVECTOR *dependent_positions;

  /* whether result was seen or not */
  int seen;
} SMT_RESULT;


/**************************************************************/
/* Function prototypes                                        */
/**************************************************************/

SMT_RESULT* smt_result_Create(int, EVECTOR*, EVECTOR*);
SMT_RESULT* smt_result_CreateUnsatWithoutEvidence(CLAUSE_DICTIONARY*);
SMT_RESULT* smt_result_CreateUnsatWithEvidence(EVECTOR*, EVECTOR*);
SMT_RESULT* smt_result_CreateSat(EVECTOR*);
SMT_RESULT* smt_result_CreateUnknown(EVECTOR*);

void smt_result_Free(SMT_RESULT*);
BOOL smt_result_isRelevant(SMT_RESULT*, struct PROOFSEARCH_HELP*, POSITION*);

CLAUSE smt_result_generateEmptyClause(struct PROOFSEARCH_HELP*, SMT_RESULT*);


/**************************************************************/
/* Inline Functions                                           */
/**************************************************************/

static __inline__
int smt_result_seen(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if the SPASS already has seen (proceed) the
           <result>, 0 otherwise.
**********************************************************/
{
  return result->seen;
}

static __inline__
void smt_result_markSeen(SMT_RESULT* result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  EFFECT:  Marks <result> as already proceeded.
**********************************************************/
{
  result->seen = 1;
}

static __inline__
int smt_result_isUnsatWithoutEvidence(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is unsatisfiable without evidence,
           0 otherwise.
**********************************************************/
{
  return (result->status == 1);
}

static __inline__
int smt_result_isUnsatWithEvidence(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is unsatisfiable witht evidence,
           0 otherwise.
**********************************************************/
{
  return (result->status == 2);
}

static __inline__
int smt_result_isUnsat(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is unsatisfiable with/without
           evidence, 0 otherwise.
**********************************************************/
{
  return (smt_result_isUnsatWithoutEvidence(result) ||
	  smt_result_isUnsatWithEvidence(result));
}

static __inline__
int smt_result_isSat(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is satisfiable, 0 otherwise.
**********************************************************/
{
  return (result->status == 3);
}

static __inline__
int smt_result_isUnknown(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is unknown, 0 otherwise.
**********************************************************/
{
  return (result->status == 4);
}

static __inline__
int smt_result_isError(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: 1 if result is error, 0 otherwise.
**********************************************************/
{
  return (result->status == 5);
}

static __inline__
EVECTOR* smt_result_getExplanation(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: The unsatisfiable core.
**********************************************************/
{
#ifdef CHECK
  if (smt_result_isUnsat(result) != 1) {
    misc_StartErrorReport();
    misc_ErrorReport("In smt_result_getExplanation:\n");
    misc_ErrorReport("Result under consideration is not an \"unsatisfiable\"!\n");
    misc_FinishErrorReport();
  }
#endif

  return result->explanation;
}

static __inline__
POSITION* smt_result_getHighestPosition(SMT_RESULT *result)
/*********************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: The position of the highest branch that was
           considered and refers to the result.
**********************************************************/
{
  return ((POSITION*) evector_getMaxElement(result->dependent_positions));
}

static __inline__
EVECTOR* smt_result_getDependentBranches(SMT_RESULT *result)
/**************************************************************
  INPUT:   An SMT_RESULT object.
  RETURNS: The evector of dependent positions.
***************************************************************/
{
  return result->dependent_positions;
}
#endif
