/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               CONNECTION TO YICES                      * */
/* *                                                        * */
/* *  $Module: smt_yices                                    * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Uwe Waldmann, Stephan Zimmer     * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: 2007                                              * */
/* $Author: Uwe Waldmann, Stephan Zimmer                    * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Uwe Waldmann, Stephan Zimmer               * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: uwe@mpi-inf.mpg.de                  * */
/* *                    st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "smt_yices.h"
#include "search.h"
#include <errno.h>

/**************************************************************/
/* Module internal variables                                  */
/**************************************************************/

#define _SMTYICES_BIGSTRING_SIZE 8192
#define _SMTYICES_IDPREFIX "wq"
char smt_yices_big_string[_SMTYICES_BIGSTRING_SIZE];
LIST smt_yices_keywords;
LIST smt_yices_error_return_codes;


/**************************************************************/
/* Implementations                                            */
/**************************************************************/

SMT_INTERFACE* smt_yices_getInterface(void)
/*********************************************************
  RETURNS: The created instance of the interface for
           addressing the SMT functionalities from
           dpconnect.
  EFFECT:  Makes the internal functionality usable for
           dpconnect by establishing an instance of
           the datatype SMT_INTERFACE.
**********************************************************/
{
  SMT_INTERFACE *local;

  /* what are the capabilites of yices? */
  local = smt_interface_Create("yices",  /* name */
			       1,        /* unsatf. cores supp. */
			       1         /* non-gr. export supp. */
			       );

  /* provide pointers to the implementation of the functions that
     are required by the interface */
  local->smt_InitProver = &smt_yices_Init;
  local->smt_FreeProver = &smt_yices_Free;
  local->smt_clauseToString = &smt_yices_YicesClauseToString;
  local->smt_getCommandLineArguments = &smt_yices_getCommandLineArguments;
  local->smt_parseSMTOutput = &smt_yices_parseSMTOutput;
  local->smt_addAssertion = &smt_yices_printAssertion;
  local->smt_printCheck = &smt_yices_printCheck;
  local->smt_printSPASSSignature = &smt_yices_printSPASSSignature;
  local->smt_printComment = &smt_yices_printComment;
  local->smt_prepareNonFilteredClauseInput = &smt_yices_prepareNonFilteredClauseInput;
  local->smt_SigToString = &smt_yices_YicesSigToString;

  return local;
}

void smt_yices_Init(void)
/*********************************************************
  EFFECT:  Initializes the module, in particular sets up
           the list of forbidden variable names for the
           SMT procedure.
**********************************************************/
{
  /* build the list of Yices keywords */
  smt_yices_keywords = list_Nil();
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("assert"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("assert+"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("retract"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("check"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("max-sat"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("push"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("pop"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("echo"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("reset"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("include"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("status"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("exit"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("help"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("set-evidence!"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("set-verbosity!"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("set-arith-only!"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("define-type"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("subtype"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("subrange"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("tuple"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("record"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("datatype"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("scalar"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bitvector"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("and"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("or"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("not"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("let"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("forall"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("exists"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("update"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("lambda"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("mk-tuple"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("select"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("mk-record"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("mk-bv"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-concat"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-extract"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-shift-left0"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-shift-left1"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-shift-right0"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-shift-right1"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-sign-extend"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-and"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-or"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-xor"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-not"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-add"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-sub"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-mul"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-neg"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-lt"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-le"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-gt"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-ge"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-slt"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-sle"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-sgt"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bv-sge"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("dump-context"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("int"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("bool"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("real"),
				 smt_yices_keywords);
  smt_yices_keywords = list_Push((POINTER)string_StringCopy("nat"),
				 smt_yices_keywords);

  /* TO DO: add errorneous return codes later on */
  smt_yices_error_return_codes = list_Nil();
}

void smt_yices_Free(void)
/*********************************************************
  EFFECT:  Frees the module.
**********************************************************/
{
  LIST Scan;

  for(Scan = smt_yices_keywords; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    string_StringFree((char*)list_Car(Scan));
  }

  list_Delete(smt_yices_keywords);

  list_Delete(smt_yices_error_return_codes);
}

EVECTOR* smt_yices_getCommandLineArguments(struct PROOFSEARCH_HELP* Search)
/*********************************************************
  INPUT:   A proofsearch object.
  RETURNS: An expandable vector of command line arguments
           suitable for the current configuration as
           specified by <Search>.
**********************************************************/
{
  EVECTOR *local;
  FLAGSTORE store;
  char temp[30];

  /* abbreviation */
  store = prfs_Store(Search);

  local = evector_createEvectorString(15, 2);

  /* here some furhter fine tuning can be done */
  sprintf(temp, "--max-instances=%d", 30000);
  evector_insertAppendedString(local, temp);
  sprintf(temp, "--max-patterns=%d", 10);
  evector_insertAppendedString(local, temp);

  /* do we need evidence? */
  if (flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) == flag_ON) {
    evector_insertAppendedString(local, "-e");
    evector_insertAppendedString(local, "--verbose=2");
  }

  /* name of input file as last argument */
  evector_insertAppendedString(local, (prfs_DpConnect(Search))->smt_input_filename);

  return local;
}

int smt_yices_parseSMTOutput(struct PROOFSEARCH_HELP* Search,
			     char *output_filename,
			     EVECTOR** explanation,
			     EVECTOR** dep_positions, int smt_exitstatus)
/*********************************************************
  INPUT:   A proofsearch object, a file name, two
           pointers to  expandable vectors used as return
           value, and an integer giving the exit status
           of the SMT procedure.
  RETURNS: 1: unsat, 2: sat, 3: unknown, 4: error or
           unrecognized; depending on the flag
           <DpcEvidence> <explanation> is filled with a
           list of SPASS clauses that are the reason the
           for empty clause.
  EFFECT:  Analyses the output of the SMT procedure.
  CAUTION: *NEVER* call this routine directly, instead
           use <dpconnect_checkForExternalSMTResult>.
**********************************************************/
{
  FILE *output;
  int result, i, j, current_index;
  char *explanation_start;
  EVECTOR *unsat_core;
  CLAUSE_DICTIONARY *dictionary;
#ifdef CHECK
  int max_explanation_id;
#endif

  /* initialization */
  *explanation = (EVECTOR*) NULL;
  *dep_positions = (EVECTOR*) NULL;

  /* abbreviation */
  dictionary = (prfs_DpConnect(Search))->clause_dictionary;

#ifdef CHECK
  /* check whether file with output exists */
  if (misc_ExistsFile(output_filename) == 0) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In dpconnect_parseSMTOutput:");
    misc_ErrorReport("\n No file containing SMT output existing.\n");
    misc_FinishErrorReport();
    return 4;
  }
#endif

  /* first of all, did we get a correct exit status? */
  if (list_PointerMember(smt_yices_error_return_codes, (POINTER) smt_exitstatus))
    return 4;

  /* try to inspect output */
  output = fopen(output_filename, "r");
  if (output == NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_parseSMTOutput:");
    misc_ErrorReport("\n Could not open file for reading SMT output.\n");
    misc_FinishErrorReport();
  }

  result = -1;
  while (NULL != fgets(smt_yices_big_string, _SMTYICES_BIGSTRING_SIZE, output)) {
    if (strstr(smt_yices_big_string, "unsat core ids: ") != NULL) {
      result = 1;
    } else if (strstr(smt_yices_big_string, "unsat\n") != NULL) {
      if (flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) == flag_OFF) {
	result = 1;   /* "unsat" found, no explanation needed */
      } else {
        result = -2;  /* "unsat" found, explanation needed, but not found yet */
      }
    } else if (strstr(smt_yices_big_string, "sat\n") != NULL) {
      result = 2;
    } else if (strstr(smt_yices_big_string, "unknown\n") != NULL) {
      result = 3;
    } else if (strstr(smt_yices_big_string, "error\n") != NULL) {
      result = 4;
    }

    if (result > 0)
      /* we got what we needed */
      break;
  }

  if (flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) != flag_OFF
      && result == 1) {
    /* we found "unsat" and an explanation (and need it) */

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("In smt_yices_parseSMTOutput: Got the following result from yices:\n%s\n",
	   smt_yices_big_string);
    smt_debug_end();
#endif

    /* grap the explanation out of the output */
    explanation_start = ((char*) smt_yices_big_string) + strlen("unsat core ids: ");

    unsat_core = evector_createFromStringInt(explanation_start, " ");

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("We converted this to:\n");
    evector_printEVectorInt(unsat_core);
    printf("\n");
    smt_debug_end();
#endif

#ifdef CHECK
    max_explanation_id = evector_getMaxElementInt(unsat_core);
    if (max_explanation_id >
	clause_dictionary_getNumberOfElements(dictionary)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_yices_parseSMTOutput:");
      misc_ErrorReport("\n Inconsistent datastructures.\n");
      misc_FinishErrorReport();
    }
#endif

    /* prepare datastructure to take the explanation */
    *dep_positions =
      evector_createEvector(clause_dictionary_getNumberOfElements(dictionary),
			    2, &position_evector_equal, &position_evector_tree_order);
    *explanation =
      evector_createEvectorInt(10+clause_dictionary_getNumberOfElements(dictionary),
			       2);

    /* create explanation:
       (1) look in dictionary for corresponding SPASS clause numbers
       (2) look in dictionary for dependent positions and accumulate them */
    for (i=0; i<evector_getNumberOfElements(unsat_core); ++i) {
      current_index = evector_getNthInt(unsat_core, i)-1;

      /* (1) insert SPASS clause number */
      evector_insertAppendedInt(*explanation,
				clause_dictionary_getNthClauseNo(dictionary, current_index));

      /* (2) accumulate dependent positions */
      for(j=0;
	  j<evector_getNumberOfElements(clause_dictionary_getNthDependentPositions(dictionary,
										   current_index));
	  ++j) {
	if (!evector_isMember(*dep_positions,
			      (POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)))
	  evector_insertAppended(*dep_positions,
				 position_copyPosition((POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)));
      }
    }
    /* free memory for temporary allocated evector */
    evector_destroyEvectorInt(unsat_core);
#ifdef CHECK
    if (evector_getNumberOfElements(*explanation) == 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_yices_parseSMTOutput:");
      misc_ErrorReport("\n Got an empty explanation.\n");
      misc_FinishErrorReport();
    }
#endif

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("This yields the following SPASS explanation:\n");
    evector_printEVectorInt(*explanation);
    printf("With the following positions the empty clause depends on:\n");
    evector_printEVector(*dep_positions, &position_evector_print);
    printf("\n");
    smt_debug_end();
#endif

  } else if (flag_GetFlagValue(prfs_Store(Search), flag_DPCEVIDENCE) != flag_OFF
             && result == -2) {
    /* we found "unsat" and need an explanation, but didn't find one, */
    /* so we return the entire dictionary instead */

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("In smt_yices_parseSMTOutput: Got \"unsat\" but no unsat core from yices\n");
    smt_debug_end();
#endif

    result = 1;

    /* prepare datastructure to take the explanation */
    *dep_positions =
      evector_createEvector(clause_dictionary_getNumberOfElements(dictionary),
			    2, &position_evector_equal, &position_evector_tree_order);
    *explanation =
      evector_createEvectorInt(10+clause_dictionary_getNumberOfElements(dictionary),
			       2);

    /* create explanation:
       (1) get all SPASS clause numbers from dictionary
       (2) look in dictionary for dependent positions and accumulate them */
    for (current_index=0; current_index<clause_dictionary_getNumberOfElements(dictionary); ++current_index) {

      /* (1) insert SPASS clause number */
      evector_insertAppendedInt(*explanation,
				clause_dictionary_getNthClauseNo(dictionary, current_index));

      /* (2) accumulate dependent positions */
      for(j=0;
	  j<evector_getNumberOfElements(clause_dictionary_getNthDependentPositions(dictionary,
										   current_index));
	  ++j) {
	if (!evector_isMember(*dep_positions,
			      (POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)))
	  evector_insertAppended(*dep_positions,
				 position_copyPosition((POSITION*) evector_getNth(clause_dictionary_getNthDependentPositions(dictionary, current_index), j)));
      }
    }
#ifdef CHECK
    if (evector_getNumberOfElements(*explanation) == 0) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In smt_yices_parseSMTOutput:");
      misc_ErrorReport("\n Got an empty explanation.\n");
      misc_FinishErrorReport();
    }
#endif

#ifdef SMT_DEBUG
    smt_debug_start();
    printf("This yields the following SPASS explanation:\n");
    evector_printEVectorInt(*explanation);
    printf("With the following positions the empty clause depends on:\n");
    evector_printEVector(*dep_positions, &position_evector_print);
    printf("\n");
    smt_debug_end();
#endif
  }

  /* close file containing the input */
  fclose(output);
  /* relax, ignore any errors */
  errno = 0;

  if (result < 0) {
    /* read all output, but could not match any of the patterns */
    /* => treat it as "unknown" */
    result = 3;
  }

#ifdef SMT_DEBUG
  smt_debug_start();
  printf("In smt_yices_parseSMTOutput: Got the following result from yices:\n%s\n",
	 smt_yices_big_string);
  smt_debug_end();
#endif

  return result;
}

void smt_yices_printAssertion(struct PROOFSEARCH_HELP* Search, FILE *export_file, char* formula, int spass_id, int position)
/*********************************************************
  INPUT:   A proofsearch object, a pointer to a file
           stream, a string representing the formula, the
           SPASS clause id and the number indicating how
           many clauses the SMT procedure has seen so
           far.
  EFFECT:  Appends a yices instruction to the current
           yices input (in file <export_file>)
           which asserts the formula given by
           the string <formula>. Depending on the value of
           <DpcEvidence> either "assert" or "assert+" is
           used.
  CAUTION: The file has to be opened before.
**********************************************************/
{
#ifdef CHECK
  if (export_file == (FILE*) NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_printAssertion:");
    misc_ErrorReport("\n Error providing yices output.\n");
    misc_FinishErrorReport();
  }
#endif

  if (flag_GetFlagValue(prfs_Store(Search),
			flag_DPCEVIDENCE) == flag_ON)
    fprintf(export_file, "(assert+ ");
  else
    fprintf(export_file, "(assert ");

  fprintf(export_file, "%s) ;; SPASS clause ID is %d\n", formula, spass_id);
}

void smt_yices_printCheck(struct PROOFSEARCH_HELP* Search, FILE *export_file)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream.
  EFFECT:  Appends a yices instruction to the current
           yices input (in file <export_file>)
           which asks the DP for a check.
**********************************************************/
{
#ifdef CHECK
  if (export_file ==  NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_printAssertion:");
    misc_ErrorReport("\n Error providing yices output.\n");
    misc_FinishErrorReport();
  }
#endif

  fprintf(export_file, ";; check the context\n(check)\n");
}

void smt_yices_printComment(struct PROOFSEARCH_HELP* Search, FILE *export_file, char *comment)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream and a string.
  EFFECT:  Appends a yices comment to the current yices
           input (in file <export_file>).
**********************************************************/
{
  fprintf(export_file, ";; %s\n", comment);
}

void smt_yices_prepareNonFilteredClauseInput(struct PROOFSEARCH_HELP *Search, FILE *f)
/*********************************************************
  INPUT:   A proofsearch object and a pointer to a file
           stream.
  EFFECT:  Inserts some default stuff to the non-filtered
           input of the SMT procedure.
**********************************************************/
{
  /* for yices nothing has to be done. */
}

void smt_yices_printSPASSSignature(struct PROOFSEARCH_HELP* Search, FILE* export_file)
/*********************************************************
  INPUT:   A proofsearch object, and a pointer to a file
           stream.
  EFFECT:  Appends a comment to the yices input (in file
           <export_file>).
**********************************************************/
{
#ifdef CHECK
  if (export_file ==  NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_printAssertion:");
    misc_ErrorReport("\n Error providing yices output.\n");
    misc_FinishErrorReport();
  }
#endif

  fprintf(export_file, ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
  fprintf(export_file, ";; This yices input was automatically created ;;\n");
  fprintf(export_file, ";; by SPASS+T.                                ;;\n");
  fprintf(export_file, ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n\n");
}


/**************************************************************/
/* Conversion Functions                                       */
/**************************************************************/

void smt_yices_YicesClausePrint(FILE* File, CLAUSE Clause) {
/**************************************************************
  INPUT:   A file and a clause.
  SUMMARY: Prints any clause to File.
  CAUTION: Uses the other clause_Output functions.
***************************************************************/
  int     n,j;
  LITERAL Lit;
  TERM    Atom;

  n = clause_Length(Clause);

  /* > SZ < */

  fputs("(assert (or ", File);
  if (n == 0) {
    fputs("false", File);
  } else {
    for (j = 0; j < n; j++) {
      Lit  = clause_GetLiteral(Clause,j);
      Atom = clause_LiteralAtom(Lit);
      if (clause_LiteralIsNegative(Lit)) {
        fputs("(not ", File);
        smt_yices_YicesAtomPrint(File, Atom);
        putc(')', File);
      } else {
	smt_yices_YicesAtomPrint(File, Atom);
      }
      if (j <= (n-2)) {
        fputs(" ", File);
      }
    }
  }
  fputs("))\n", File);
  fflush(File);
}

void smt_yices_YicesPrintSig(FILE* File, const char* Type) {
/**************************************************************
prints the signature to the file in Yices syntax using <Type>
as the only argument and result type.
***************************************************************/
  if (symbol_SignatureExists()) {
    int          i, Index;
    SYMBOL       Info;
    SIGNATURE    Entry;

    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Oops: This shouldn't happen (smt_yices_YicesPrintSig)\n");
    misc_FinishUserErrorReport();

    for (Index = 1 ; Index < symbol_ACTINDEX ; Index++) {
      Entry  = symbol_Signature(Index);
      if (Entry != NULL) {
	Info = Entry->info;

	switch (symbol_Type(Info)) {
	case symbol_CONSTANT:
          if (string_StringIsNumber(Entry->name)
              ||  string_StringIsNumber((Entry->name)+1)) {
            continue;
	  }
          fprintf(File, "(define %s::%s)\n", Entry->name, Type);
          break;
	case symbol_FUNCTION:
          if (symbol_Equal(Info, fol_ThSucc())
              || symbol_Equal(Info, fol_ThPred())
              || symbol_Equal(Info, fol_ThPlus())
              || symbol_Equal(Info, fol_ThMinus())
              || symbol_Equal(Info, fol_ThUminus())
              /* we declare "times" in addition to the built-in "*",
                 so that we can use it in non-linear contexts */
              /* || symbol_Equal(Info, fol_ThTimes()) */
              /* we declare "divide" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
              /* || symbol_Equal(Info, fol_ThDivide()) */
              /* we declare "divide_nat" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
	      /* || symbol_Equal(Info, fol_ThDivideNat()) */
              || symbol_Equal(Info, fol_ThRemainder())) {
            continue;
	  }
          fprintf(File, "(define %s::(-> ", Entry->name);
          for (i = Entry->arity; i >= 0; i--) {
            fprintf(File, "%s%s", Type, i == 0 ? "))\n" : " ");
	  }
	  break;
        case symbol_PREDICATE:
          if (symbol_Equal(Info, fol_Equality())
              || symbol_Equal(Info, fol_ThLess())
              || symbol_Equal(Info, fol_ThLesseq())
              || symbol_Equal(Info, fol_ThGreater())
              || symbol_Equal(Info, fol_ThGreatereq())
              || symbol_Equal(Info, fol_ThGround())
              || symbol_Equal(Info, fol_True())
              || symbol_Equal(Info, fol_False())) {
            continue;
	  }
          fprintf(File, "(define %s::(-> ", Entry->name);
          for (i = Entry->arity; i > 0; i--) {
            fprintf(File, "%s%s", Type, " ");
	  }
          fprintf(File, "bool))\n");
	  break;
	case symbol_JUNCTOR:
	  ;
	}
      }
    }
  }
}


void smt_yices_YicesAtomPrint(FILE* File, TERM Atom) {
/**************************************************************
  INPUT:   A file and an atom.
  SUMMARY: Prints any atom to File.
  CAUTION: Uses the other clause_Output functions.
***************************************************************/
  SYMBOL TopSymbol;

  TopSymbol = term_TopSymbol(Atom);

  if (term_IsComplex(Atom)) {
    putc('(', File);
    if (TopSymbol == fol_Equality()) {
      fputs("=", File);
    } else if (TopSymbol == fol_ThLess()) {
      fputs("<", File);
    } else if (TopSymbol == fol_ThLesseq()) {
      fputs("<=", File);
    } else if (TopSymbol == fol_ThGreater()) {
      fputs(">", File);
    } else if (TopSymbol == fol_ThGreatereq()) {
      fputs(">=", File);
    } else {
      symbol_FPrint(File, TopSymbol);
    }
    putc(' ', File);
    smt_yices_YicesTermListPrint(File, term_ArgumentList(Atom));
    putc(')', File);
  } else {
    /* term_IsConstant(Atom) */
    symbol_FPrint(File, TopSymbol);
  }
}


void smt_yices_YicesTermPrint(FILE* File, TERM Term) {
/**************************************************************
  INPUT:   A file and a term.
  RETURNS: none.
  SUMMARY: Prints the term to the file.
  CAUTION: Uses the other term output functions.
***************************************************************/
  SYMBOL TopSymbol;

  misc_StartUserErrorReport();
  misc_UserErrorReport("\n Oops: This shouldn't happen (smt_yices_YicesTermPrint)\n");
  misc_FinishUserErrorReport();

  if (term_IsComplex(Term)) {
    putc('(', File);
    TopSymbol = term_TopSymbol(Term);

    if (TopSymbol == fol_ThSucc()) {
      fputs("+ ", File);
      smt_yices_YicesTermPrint(File, term_FirstArgument(Term));
      fputs(" 1)", File);
      return;
    } else if (TopSymbol == fol_ThPred()) {
      fputs("- ", File);
      smt_yices_YicesTermPrint(File, term_FirstArgument(Term));
      fputs(" 1)", File);
      return;
    } else if (TopSymbol == fol_ThUminus()) {
      fputs("- 0", File);
    } else if (TopSymbol == fol_ThPlus()) {
      fputs("+", File);
    } else if (TopSymbol == fol_ThMinus()) {
      fputs("-", File);
    } else if (TopSymbol == fol_ThTimes()
               && (term_IsNumeric(term_FirstArgument(Term))
                   || term_IsNumeric(term_SecondArgument(Term)))) {
      TERM t1, t2;
      if (term_IsNumeric(term_FirstArgument(Term))) {
        t1 = term_FirstArgument(Term);
        t2 = term_SecondArgument(Term);
      } else {
        t1 = term_SecondArgument(Term);
        t2 = term_FirstArgument(Term);
      }
      fputs("* ", File);
      smt_yices_YicesTermPrint(File, t1);
      fputs(" ", File);
      smt_yices_YicesTermPrint(File, t2);
      putc(')', File);
      return;
    } else {
      fputs(symbol_Name(TopSymbol), File);
    }
    fputs(" ", File);
    smt_yices_YicesTermListPrint(File,term_ArgumentList(Term));
    putc(')', File);

  } else if (term_IsVariable(Term)) {
      putc('0', File);
  } else {

    /* term_IsConstant(Term) */
    symbol_FPrint(File,term_TopSymbol(Term));
  }
}


void smt_yices_YicesTermListPrint(FILE* File, LIST List) {
/**************************************************************
  INPUT:   A file and a list of terms.
  RETURNS: None.
***************************************************************/
  for (; !list_Empty(List); List=list_Cdr(List)) {
    smt_yices_YicesTermPrint(File,list_Car(List));
    if (!list_Empty(list_Cdr(List)))
      putc(' ', File);
  }
}

BOOL dpconnect_VariableIDsEqual(POINTER p1, POINTER p2)
{
  return (*((NAT*) p1) == *((NAT*) p2));
}

ESTRING* smt_yices_YicesQuantifierPrefixToString(CLAUSE Clause, char *type)
/**************************************************************
  INPUT:   A clause, the type of the variables.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts the quantifier prefix into string, using
           <type> for all variables.
***************************************************************/
{
  ESTRING *estring;
  char convert[50];
  LIST vars, iter;

  estring = estring_EmptyString(200);
  estring_AppendString(estring, "(forall ( ");

  vars =  clause_ListOfVariables(Clause);
  vars = list_DeleteDuplicates(vars, &dpconnect_VariableIDsEqual);

  for (iter=vars; !list_Empty(iter); iter=list_Cdr(iter)) {
    strcpy(convert, "");
    sprintf(convert, "%d", *((NAT *)list_Car(iter)));

    estring_AppendString(estring, " U_");
    estring_AppendString(estring, convert);
    estring_AppendString(estring, " :: ");
    estring_AppendString(estring, type);
  }

  estring_AppendString(estring, " )");

  list_Delete(vars);

  return estring;
}

ESTRING* smt_yices_YicesClauseToString(CLAUSE Clause, char *type)
/**************************************************************
  INPUT:   A clause, a type
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Clause> into the format of the external
           SMT procedure; variables are assumed to be of type
           <type>.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING *estring, *qprefix_estring;
  int     n,j;
  LITERAL Lit;
  TERM    Atom;
  BOOL    numericTheory;

  numericTheory = (strcmp(type, "data") != 0);

  n = clause_Length(Clause);

  estring = estring_EmptyString(150);

  /* > SZ < */

  /* do we have to add a quantifier prefix? */
  if (clause_ContainsVariables(Clause)) {
    qprefix_estring = smt_yices_YicesQuantifierPrefixToString(Clause, type);
    estring_AppendEStringDelete(estring, qprefix_estring);
  }

  estring_AppendString(estring, " (or ");
  /* fputs("(assert (or ", File); */
  if (n == 0) {
    estring_AppendString(estring, "false");
    /* fputs("false", File); */
  } else {
    for (j = 0; j < n; j++) {
      Lit  = clause_GetLiteral(Clause,j);
      Atom = clause_LiteralAtom(Lit);
      if (clause_LiteralIsNegative(Lit)) {
	estring_AppendString(estring, "(not ");
	/* fputs("(not ", File); */

	estring_AppendEStringDelete(estring, smt_yices_YicesAtomToString(Atom, numericTheory));
	/* smt_yices_YicesAtomPrint(File, Atom); */

	estring_AppendString(estring, ")");
        /* putc(')', File); */
      } else {
	estring_AppendEStringDelete(estring, smt_yices_YicesAtomToString(Atom, numericTheory));
	/* smt_yices_YicesAtomPrint(File, Atom); */
      }
      if (j <= (n-2)) {
	estring_AppendString(estring, " ");
        /* fputs(" ", File); */
      }
    }
  }

  estring_AppendString(estring,
		       (clause_ContainsVariables(Clause)  ? "))" : ")"));

  /* fputs("))\n", File); */
  /* fflush(File); */

  return estring;
}

ESTRING* smt_yices_YicesFormulaToString(TERM Formula, char *type, BOOL numericTheory)
/**************************************************************
  INPUT:   A formula (as TERM), a type, and a boolean indicating whether some numeric theory is in effect
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Formula> into the format of the external
           SMT procedure; variables are assumed to be of type
           <type>.
           Uses the other converting functions.
  CAUTION: <Formula> can be an arbitrary formula (probably
           most of the times not in CNF, since FLOTTER was not
           invoked so far).
***************************************************************/
{
  SYMBOL Top;
  ESTRING *result;
  char convert[20];

  result = estring_EmptyString(100);

  Top = term_TopSymbol(Formula);

  if (symbol_Equal(Top, fol_And()) || symbol_Equal(Top, fol_Or()) ||
      symbol_Equal(Top, fol_Equiv()) || symbol_Equal(Top, fol_Implies()) ||
      symbol_Equal(Top, fol_Implied())) {
    LIST Scan;
    if (symbol_Equal(Top, fol_And()))
      estring_AppendString(result, "(and ");
    else if (symbol_Equal(Top, fol_Or()))
      estring_AppendString(result, "(or ");
    else if (symbol_Equal(Top, fol_Equiv()))
      estring_AppendString(result, "(= ");  /* bi-implication is represented by equality */
    else /* (symbol_Equal(Top, fol_Implies())) || (symbol_Equal(Top, fol_Implied())) */
      estring_AppendString(result, "(implies ");

    if (symbol_Equal(Top, fol_Implied())) {
      estring_AppendEStringDelete(result,
				  smt_yices_YicesFormulaToString(term_SecondArgument(Formula), type, numericTheory));
      estring_AppendString(result, " ");
      estring_AppendEStringDelete(result,
				  smt_yices_YicesFormulaToString(term_FirstArgument(Formula), type, numericTheory));
    } else
      for (Scan=term_ArgumentList(Formula);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	estring_AppendEStringDelete(result,
				    smt_yices_YicesFormulaToString((TERM) list_Car(Scan), type, numericTheory));
	estring_AppendString(result, " ");
      }
    estring_AppendString(result, ")");
  } else if (symbol_Equal(Top, fol_Not())) {
    estring_AppendString(result, "(not ");

/*     for (Scan=term_ArgumentList(Formula);!list_Empty(Scan);Scan=list_Cdr(Scan)) { */
    estring_AppendEStringDelete(result,
				smt_yices_YicesFormulaToString(term_FirstArgument(Formula),
							       type, numericTheory));
    estring_AppendString(result, ")");
  } else if (symbol_Equal(Top, fol_True())) {
    estring_AppendString(result, "true");
  } else if (symbol_Equal(Top, fol_False())) {
    estring_AppendString(result, "false");
  } else if (fol_IsQuantifier(Top)) {
    LIST Scan;

    if (symbol_Equal(Top, fol_All()))
      estring_AppendString(result, "(forall (");
    else
      estring_AppendString(result, "(exists (");
    for (Scan=fol_QuantifierVariables(Formula); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      if (term_IsVariable((TERM) list_Car(Scan))) {
	sprintf(convert, "U_%d", *((NAT*) list_Car(Scan)));
	estring_AppendString(result, convert);
	estring_AppendString(result, " :: ");
	estring_AppendString(result, type);
	estring_AppendString(result, " ");
      }
#ifdef CHECK
      else {
	misc_StartErrorReport();
	misc_ErrorReport("\n In smt_yices_YicesFormulaToString:");
	misc_ErrorReport("\n Non-variable argument under quantifier.\n");
	misc_FinishErrorReport();
      }
#endif
    }
    estring_AppendString(result, ") ");
    estring_AppendEStringDelete(result,
				smt_yices_YicesFormulaToString((TERM) term_SecondArgument(Formula), type, numericTheory));

    estring_AppendString(result, ")");
  } else {
    estring_AppendEStringDelete(result, smt_yices_YicesAtomToString(Formula, numericTheory));
  }

  return result;
}

ESTRING* smt_yices_YicesAtomToString(TERM Atom, BOOL numericTheory)
/**************************************************************
  INPUT:   An atom and a boolean indicating whether some numeric theory is in effect
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Atom> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  SYMBOL TopSymbol;
  ESTRING *estring;

  estring = estring_EmptyString(150);

  TopSymbol = term_TopSymbol(Atom);

  if (term_IsComplex(Atom)) {
    estring_AppendString(estring, "(");
    /* putc('(', File); */
    if (TopSymbol == fol_Equality()) {
      estring_AppendString(estring, "=");
      /* fputs("=", File); */
    } else if (! numericTheory) {
      estring_AppendString(estring, _SMTYICES_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
    } else if (TopSymbol == fol_ThLess()) {
      estring_AppendString(estring, "<");
      /* fputs("<", File); */
    } else if (TopSymbol == fol_ThLesseq()) {
      estring_AppendString(estring, "<=");
      /* fputs("<=", File); */
    } else if (TopSymbol == fol_ThGreater()) {
      estring_AppendString(estring, ">");
      /* fputs(">", File); */
    } else if (TopSymbol == fol_ThGreatereq()) {
      estring_AppendString(estring, ">=");
      /* fputs(">=", File); */
/*       /\* SZ > *\/ */
/*     } else if (TopSymbol == fol_ThArrayEqual()) { */
/*       estring_AppendString(estring, "="); */
/*       /\* < SZ *\/ */
    } else {
      estring_AppendString(estring, _SMTYICES_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
      /* symbol_FPrint(File, TopSymbol); */
    }
    estring_AppendString(estring, " ");
    /* putc(' ', File); */
    estring_AppendEStringDelete(estring,
				smt_yices_YicesTermListToString(term_ArgumentList(Atom), numericTheory));
    /* smt_yices_YicesTermListPrint(File, term_ArgumentList(Atom)); */
    estring_AppendString(estring, ")");
    /* putc(')', File); */
  } else if (TopSymbol == fol_True())
    estring_AppendString(estring, "true");
  else if (TopSymbol == fol_False())
    estring_AppendString(estring, "false");
  else {
    /* term_IsConstant(Atom) */
    estring_AppendString(estring, _SMTYICES_IDPREFIX);
    estring_AppendString(estring, symbol_Name(TopSymbol));
  }

  return estring;
}

ESTRING* smt_yices_YicesTermToString(TERM Term, BOOL numericTheory)
/**************************************************************
  INPUT:   A term and a boolean indicating whether some numeric theory is in effect
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <Term> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING *estring;
  char convert[20];
  SYMBOL TopSymbol;

  estring = estring_EmptyString(150);

  if (term_IsComplex(Term)) {
    estring_AppendString(estring, "(");
    /* putc('(', File); */

    TopSymbol = term_TopSymbol(Term);

    if (! numericTheory) {
      estring_AppendString(estring, _SMTYICES_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
    } else if (TopSymbol == fol_ThSucc()) {
      estring_AppendString(estring, "+ ");
      /* fputs("+ ", File); */
      estring_AppendEStringDelete(estring,
				  smt_yices_YicesTermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " 1)");
      /* fputs(" 1)", File); */
      return(estring);
    } else if (TopSymbol == fol_ThPred()) {
      estring_AppendString(estring, "- ");
      /* fputs("- ", File); */
      estring_AppendEStringDelete(estring,
				  smt_yices_YicesTermToString(term_FirstArgument(Term), numericTheory));
      estring_AppendString(estring, " 1)");
      /* fputs(" 1)", File); */
      return estring;
    } else if (TopSymbol == fol_ThUminus()) {
      estring_AppendString(estring, "- 0");
      /* fputs("- 0", File); */
    } else if (TopSymbol == fol_ThPlus()) {
      estring_AppendString(estring, "+");
      /* fputs("+", File); */
    } else if (TopSymbol == fol_ThMinus()) {
      estring_AppendString(estring, "-");
      /* fputs("-", File); */
    } else if (TopSymbol == fol_ThTimes()
               && (term_IsNumeric(term_FirstArgument(Term))
                   || term_IsNumeric(term_SecondArgument(Term)))) {
      TERM t1, t2;
      if (term_IsNumeric(term_FirstArgument(Term))) {
        t1 = term_FirstArgument(Term);
        t2 = term_SecondArgument(Term);
      } else {
        t1 = term_SecondArgument(Term);
        t2 = term_FirstArgument(Term);
      }
      estring_AppendString(estring, "* ");
      /* fputs("* ", File); */
      estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t1, numericTheory));
      estring_AppendString(estring, " ");
      /* fputs(" ", File); */
      estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t2, numericTheory));
      estring_AppendString(estring, ")");
      /* putc(')', File); */
      return estring;
/*     } else if (TopSymbol == fol_ThArraySelect()) { */
/*       TERM t1, t2; */
/*       t1 = term_FirstArgument(Term); */
/*       t2 = term_SecondArgument(Term); */

/*       estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t1, numericTheory)); */
/*       estring_AppendString(estring, " "); */
/*       estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t2, numericTheory)); */
/*       estring_AppendString(estring, ")"); */
/*       return estring; */
/*     } else if (TopSymbol == fol_ThArrayStore()) { */
/*       TERM t1, t2, t3; */
/*       t1 = term_FirstArgument(Term); */
/*       t2 = term_SecondArgument(Term); */
/*       t3 = term_ThirdArgument(Term);       */

/*       estring_AppendString(estring, "update "); */
/*       estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t1, numericTheory)); */
/*       estring_AppendString(estring, " ("); */
/*       estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t2, numericTheory)); */
/*       estring_AppendString(estring, ") "); */
/*       estring_AppendEStringDelete(estring, smt_yices_YicesTermToString(t3, numericTheory)); */
/*       estring_AppendString(estring, ")"); */
/*       return estring; */
/*     } else if (((TopSymbol == fol_ThDivide())/\*  || (TopSymbol == fol_ThDivideNat()) *\/) && */
/* 	       (term_IsNumeric(term_FirstArgument(Term)) &&  */
/* 		term_IsNumeric(term_SecondArgument(Term)))) { */
/*       /\* we directly export a term of the form "divide\(_nat\)(3,4)" (in any case this is  */
/* 	 done later after arithmetic simplification) *\/ */
/*       estring_AppendString(estring, "/"); */
    } else {
      estring_AppendString(estring, _SMTYICES_IDPREFIX);
      estring_AppendString(estring, symbol_Name(TopSymbol));
      /* fputs(symbol_Name(TopSymbol), File); */
    }
    estring_AppendString(estring, " ");
    /* fputs(" ", File); */
    estring_AppendEStringDelete(estring,
				smt_yices_YicesTermListToString(term_ArgumentList(Term), numericTheory));

    estring_AppendString(estring, ")");
    /* putc(')', File); */
  } else if (term_IsVariable(Term)) {
    /* > SZ < */
    sprintf(convert, "U_%d", *((NAT*) Term));
    estring_AppendString(estring, convert);
    /* putc('0', File); */
  } else {
    /* term_IsConstant(Term) */
    if (symbol_denotesInt(term_TopSymbol(Term)) ||
	symbol_denotesRat(term_TopSymbol(Term))) {
      if (! numericTheory) {
        estring_AppendString(estring, _SMTYICES_IDPREFIX);
        /* note that integer numbers can occur outside of a numeric theory, but rationals can't */
      }
      estring_AppendEStringDelete(estring,
				  smt_yices_YicesNumericSymbolToString(term_TopSymbol(Term)));
    } else {
      estring_AppendString(estring, _SMTYICES_IDPREFIX);
      estring_AppendString(estring, symbol_Name(term_TopSymbol(Term)));
    }
  }

  return estring;
}

ESTRING* smt_yices_YicesTermListToString(LIST List, BOOL numericTheory)
/**************************************************************
  INPUT:   A list of terms and a boolean indicating whether some numeric theory is in effect
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts <List> into the format of the external
           SMT procedure.
           Uses the other converting functions.
***************************************************************/
{
  ESTRING* estring;

  estring = estring_EmptyString(150);

  for (; !list_Empty(List); List=list_Cdr(List)) {
    estring_AppendEStringDelete(estring,
				smt_yices_YicesTermToString(list_Car(List), numericTheory));
    if (!list_Empty(list_Cdr(List)))
      estring_AppendString(estring, " ");
  }

  return estring;
}

void     smt_yices_checkIdentifier(char *Name)
/**************************************************************
  INPUT:   A name intended to be used for an identifier.
  EFFECT:  Nothing if <Name> can be used for the name of a
           signature symbol, otherwise, if it is a reserved
           keyword for Yices, an error is reported and the
           program terminates.
  CAUTION: Function "smt_yices_Init" must have been called
           before.
***************************************************************/
{
  LIST Scan;

  for(Scan = smt_yices_keywords; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (string_Equal((char*) list_Car(Scan), Name)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Error, the name \"%s\" is a reserved %s keyword.\n",
			   Name, "Yices");
      misc_FinishUserErrorReport();
    }
  }
}

ESTRING* smt_yices_YicesSigToString(const char* theory)
/**************************************************************
  INPUT:   A theory type specification.
  RETURNS: The result of the conversion as an expandable
           string.
  SUMMARY: Converts the signature to the format of the
           external SMT procedure. Uses <Type> as base type
           for free function symbols and predicates.
***************************************************************/
{
  ESTRING* estring;
  char Type[50];
  BOOL numericTheory;

  /* set type appropriately */
  if (strcmp(theory, "data") == 0) {
    strcpy(Type, "data");
    numericTheory = FALSE;
  } else if (strcmp(theory, "int") == 0) {
    strcpy(Type, "int");
    numericTheory = TRUE;
  } else if (strcmp(theory, "real") == 0) {
    strcpy(Type, "real");
    numericTheory = TRUE;
  }
#ifdef CHECK
  else {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_YicesSigToString:");
    misc_ErrorReport("\n Wrong theory type specification given.\n");
    misc_FinishErrorReport();
  }
#endif

  /* initialize the string to be returned */
  estring = estring_EmptyString(150);

  estring_AppendString(estring, ";; signature \n\n");

  if (! numericTheory) {
    estring_AppendString(estring, "(define-type data)\n\n");
  }

  if (symbol_SignatureExists()) {
    int          i, Index;
    SYMBOL       Info;
    SIGNATURE    Entry;

    for (Index = 1 ; Index < symbol_ACTINDEX ; Index++) {
      Entry  = symbol_Signature(Index);
      if (Entry != NULL) {
	Info = Entry->info;

	switch (symbol_Type(Info)) {
	case symbol_CONSTANT:
          if (symbol_denotesInt(Info) || symbol_denotesRat(Info)) {
            if (! numericTheory) {
              estring_AppendString(estring, "(define ");
              estring_AppendString(estring, _SMTYICES_IDPREFIX);
              estring_AppendEStringDelete(estring,
                                          smt_yices_YicesNumericSymbolToString(Info));
              estring_AppendString(estring, " :: ");
              estring_AppendString(estring, Type);
              estring_AppendString(estring, ")\n");
	    } else {
              ;
	    }
	  } else {
            estring_AppendString(estring, "(define ");
            estring_AppendString(estring, _SMTYICES_IDPREFIX);
            estring_AppendString(estring, Entry->name);
	    estring_AppendString(estring, " :: ");
	    estring_AppendString(estring, Type);
	    estring_AppendString(estring, ")\n");
	  }
	  break;
	case symbol_FUNCTION:
          if (numericTheory && (
                 symbol_Equal(Info, fol_ThSucc())
              || symbol_Equal(Info, fol_ThPred())
              || symbol_Equal(Info, fol_ThPlus())
              || symbol_Equal(Info, fol_ThMinus())
              || symbol_Equal(Info, fol_ThUminus())
	      /* SZ > */
/* 	      || symbol_Equal(Info, fol_ThArraySelect()) */
/* 	      || symbol_Equal(Info, fol_ThArrayStore()) */
	      /* < SZ */

              /* we declare "times" in addition to the built-in "*",
                 so that we can use it in non-linear contexts */
              /* || symbol_Equal(Info, fol_ThTimes()) */
              /* we declare "divide" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
              /* || symbol_Equal(Info, fol_ThDivide()) */
              /* we declare "divide_nat" in addition to the built-in "/",
                 so that we can use it for divisions with non-constant divisor */
	      /* || symbol_Equal(Info, fol_ThDivideNat()) */
              /* || symbol_Equal(Info, fol_ThRemainder()))  */
	      )) {
            continue;
	  }
          estring_AppendString(estring, "(define ");
          estring_AppendString(estring, _SMTYICES_IDPREFIX);
          estring_AppendString(estring, Entry->name);
          estring_AppendString(estring, " :: (-> ");
          /* fprintf(File, "(define %s :: (-> ", Entry->name); */
	  for (i = Entry->arity; i >= 0; i--) {
	    estring_AppendString(estring, Type);
	    estring_AppendString(estring, ((i == 0) ? "))\n" : " "));
	    /* fprintf(File, "%s%s", Type, i == 0 ? "))\n" : " "); */
	  }
	  break;
        case symbol_PREDICATE:
          if (symbol_Equal(Info, fol_Equality())) {
            continue;
	  }
          if (numericTheory && (
                 symbol_Equal(Info, fol_ThLess())
              || symbol_Equal(Info, fol_ThLesseq())
              || symbol_Equal(Info, fol_ThGreater())
              || symbol_Equal(Info, fol_ThGreatereq())
              || symbol_Equal(Info, fol_ThGround())
              || symbol_Equal(Info, fol_True())
              || symbol_Equal(Info, fol_False())
	      /* SZ > */
/* 	      || symbol_Equal(Info, fol_ThArrayEqual()) */
	      /* < SZ */
	      )) {
            continue;
	  }
	  estring_AppendString(estring, "(define ");
          estring_AppendString(estring, _SMTYICES_IDPREFIX);
          estring_AppendString(estring, Entry->name);
	  /* > SZ < predicates of arity 0 should be exported in the form " :: bool"
	            rather than " :: (-> bool)" */
	  if (Entry->arity == 0) {
	    estring_AppendString(estring, " :: ");
	    estring_AppendString(estring, "bool)\n");
	  } else {
	    estring_AppendString(estring, " :: (-> ");
	    /* fprintf(File, "(define %s :: (-> ", Entry->name); */
	    for (i = Entry->arity; i > 0; i--) {
	      estring_AppendString(estring, Type);
	      estring_AppendString(estring, " ");
	      /* fprintf(File, "%s%s", Type, " "); */
	    }
	    estring_AppendString(estring, "bool))\n");
	    /* fprintf(File, "bool))\n"); */
	  }
	  break;
	case symbol_JUNCTOR:
	  ;
	}
      }
    }
  }

  return estring;
}

ESTRING* smt_yices_YicesNumericSymbolToString(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: The string representation of the numeric symbol
           <Symbol>.
  SUMMARY: If symbol denotes an integer number, just <Symbol>,
           if it encodes a rational number <n/m> then <(n/m)>,
           otherwise an error.
  CAUTION: The function only accepts arithmetic symbols as
           input.
***************************************************************/
{
  ESTRING *result;
  MPN *r;

  result = estring_EmptyString(10);

  r = symbol_getValue(Symbol);

  if (symbol_denotesInt(Symbol)) {
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_numref(*r)));
  } else if (symbol_denotesRat(Symbol)) {
    estring_AppendString(result, "( / ");
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_numref(*r)));
    estring_AppendString(result, " ");
    estring_AppendStringDeleteMPN(result, mpz_get_str(NULL, 10, mpq_denref(*r)));
    estring_AppendString(result, ")");
  } else {
    misc_StartErrorReport();
    misc_ErrorReport("\n In smt_yices_YicesNumericSymbolToString:");
    misc_ErrorReport("\n Argument is not a numeric symbol.\n");
    misc_FinishErrorReport();
  }

  return result;
}

void smt_yices_FPrintCnfYices(FILE* File, LIST Clauses, char *type, int evidence)
/**************************************************************
  INPUT:   A file, a list of clauses and a flag store, a flag
           whether unsatisfiable cores are wanted or not.
  RETURNS: Nothing.
  SUMMARY: Prints the clauses to <File> in a format readable by Otter.
***************************************************************/
{
  LIST   scan;
  ESTRING *result;

  result = estring_EmptyString(10);

/*   fol_FPrintOtterOptions(File, Equality, */
/* 			 flag_GetFlagValue(Flags, flag_TDFG2OTTEROPTIONS)); */

  if (Clauses)
    for (scan=Clauses;!list_Empty(scan);scan=list_Cdr(scan)) {
      estring_AppendString(result,
			   ((evidence==1) ? "(assert+ " : "(assert "));
      estring_AppendEStringDelete(result,
				  smt_yices_YicesClauseToString(list_Car(scan),
								type));
      estring_AppendString(result, ")\n");
    }

  fprintf(File, "%s", estring_GetString(result));

  estring_StringFree(result);
}

void smt_yices_FPrintYices(FILE* File, LIST Formulae, FLAG_TDFG2OTTEROPTIONSTYPE Option,
			   char *type, int evidence)
/**************************************************************
  INPUT:   A file, a list of pairs (label.formula), an option
           flag, and a flag whether unsatisfiable cores are
           wanted.
  RETURNS: Nothing.
  SUMMARY: Prints a  the respective formulae in Yices format to <File>.
***************************************************************/
{
  LIST   Scan;
  ESTRING *result;
  BOOL   numericTheory;

  /* fol_FPrintOtterOptions(File, Equality, Option); */

  numericTheory = (strcmp(type, "data") != 0);

  result = estring_EmptyString(10);

  if (!list_Empty(Formulae)) {
    for (Scan=Formulae;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      estring_AppendString(result,
			   ((evidence==1) ? "(assert+ ": "(assert "));
      estring_AppendEStringDelete(result,
				  smt_yices_YicesFormulaToString(list_PairSecond(list_Car(Scan)),
								 type, numericTheory));
      estring_AppendString(result, ")\n");
    }
  }

  fprintf(File, "%s", estring_GetString(result));
  fprintf(File, "\n");

  estring_StringFree(result);
}

