/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *      Connection to GNU Multiprecision Library          * */
/* *                                                        * */
/* *  $Module:   SPASSGMP                                   * */
/* *                                                        * */
/* *  Copyright (C) 2007                                    * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: July 2007                                         * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "spass_gmp.h"
#include "memory.h"

/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

#ifdef SPASS_GMP_DEBUG
/* how big is the memory overhead created by using a GMP libary */
long gmp_memory_allocated_total;
long gmp_memory_used_balance;
long gmp_memory_used_pike;
#endif


/**************************************************************/
/* Internal Functions                                         */
/**************************************************************/

void* gmp_MallocInternal(size_t size)
{
  void *result;

  result = memory_Malloc(size);

#ifdef SPASS_GMP_DEBUG
  gmp_memory_allocated_total +=size;
  gmp_memory_used_balance +=size;
  if (gmp_memory_used_balance > gmp_memory_used_pike)
    gmp_memory_used_pike = gmp_memory_used_balance;
#endif

  return result;
}

void gmp_FreeInternal(void *ptr, size_t size)
{
  memory_Free(ptr, size);

#ifdef SPASS_GMP_DEBUG
  gmp_memory_used_balance -= size;
#endif
}

void* gmp_ReallocInternal(void *ptr, size_t old_size, size_t new_size)
{
  void *result;

  if (ptr == NULL) {
    /* following the ANSI-C standard */
    result = gmp_MallocInternal(new_size);
  } else {
    /* currently SPASS memory management does not support reallocation, so we
       have to simulate it here */

    /* allocate new memory */
    result = gmp_MallocInternal(new_size);

    memcpy(result, ptr, ((old_size < new_size) ? old_size : new_size));

    /* free old memory */
    gmp_FreeInternal(ptr, old_size);
  }

  return result;
}


/**************************************************************/
/* Function Definitions                                       */
/**************************************************************/

void gmp_Init(void)
/**********************************************************
  EFFECT:  Initializes the SPASS GMP module by
           initializing the GNU MP library, in particular
           regarding the memory functions: we advice the
           library to use the SPASS memory management.
***********************************************************/
{
#ifdef SPASS_GMP_DEBUG
  /* initialize internal SPASS_GMP memory bookkeeping */
  gmp_memory_allocated_total = 0;
  gmp_memory_used_pike = 0;
  gmp_memory_used_balance = 0;
#endif

  /* advice GMP to use SPASS memory management */
  mp_set_memory_functions(&gmp_MallocInternal, &gmp_ReallocInternal, &gmp_FreeInternal);

}

void gmp_Free(void)
/**********************************************************
  EFFECT:  If compiler switch <SPASS_GMP_DEBUG> is
           activated, memory statistics regarding GMP
           using is printed. Otherwise, nothing.
***********************************************************/
{
#ifdef SPASS_GMP_DEBUG
  FILE *File;
#endif

#ifdef SPASS_GMP_DEBUG
  /* print status information to stdout */
  File = stdout;
  fputs("\n###\n", File);
  fprintf(File,"### GMP maximum memory usage: %ld Bytes\n",
	  gmp_memory_used_pike/* /memory__KILOBYTE */);
  fprintf(File,"### GMP memory allocated in total: %ld Bytes\n",
	  gmp_memory_allocated_total/* /memory__KILOBYTE */);
  fprintf(File,"### GMP memory balance:  %ld Bytes\n",
	  gmp_memory_used_balance);
  fputs("###\n", File); fflush(File);
#endif
}

char* gmp_MPNToName(MPN r)
/**********************************************************
  INPUT:   An MP number.
  RETURNS: The name of the corresponding function symbol.
  EFFECT:  As convention a rational MP number <n/d> has
           signature name "n/d", an integer MP number just
           "n".
  CAUTION: Memory for the string is allocated.
***********************************************************/
{
  char *num, *denom, *result;

  num = mpz_get_str(NULL, 10, mpq_numref(r));
  denom = mpz_get_str(NULL, 10, mpq_denref(r));

  if (gmp_IsMPNRational(r)) {
    result = memory_Malloc(sizeof(char) * (2+strlen(num)+strlen(denom)));
    sprintf(result, "%s/%s", num, denom);
  } else {
    result = memory_Malloc(sizeof(char) * (1+strlen(num)));
    sprintf(result, "%s", num);
  }

  /* use the internal SPASS_GMP library function in order to maintain memory book keeping */
  gmp_FreeInternal(num, sizeof(char)*(1+strlen(num)));
  gmp_FreeInternal(denom, sizeof(char)*(1+strlen(denom)));

  return result;
}

MPN* gmp_NameToMPN(const char *String)
/**********************************************************
  INPUT:   A signature symbol name.
  RETUNRS: The MP number represented by the
           signature name string <String>.
  EFFECT:  Uses the function <gmp_CreateMPN>.
***********************************************************/
{
  signed long int num;
  unsigned long int denom;

  if (strstr(String, "/") == NULL) {
    /* integer MP number since no denominator is represented in string */
    sscanf(String, "%ld", &num);
    denom = 1;
  } else
    /* rational MP number */
    sscanf(String, "%ld/%lud", &num, &denom);

  return (gmp_CreateMPN(num, denom));
}

MPN* gmp_AllocMPN(void)
/**************************************************************
  RETURNS: A newly allocated pointer to a MPN.
  EFFECT:  Allocated memory for the pointer and GMP memory,
           i.e. uses the GMP "Init" function.
  CAUTION: Use <gmp_FreeRat> to disallocate the MP number
           again.
***************************************************************/
{
  MPN *result;

  result = (MPN*) memory_Malloc(sizeof(MPN));

  mpq_init(*result);

  return result;
}

MPN* gmp_CreateMPN(signed long int num, unsigned long int denom)
/**********************************************************
  INPUT:   Two longs, one signed an one unsigned.
  RETUNRS: The by the GMP libary allocated MP number
           <num/denom> (which could be an MP integer or an
           MP rational).
  EFFECT:  Uses <gmp_AllocMPN>.
  CAUTION: Use <gmp_FreeRat> to disallocate the MP number
           again.
           Negative ratioanl MP numbers must be allocated
           with negative numerator and positive
           denominator.
***********************************************************/
{
  MPN *result;

  result = gmp_AllocMPN();

  mpq_set_si(*result, num, denom);
  mpq_canonicalize(*result);

  return result;
}

void gmp_FreeMPN(MPN* r)
/**********************************************************
  INPUT:   A poiner to an MP number.
  EFFECT:  Advices the GMP libary to free its
           representation of <r>; additioanlly the slot
           for <r> is freed.
***********************************************************/
{
  /* first free internal GMP representation ...*/
  mpq_clear(*r);

  /* ...then associated slot */
  memory_Free(r, sizeof(MPN));
}

BOOL gmp_IsMPNRational(MPN r)
/**********************************************************
  INPUT:   A rational number.
  EFFECT:  If <r> is actually a rational number, ie. if
           <r>'s denominator is unequal to 1.
  CAUTION: <r> is not subject to canoncalization. If it is
           not canonicalzed before, the function might
           deliver wrong results.
***********************************************************/
{
  mpz_t one;
  BOOL result;

  mpz_init(one);
  mpz_set_si(one, 1);

  if (mpz_cmp(mpq_denref(r), one) == 0)
    result = FALSE;
  else
    result = TRUE;

  mpz_clear(one);

  return result;
}

BOOL gmp_TryMPNToInt(MPN mpn, int *i)
/**********************************************************
  INPUT:   A multiprecesion number, an integer number <i>
           used as one of the return values.
  RETURNS: 0 if <mpn> is either a non-integer number (thus
           a canonicalized rational number with
           denominator != 1) or an integer which is too
           big to fit into the standard C int type; 1
           otherwise.
  EFFECT:  If <mpn> can fit into <i>, then <mpn> is
           converted into <i>.
  CAUTION: <mpn> must be canonicalized.
***********************************************************/
{
  BOOL result;
  signed long int sli_tmp;

  if (gmp_IsMPNRational(mpn))
    result = FALSE;
  else if (mpz_fits_sint_p(mpq_numref(mpn)) == 0)
    result = FALSE;
  else {
    result = TRUE;
    sli_tmp = mpz_get_si(mpq_numref(mpn));
    *i = (int) sli_tmp;
  }

  return result;
}

BOOL gmp_StringIsMPN(char* String)
/**********************************************************
  INPUT:   A string.
  RETURNS: Wehther <String> can be converted to a MPN in
           our understanding, i.e. has the form
           "(-)\d+(/\d+))" i.e. is a fraction, TRUE is
           returned or FALSE otherwise.
  EFFECT:  It uses the GMP function <mpq_set_str>.
***********************************************************/
{
  int result;
  mpq_t temp;

  mpq_init(temp);

  result = mpq_set_str(temp, String, 10);

  mpq_clear(temp);

  if (result == 0)
    return TRUE;
  else
    return FALSE;
}

mpz_t* gmp_RoundUpwards(MPN r)
/**************************************************************
  INPUT:   An MPN number.
  RETURNS: The upwardly rounded number <r>.
***************************************************************/
{
  mpz_t *result;

  result = (mpz_t*) memory_Malloc(sizeof(mpz_t));

  mpz_init(*result);
  mpz_cdiv_q(*result, mpq_numref(r), mpq_denref(r));

  return(result);
}
