/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *       CONVERTER FROM DFG SYNTAX TO YICES SYNTAX        * */
/* *                                                        * */
/* *  $Module:   TDFG2YICES                                 * */
/* *                                                        * */
/* *  Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001      * */
/* *  MPI fuer Informatik, Stephan Zimmer                   * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision:                                               * */
/* $State:                                                  * */
/* $Date: August 2007                                       * */
/* $Author: Stephan Zimmer                                  * */
/* *                                                        * */
/* *                                                        * */
/* * HIGHLY based on tdfg2otter.c                           * */
/* * Authors: Thomas Hillenbrand <hillen@mpi-inf.mpg.de>    * */
/* *          Dalibor Topic <topc@mpi-inf.mpg.de>           * */
/* *          Christoph Weidenbach <weidenb@mpi-inf.mpg.de> * */
/* *                                                        * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Stephan Zimmer                             * */
/* *             MPI fuer Informatik                        * */
/* *             Stuhlsatzenhausweg 85                      * */
/* *             66123 Saarbruecken                         * */
/* *             Email: st.zimmer@web.de                    * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfg.h"
#include "foldfg.h"
#include "clause.h"
#include "options.h"
#include "smt_yices.h"

#include <errno.h>

#define DFG2YICES__VERSION "0.7"

/**************************************************************/
/* Main Function                                              */
/**************************************************************/

int main(int argc, const char* argv[])
{
  LIST       Scan, Scan1, Clauses, AxiomList,ConjectureList, SortList, UserPrecedence;
  FILE       *Output,*Input;
  const char *Filename;
  char       ConLabel[] = "Conjecture";
  FLAGSTORE  Flags;
  PRECEDENCE Precedence;
  ESTRING    *signature;
  char       type[20];

  memory_Init(memory__UNLIMITED);
  /* SZ > */
  gmp_Init();
  /* < SZ */
  atexit(memory_FreeAllMem);
  symbol_Init(TRUE);
  stack_Init();
  term_Init();
  flag_Init();
  /* SZ > */
  smt_yices_Init();
  /* < SZ */

  Flags      = flag_CreateStore();
  flag_InitStoreByDefaults(Flags);
  Precedence = symbol_CreatePrecedence();

  fol_Init(Precedence);
  clause_Init();
  opts_Init();


  opts_DeclareSPASSFlagsAsOptions();
  if (!opts_Read(argc, argv))
    return EXIT_FAILURE;

  opts_SetFlags(Flags);
  opts_Free();

  if (!flag_GetFlagValue(Flags, flag_STDIN) && argc != opts_Indicator()+2) {
    fputs("\n\t          tdfg2yices Version ", stdout);
    fputs(DFG2YICES__VERSION, stdout);
    puts("\n\t       Usage: tdfg2yices [options] <input-file> <output-file>\n");
    return EXIT_FAILURE;
  }

  AxiomList = ConjectureList = SortList = UserPrecedence = list_Nil();

  if (flag_GetFlagValue(Flags, flag_DPCTHEORY) == 1)
    strcpy(type, "int");
  else if (flag_GetFlagValue(Flags, flag_DPCTHEORY) == 2)
    strcpy(type, "real");
  else {
    printf("Please specify the background theory using option \"DpcTheory\"\n");
    return EXIT_FAILURE;
  }

  if (!flag_GetFlagValue(Flags, flag_STDIN)) {
    Input = misc_OpenFile(argv[opts_Indicator()],"r");

    Clauses = dfg_DFGParser(Input,Flags,Precedence,&AxiomList,&ConjectureList,
			    &SortList, &UserPrecedence);

    misc_CloseFile(Input,argv[opts_Indicator()]);
    AxiomList = list_Nconc(AxiomList, SortList);
    if (!list_Empty(ConjectureList)) {
      /* Build conjecture formula and negate it: conjectures are taken disjunctively!!*/
      for (Scan = ConjectureList; !list_Empty(Scan); Scan = list_Cdr(Scan))
	list_Rplacd(list_Car(Scan), (LIST)term_Create(fol_Not(),
						      list_List(list_PairSecond(list_Car(Scan)))));
      if (!list_Empty(list_Cdr(ConjectureList))) {
	for (Scan = ConjectureList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  Scan1 = list_Car(Scan);
	  list_Rplaca(Scan, list_PairSecond(list_Car(Scan)));
	  if (list_PairFirst(Scan1) != NULL)
	    string_StringFree(list_PairFirst(Scan1));  /* Free the label */
	  list_PairFree(Scan1);                      /* Free the pair */
	}
	ConjectureList = list_List(list_PairCreate(string_StringCopy(ConLabel),
						   term_Create(fol_Or(), ConjectureList)));
      }
    }


    AxiomList = list_Nconc(ConjectureList, AxiomList);
    Filename = argv[opts_Indicator()+1];
    Output = misc_OpenFile(Filename,"w");
    fputs("\n;; Generated by tdfg2yices ", Output);
    fputs(DFG2YICES__VERSION, Output);
    fputs("\n\n", Output);

    signature = smt_yices_YicesSigToString(type);

    fprintf(Output, estring_GetString(signature));
    fprintf(Output, "\n");
    estring_StringFree(signature);

    fputs("\n;; formulae\n\n", Output);

    if (list_Empty(AxiomList))
      smt_yices_FPrintCnfYices(Output, Clauses, type,
			       flag_GetFlagValue(Flags, flag_DPCEVIDENCE));
    else
      smt_yices_FPrintYices(Output, AxiomList,
			    flag_GetFlagValue(Flags, flag_TDFG2OTTEROPTIONS), type,
			    flag_GetFlagValue(Flags, flag_DPCEVIDENCE));
    dfg_DeleteFormulaPairList(AxiomList);

    fprintf(Output, "\n\n;;check the context\n(check)\n");


    misc_CloseFile(Output,Filename);
  }
  else {
    Clauses   = dfg_DFGParser(stdin,Flags,Precedence,&AxiomList,&ConjectureList,
			      &SortList, &UserPrecedence);
    AxiomList = list_Nconc(AxiomList, SortList);
    if (!list_Empty(ConjectureList)) {
      /* Build conjecture formula and negate it: conjectures are taken disjunctively!!*/
      for (Scan = ConjectureList; !list_Empty(Scan); Scan = list_Cdr(Scan))
	list_Rplacd(list_Car(Scan), (LIST)term_Create(fol_Not(),
						      list_List(list_PairSecond(list_Car(Scan)))));
      if (!list_Empty(list_Cdr(ConjectureList))) {
	for (Scan = ConjectureList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  Scan1 = list_Car(Scan);
	  list_Rplaca(Scan, list_PairSecond(list_Car(Scan)));
	  if (list_PairFirst(Scan1) != NULL)
	    string_StringFree(list_PairFirst(Scan1));  /* Free the label */
	  list_PairFree(Scan1);                      /* Free the pair */
	}
	ConjectureList = list_List(list_PairCreate(string_StringCopy(ConLabel),
						   term_Create(fol_Or(), ConjectureList)));
      }
    }
    AxiomList = list_Nconc(ConjectureList, AxiomList);
    Output    = stdout;
    fputs("\n;; Generated by tdfg2yices ", Output);
    fputs(DFG2YICES__VERSION, Output);
    fputs("\n\n", Output);

    signature = smt_yices_YicesSigToString(type);

    fprintf(Output, estring_GetString(signature));
    fprintf(Output, "\n");
    estring_StringFree(signature);

    fputs("\n;; formulae\n\n", Output);

    if (list_Empty(AxiomList))
      smt_yices_FPrintCnfYices(Output, Clauses, type,
			       flag_GetFlagValue(Flags, flag_DPCEVIDENCE));
    else
      smt_yices_FPrintYices(Output, AxiomList,
			    flag_GetFlagValue(Flags, flag_TDFG2OTTEROPTIONS), type,
			    flag_GetFlagValue(Flags, flag_DPCEVIDENCE));
    dfg_DeleteFormulaPairList(AxiomList);

    fprintf(Output, "\n\n;; check the context\n(check)\n");
  }

  clause_DeleteClauseList(Clauses);
  /*symbol_Dump();*/
  flag_DeleteStore(Flags);
  symbol_DeletePrecedence(Precedence);
  list_Delete(UserPrecedence);

  /* SZ > */
  smt_yices_Free();
  /* < SZ */
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();

  gmp_Free();

#ifdef CHECK
  memory_Print();
#endif
  return 0;
}
